/* CRYPTOGRAM.H */
/* TODO: gestione caratteri non-ascii, tipologie dizionario e ibrida, gestione errori */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "substs.h"


class Cryptogram {
 public:
  Cryptogram();
  ~Cryptogram();
  char *encode(char *text);
  char *decode(char *cypher);
  void set_freqtable(FreqTable *ft) { this->ft = ft; };
  void set_lang(char *lang) { this->lang = strdup(lang); }
  void set_subst(int symbol, char c);

 private:
  int flag_obliged;
  double digram_threshold, trigram_threshold, ngram_tolerance;
  FreqTable *ft;
  TableSubst *subst, *tsubst;
  char *lang;
  //Dict *dict;

  int *convert_cypher(char *cypher, int *len);
  double *calc_cypher_freq(int *cypher, int cypher_len, int *len);
  void calc_subst(double *freqs, int freqs_len, double delta, int maxlen);
  void gen_words(int *cypher, int cypher_len, int pos, char lastchars[5]);

  static int cmp_freqs(const void *s1, const void *s2);
};

Cryptogram::Cryptogram() {
  subst = new TableSubst();
  lang = NULL;
  flag_obliged = 0;
  digram_threshold =  0.000001;     // soglia minima per i digrammi validi
  trigram_threshold = 0.00000001;   // soglia minima per i trigrammi validi
  //queue_len = 20;                 // numero di frasi da tenere nella coda dei risultati
  ngram_tolerance = 0.005;          // ammette questa percentuale di ngrammi non validi
}

Cryptogram::~Cryptogram() {
  delete subst;
  if (tsubst) delete tsubst;
  if (lang) free(lang);
}

char *Cryptogram::encode(char *text) {
  int pos=0, l = strlen(text);
  char c, *ret = (char *)malloc(l*3);

  int sym_counter = 1, sym_table[26];
  memset(sym_table, 0xff, 26*sizeof(int));

  for (int i=0; i<l; i++) {
    c = tolower(text[i]);
    if (isalpha(c)) {
      if (sym_table[c-'a'] == -1) sym_table[c-'a'] = sym_counter++;
      pos += sprintf(ret+pos, "%i ", sym_table[c-'a']);
    } else {
      ret[pos++] = c;
      if (c == ' ') ret[pos++] = ' ';
    }
  }
  ret[pos] = 0;
  ret = (char *)realloc(ret, pos+1);
  return ret;
}

char *Cryptogram::decode(char *cypher) {
  // Converte la stringa cifrata in un array numerico
  int cypher_list_len, *cypher_list;
  cypher_list = convert_cypher(cypher, &cypher_list_len);

  // Calcola le frequenze dei simboli
  int cypher_freq_len;
  double *cypher_freq;
  cypher_freq = calc_cypher_freq(cypher_list, cypher_list_len, &cypher_freq_len);

  double deltas[] = {0.02, 0.04, 0.06, 0.08, 0.10};
  int lens[]      = {8,    9,    10,   12,   14};

  // Calcola la tabella di sostituzione
  calc_subst(cypher_freq, cypher_freq_len, deltas[2], lens[2]);

  // Calcola le frasi possibili
  gen_words(cypher_list, cypher_list_len, 0, "");
//printf("%s\n", cypher);
//for (int i=0; i<cypher_freq_len; i++) printf("%i :  %f\n", i, cypher_freq[i]);

  free(cypher_freq);
  free(cypher_list);

  return NULL;
}

void Cryptogram::set_subst(int symbol, char c) {
  flag_obliged = 1;
  subst->add_subst(symbol, c, 2.0);
}

/* Converte il testo cifrato in un array numerico, di più facile gestione */
int *Cryptogram::convert_cypher(char *cypher, int *len) {
  char buff[10] = "", firstblank = 1, secondblank = 0;
  int blen = 0, l = strlen(cypher);
  int *ret = (int *)malloc(l*sizeof(int));
  *len = 0;

  for (int i=0; i<l; i++) {
    char c = cypher[i];
    if (isdigit(c)) {
      if (blen >= 9) continue;
      buff[blen++] = c;
      buff[blen] = 0;
      firstblank = 1;
    } else {
      if (firstblank) {
        ret[(*len)++] = atoi(buff);
        buff[0] = 0;
        blen = 0;
        firstblank = 0;
        secondblank = 1;
      } else if (secondblank) {
        ret[(*len)++] = 0;
        secondblank = 0;
      }
    }
  }
  ret = (int *)realloc(ret, sizeof(int)**len);
  #ifdef DEBUG
  printf("ARRAY CIFRATO:\n");
  for (int k=0; k<*len; k++) printf("%3i", ret[k]);
  printf("\n");
  fflush(stdout);
  #endif
  return ret;
}

/* Calcola le frequenze dei simboli nel testo cifrato */
double *Cryptogram::calc_cypher_freq(int *cypher, int cypher_len, int *len) {
  int i,maxsym = 0;
  double *ret;

  for (i=0; i<cypher_len; i++) if (cypher[i] > maxsym) maxsym = cypher[i];
  ret = (double *)malloc((maxsym+2)*sizeof(double));
  for (i=0; i<maxsym+1; i++) ret[i] = 0.0;

  for (i=0; i<cypher_len; i++) ret[cypher[i]] += 1.0;
  for (i=0; i<maxsym+1; i++) ret[i] /= (double)cypher_len;

  *len = maxsym+1;
  return ret;
}

/* Calcola la tabella delle sostituzioni in base alle frequenze standard */
void Cryptogram::calc_subst(double *freqs, int freqs_len, double delta, int maxlen) {
  int i = 0, j;

  struct f {
    char c[4];
    double freq;
  } freqtab[27], deltas[27], mids[27];
  char letters[28] = "abcdefghijklmnopqrstuvwxyz";

  if (!flag_obliged) subst->clear();

  FreqLang *lt = ft->get_langtable(this->lang);
  while (!lt->atend(0)) {
    lt->get_next_freq(0, (char *)(freqtab[i].c), &(freqtab[i].freq));
    i++;
  }

  subst->add_subst(0, ' ', 2.0);
  for (i=1; i<freqs_len; i++) {
    // Calcola le differenze
    for (j=1; j<=26; j++) {
      deltas[j].c[0] = freqtab[j].c[0];
      deltas[j].freq = fabs(freqtab[j].freq - freqs[i]);
    }
    qsort(deltas+1, 26, sizeof(f), cmp_freqs);
    // Calcola le sostituzioni
    if (!subst->is_present(i)) {
      mids[i].freq = 0.0;
      for (j=0; j<MIN(maxlen, freqs_len); j++)
        if (deltas[j+1].freq <= delta) {
          subst->add_subst(i, deltas[j+1].c[0], 1.0-deltas[j+1].freq);
          str_replace(deltas[j+1].c[0], '.', letters);
          mids[i].freq += 1.0-deltas[j+1].freq;
        }
    } else {
      mids[i].freq = 2.0;
    }
    mids[i].c[0] = i;
    mids[i].freq /= subst->count(i);
  }
  // Aggiunge le lettere rimaste alle altre sostituzioni
  str_remove('.', letters);
  if (letters[0]) {
    qsort(mids+1, freqs_len-1, sizeof(f), cmp_freqs);
    int l = strlen(letters);
    for (j=1; j<=MIN(l, freqs_len-1); j++) {
      int s = mids[j].c[0];
      for (int k=0; k<l; k++)
        subst->add_subst(s, letters[k], 0.001);
    }
  }
  //printf("LETTERE ESCLUSE: %s\n", letters);
  //subst->print();
}

void Cryptogram::gen_words(int *cypher, int cypher_len, int pos, char lastchars[5]) {
  int i,j, c;
  Subst sub;

  //printf("GEN: pos=%i sym=%i\n", pos, cypher[pos]);

  // Prima iterazione
  if (pos == 0) {
    if (tsubst) tsubst->clear(); else tsubst = new TableSubst();

  // Ultima iterazione
  } else if (pos >= cypher_len) {
    char text[cypher_len+1];
    for (j=0; j<cypher_len; j++) {
      SymbolSubst *sys = tsubst->get_subst(cypher[j]);
      sys->reset();
      sys->get_next(&sub);
      text[j] = sub.c;
    }
    text[cypher_len+1] = 0;
    printf("%s\n", text);
    return;
  }

//---------------------------------------------
/*
    char text[cypher_len+1];
    memset(text, '.', cypher_len);
    for (j=0; j<cypher_len; j++) {
      if (tsubst->is_present(cypher[j])) {
        SymbolSubst *sys = tsubst->get_subst(cypher[j]);
        sys->reset();
        sys->get_next(&sub);
        text[j] = sub.c;
      }
    }
    text[cypher_len+1] = 0;
    printf("%s\n", text);
*/
//---------------------------------------------

  c = cypher[pos];
  SymbolSubst *ss = new SymbolSubst(subst->get_subst(c));
  while (!ss->atend()) {
    ss->get_next(&sub);
    if (tsubst->is_char_present(sub.c)) continue;

    tsubst->add_subst(c, sub.c, sub.freq);
    gen_words(cypher, cypher_len, pos+1, "");
    tsubst->get_subst(c)->del_subst(sub.c);
  }
  delete ss;
}

int Cryptogram::cmp_freqs(const void *s1, const void *s2) {
  struct f {
    char c[4];
    double freq;
  };

  if ( ((f *)s1)->freq > ((f *)s2)->freq ) return 1;
  else if ( ((f *)s1)->freq < ((f *)s2)->freq ) return -1;
  else return 0;
}
