#include "dict.h"
#include <stdlib.h>

using std::ifstream;
using std::ostream;
using std::string;


// ---- CONSTRUCTORS ---------------------------------------------------------

// Standard constructor
Dict::Dict(): words(0), table(0), longest_word(-1), shortest_word(-1) {
}

// Constructor: read from file
Dict::Dict(const string& filename): words(0), table(0), longest_word(-1), shortest_word(-1) {
  ifstream fdic(filename.c_str());
  if (!fdic) return;
  while (!fdic.eof()) {
    string t;
    if ((words & 0xff)==0) table = (string**)realloc(table, sizeof(string*)*(words+0x100));
    getline(fdic, t);
    if (t.empty()) continue;
    table[words++] = new string(t);
    if (words == 1) longest_word = shortest_word = 0;
    if (t.length() > table[longest_word]->length()) longest_word = words-1;
    if (t.length() < table[shortest_word]->length()) shortest_word = words-1;
  }
  fdic.close();
}

// Constructor: copy
Dict::Dict(const Dict& d): words(d.words), longest_word(d.longest_word), shortest_word(d.shortest_word) {
  table = (string**)malloc(sizeof(string*)*words);
  for (int i=0; i<words; i++)
    table[i] = new string(d[i]);
}

// Destructor
Dict::~Dict() {
  for (int i=0; i<words; i++)
    delete table[i];
}


// ---- OPERATORS ------------------------------------------------------------

// <<
ostream& operator << (ostream& os, const Dict& d) {
  for (int i=0; i<d.words; i++) {
    os << *(d.table[i]) << std::endl;
  }
  return os;
}

// []
string Dict::operator [] (int i) const {
  if (i < 0 || i >= words) return string();
  return *(table[i]);
}

// =
Dict& Dict::operator = (const Dict& d) {
  for (int i=0; i<words; i++) delete table[i];
  words = d.words;
  table = (string**)realloc(table, sizeof(string*)*words);
  for (int i=0; i<words; i++)
    table[i] = new string(d[i]);
  longest_word = d.longest_word;
  shortest_word = d.shortest_word;
}

// +
Dict operator + (const Dict& d1, const Dict& d2) {
  Dict ret(d1);
  ret.merge(d2);
  return ret;
}

// -
Dict operator - (const Dict& d1, const Dict& d2) {
  Dict ret(d1);
  ret.subtract(d2);
  return ret;
}

// ==
bool operator == (const Dict& d1, const Dict& d2) {
  if (d1.words != d2.words) return false;
  for (int i=0; i<d1.words; i++)
    if (d1[i] != d2[i]) return false;
  return true;
}

// !=
bool operator != (const Dict& d1, const Dict& d2) {
  return !(d1 == d2);
}

// ---- PUBLIC METHODS -------------------------------------------------------

// Number of words in dictionary
int Dict::length() const {
  return words;
}

// Returns position of s in dict or -1 if not found
int Dict::find(const string& s) const {
  string **p = (string**)bsearch(&s, table, words, sizeof(string*), cmp_stringp);
  return p?(p-table):(-1);
}

// Returns true if the string s is in the dictionary
bool Dict::is_present(const string& s) const {
  return (find(s) >= 0);
}

// Returns the longest word in the dictionary
string Dict::get_longest() {
  if (longest_word < 0) calc_longest();
  if (longest_word < 0) return string();
  return *(table[longest_word]);
}

// Returns the shortest word in the dictionary
string Dict::get_shortest() {
  if (shortest_word < 0) calc_longest();
  if (shortest_word < 0) return string();
  return *(table[shortest_word]);
}

// Insert a word in the dictionary
void Dict::insert(const string& s) {
  int p0=0, p1=words-1, pm=0;
  while (p0 < p1) {
    pm = (p1+p0)>>1;
    if (pm == p0) {
      if (s > *(table[p1])) pm = p1;
      break;
    }
    if (s >= *(table[pm])) p0 = pm; else p1 = pm;
  }
  if (words == 0) {
    table = (string**)realloc(table, sizeof(string*)*0x100);
    table[words++] = new string(s);
    longest_word = shortest_word = 0;
  } else if (s != *(table[pm])) {
    if ((words & 0xff)==0) table = (string**)realloc(table, sizeof(string*)*(words+0x100));
    if (pm == 0 && s < *(table[0])) pm = -1;
    memmove(table+pm+2, table+pm+1, sizeof(string*)*(words-pm-1));
    table[pm+1] = new string(s);
    words++;
    if (longest_word < 0) longest_word = pm+1; else
      if (s.length() > table[longest_word]->length()) longest_word = pm+1;
    if (shortest_word < 0) shortest_word = pm+1; else
      if (s.length() > table[shortest_word]->length()) shortest_word = pm+1;
  }
}

// Removes s from dictionary
void Dict::remove(const string& s) {
  int pos = find(s);
  if (pos >= 0) {
    if (pos < (words-1)) 
      memmove(table+pos, table+pos+1, sizeof(string*)*(words-pos-1));
    words--;
    if (pos == longest_word) longest_word = -1;
    if (pos == shortest_word) shortest_word = -1;
    if (words == 0) {
      free(table);
      table = 0;
      longest_word = shortest_word = -1;
    }
  }
}

// Removes all words from dictionary
void Dict::remove_all() {
  for (int i=0; i<words; i++)
    delete table[i];
  free(table);
  table = 0;
  words = 0;
  longest_word = shortest_word = -1;
}

// Merges d with current dictionary
void Dict::merge(const Dict& d) {
  int i1=0, i2=0, twords=0;
  string **tab2 = 0;
  //longest_word = shortest_word = -1;
  // ------ finish here
  while (i1 < words && i2 < d.words) {
    string *s1 = table[i1];
    string *s2 = d.table[i2];
    if ((twords & 0xff)==0) tab2 = (string**)realloc(tab2, sizeof(string*)*(twords+0x100));
    if (*s1 <= *s2) {
      tab2[twords++] = s1;
      if (words == 1) {
        // todo
        //longest_word = shortest_word = 0;
      } else {
      }
      i1++;
      if (*s1 == *s2) i2++;
    } else {
      tab2[twords++] = new string(*s2);
      i2++;
    }
  }
  if (i1 >= words) {
    for (;i2<d.words; i2++) {
      if ((twords & 0xff)==0) tab2 = (string**)realloc(tab2, sizeof(string*)*(twords+0x100));
      tab2[twords++] = new string(d[i2]);
    }
  } else {
    for (;i1<words; i1++) {
      if ((twords & 0xff)==0) tab2 = (string**)realloc(tab2, sizeof(string*)*(twords+0x100));
      tab2[twords++] = table[i1];
    }
  }
  free(table);
  table = tab2;
  words = twords;
}

// Removes words that are commond with d
void Dict::subtract(const Dict& d) {
  int i1=0, i2=0, twords=0;
  string **tab2 = 0;
  while (i1 < words && i2 < d.words) {
    string *s1 = table[i1];
    string *s2 = d.table[i2];
    if ((twords & 0xff)==0) tab2 = (string**)realloc(tab2, sizeof(string*)*(twords+0x100));
    if (*s1 < *s2) {
      tab2[twords++] = s1;
      i1++;
    } else  {
      if (*s1 == *s2) i1++; 
      i2++;
    }
  }
  if (i2 >= d.words) {
    for (;i1<words; i1++) {
      if ((twords & 0xff)==0) tab2 = (string**)realloc(tab2, sizeof(string*)*(twords+0x100));
      tab2[twords++] = table[i1];
    }
  }
  free(table);
  table = tab2;
  words = twords;
}

// Keeps words commons with d
void Dict::intersect(const Dict& d) {
  int i1=0, i2=0, twords=0;
  string **tab2 = 0;
  while (i1 < words && i2 < d.words) {
    string *s1 = table[i1];
    string *s2 = d.table[i2];
    if (*s1 < *s2) {
      i1++;
    } else {
      if (*s1 == *s2) {
        if ((twords & 0xff)==0) tab2 = (string**)realloc(tab2, sizeof(string*)*(twords+0x100));
        tab2[twords++] = s1;
        i1++;
      }
      i2++;
    }
  }
  free(table);
  table = tab2;
  words = twords;
}

// Returns a new dict with words matching regexp re
Dict Dict::regex_match(const std::string& re, int flags) const {
  int v;
  Dict ret;
  regex_t rt;
  v = regcomp(&rt, re.c_str(), flags | REG_NOSUB);
  if (v != 0) return ret;
  for (int i=0; i<words; i++) {
    v = regexec(&rt, table[i]->c_str(), 0, NULL, 0);
    if (v == 0) ret.insert_tail(*(table[i]));
  }
  regfree(&rt);
  return ret;
}

// ---- PRIVATE METHODS ------------------------------------------------------

// Find the longest and shortest word
void Dict::calc_longest() {
  longest_word = shortest_word = -1;
  if (words == 0) return;
  longest_word = shortest_word = 0;
  for (int i=1; i<words; i++) {
    if (table[i]->length() > table[longest_word]->length()) longest_word = i;
    if (table[i]->length() < table[shortest_word]->length()) shortest_word = i;
  }
}

void Dict::insert_tail(const string& s) {
  if ((words & 0xff)==0) table = (string**)realloc(table, sizeof(string*)*(words+0x100));
  table[words++] = new string(s);
}

// Compares 2 strings pointed by s1 and s2 ~~
int Dict::cmp_stringp(const void *s1, const void *s2) {
  return ((string*)s1)->compare(**((string**)s2));
}
