/* SUBSTS.H */

/* Una sostituzione */
struct Subst {
  char c;
  double freq;
};

/* Gestisce le sostituzioni per un simbolo */
class SymbolSubst {
 public:
  SymbolSubst(int symbol);
  SymbolSubst(SymbolSubst *ss);
  int add_subst(char c, double freq);
  int del_subst(char c);
  int is_present(char c);
  int count() { return subst_len; }
  int atend() { return ptr >= subst_len; }
  void get_next(Subst *s);
  void reset() { ptr = 0; }
  void sort(int reverse);

  /* ---DEBUG------------------- */
  void print();

 private:
  int ptr;
  int symbol;
  int subst_len;
  int subst_pos[27];
  Subst subst[27];

  void sync_indexes();
  static int cmp(const void *s1, const void *s2);
  static int rcmp(const void *s1, const void *s2);
};

/*--DEBUG------------*/
void SymbolSubst::print() {
  printf("SYMBOL: %i\n", symbol);
  for (int i=0; i<subst_len; i++)
    printf("%c %f,  ", subst[i].c, subst[i].freq);
  printf("\n");
}
/*------------------------*/

SymbolSubst::SymbolSubst(int symbol) {
  this->symbol = symbol;
  subst_len = 0;
  ptr = 0;
  memset(subst, '.', sizeof(subst));
  memset(subst_pos, 0xff, sizeof(subst_pos));
}

SymbolSubst::SymbolSubst(SymbolSubst *ss) {
  this->symbol = ss->symbol;
  this->subst_len = ss->subst_len;
  ptr = 0;
  memcpy(this->subst_pos, ss->subst_pos, sizeof(this->subst_pos));
  memcpy(this->subst, ss->subst, sizeof(this->subst));
}


int SymbolSubst::add_subst(char c, double freq) {
  int i = (c == ' ')?0:(c-'a'+1);
  if (subst_pos[i] != -1) {
    subst[subst_pos[i]].freq = freq;
  } else {
    subst_pos[i] = subst_len;
    subst[subst_len].c = c;
    subst[subst_len++].freq = freq;
  }
  return i;
}

int SymbolSubst::del_subst(char c) {
  int i = (c == ' ')?0:(c-'a'+1);
  int p = subst_pos[i];
  if (p == -1) return i;
  memmove(subst+p, subst+p+1, sizeof(Subst)*(subst_len-p-1));
  if (ptr > p) ptr--;
  subst_pos[i] = -1;
  subst_len--;
  sync_indexes();
  return i;
}

void SymbolSubst::get_next(Subst *s) {
  if (atend()) return;
  *s = subst[ptr++];
}

int SymbolSubst::is_present(char c) {
  int i = (c == ' ')?0:(c-'a'+1);
  return (subst_pos[i] != -1);
}

void SymbolSubst::sort(int reverse) {
  qsort(subst, subst_len, sizeof(Subst), reverse?&rcmp:&cmp);
  sync_indexes();
}

void SymbolSubst::sync_indexes() {
  for (int i=0; i<subst_len; i++) subst_pos[(subst[i].c == ' ')?0:(subst[i].c-'a'+1)] = i;
}

int SymbolSubst::cmp(const void *s1, const void *s2) {
  Subst *su1 = (Subst *)s1;
  Subst *su2 = (Subst *)s2;
  if (su1->freq < su2->freq) return -1;
  else if (su1->freq > su2->freq) return +1;
  else return 0;
}

int SymbolSubst::rcmp(const void *s1, const void *s2) {
  Subst *su1 = (Subst *)s1;
  Subst *su2 = (Subst *)s2;
  if (su1->freq < su2->freq) return +1;
  else if (su1->freq > su2->freq) return -1;
  else return 0;
}


/* Una tabella di sostituzioni */
class TableSubst {
 public:
  TableSubst();
  ~TableSubst();
  void clear();
  void add_subst(int symbol, char c, double freq);
  SymbolSubst *get_subst(int symbol) { return subst[symbol]; }
  int is_present(int symbol) { return (subst[symbol] != NULL); }
  int is_char_present(char c);
  int count(int symbol);

  /* ---DEBUG-------------------*/
  void print();
  //-----------------------------

 private:
  SymbolSubst *subst[27];
};

/* DEBUG-----------------------*/
void TableSubst::print() {
  for (int i=0; i<27; i++) if (subst[i]) subst[i]->print();
}
//-------------------------

TableSubst::TableSubst() {
  for (int i=0; i<27; i++) subst[i] = NULL;
}

TableSubst::~TableSubst() {
  for (int i=0; i<27; i++) if (subst[i]) delete subst[i];
}

void TableSubst::clear() {
  for (int i=0; i<27; i++) {
    if (subst[i]) delete subst[i];
    subst[i] = NULL;
  }
}

void TableSubst::add_subst(int symbol, char c, double freq) {
  if (subst[symbol] == NULL) subst[symbol] = new SymbolSubst(symbol);
  subst[symbol]->add_subst(c, freq);
}

int TableSubst::count(int symbol) {
  if (subst[symbol]) return subst[symbol]->count();
  else return 0;
}

int TableSubst::is_char_present(char c) {
  for (int i=0; i<27; i++)
    if (subst[i] && subst[i]->is_present(c)) return 1;
  return 0;
}
