/* CLASS VIRTUALSCREEN_SDL */

#include <stdlib.h>
#include <stdio.h>
#include "class_virscrsdl.h"

#include "../include/mmx.h"

#define min(a,b)	(((a)<(b))?(a):(b))
#define max(a,b)	(((a)>(b))?(a):(b))

union rgbint {
  Uint32 c;
  struct {
    Uint8 b,g,r,a;
  };
};


VirtualScreenSDL::VirtualScreenSDL(int sizex, int sizey) {
  if (!(SDL_WasInit(SDL_INIT_VIDEO)&SDL_INIT_VIDEO)) SDL_Init(SDL_INIT_VIDEO);
  screen = SDL_CreateRGBSurface(SDL_HWSURFACE, sizex, sizey, 32, 0xFF0000, 0xFF00, 0xFF, 0xFF000000);
  if (screen == NULL) CatchError(VSSDL_ERR_NULLBUFF, NULL);
  IsLocked = 0;
  FreeSurface = 1;
  SkipLineFirst = 0;
  UseMMX = isMMX2();
#ifndef VSSDL_NO_TTF
  fontsLoaded = 0;
  fonts = NULL;
#endif
  DrawMode = VS_DM_NORMAL;
  TextMode = VS_TM_SOLID;
  TextAlign = VS_TA_LEFT;
  Lock();
}

VirtualScreenSDL::VirtualScreenSDL(SDL_Surface *screen) {
  if (screen == NULL) CatchError(VSSDL_ERR_NULLBUFF, NULL);
  this->screen = screen;
  IsLocked = 0;
  FreeSurface = 0;
  SkipLineFirst = 0;
  UseMMX = isMMX2();
#ifndef VSSDL_NO_TTF
  fontsLoaded = 0;
  fonts = NULL;
#endif
  DrawMode = VS_DM_NORMAL;
  TextMode = VS_TM_SOLID;
  TextAlign = VS_TA_LEFT;
  Lock();
}

#ifndef VSSDL_NO_IMAGE
VirtualScreenSDL::VirtualScreenSDL(const char *filename) {
  if (!(SDL_WasInit(SDL_INIT_VIDEO)&SDL_INIT_VIDEO)) SDL_Init(SDL_INIT_VIDEO);
  screen = IMG_Load(filename);
  if (screen == NULL) CatchError(VSSDL_ERR_NULLBUFF, NULL);
  IsLocked = 0;
  FreeSurface = 1;
  SkipLineFirst = 0;
  UseMMX = isMMX2();
  fontsLoaded = 0;
  fonts = NULL;
  DrawMode = VS_DM_NORMAL;
  TextMode = VS_TM_SOLID;
  TextAlign = VS_TA_LEFT;
  Lock();
}
#endif

VirtualScreenSDL::~VirtualScreenSDL() {
  Unlock();
#ifndef VSSDL_NO_TTF
  for (int i=0; i<fontsLoaded; i++) TTF_CloseFont(*(fonts+i));
  if (fonts != NULL) free(fonts);
#endif
  if (FreeSurface) SDL_FreeSurface(screen);
}

int VirtualScreenSDL::ChangeSurface(SDL_Surface *newscreen) {
  if (FreeSurface) SDL_FreeSurface(screen);
  screen = newscreen;
  return(0);
}

int VirtualScreenSDL::SetDisplayFormat() {
  SDL_Surface *newsurf = SDL_DisplayFormat(screen);
  if (newsurf == NULL) return(-1);
  return ChangeSurface(newsurf);
}

int VirtualScreenSDL::CatchError(int n, void *data) {
  if (n == 0) return(0);
  fprintf(stderr, "VirtualScreenSDL error #%i.\n", n);
  fprintf(stderr, "Last SDL error: %s\n", SDL_GetError());
  //exit(n);
  abort();
}

int VirtualScreenSDL::Lock() {
  int ret = 0;
  if (IsLocked) return(0);
  if (SDL_MUSTLOCK(screen)) ret = SDL_LockSurface(screen);
  if (ret == 0) {
    IsLocked = 1;
  } else {
    IsLocked = 0;
    CatchError(VSSDL_ERR_NOLOCK, NULL);
  }
  return(ret);
}

int VirtualScreenSDL::Unlock() {
  if (!IsLocked) return(0);
  if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);
  IsLocked = 0;
  return(0);
}

int VirtualScreenSDL::EnableMMX() {
  if (UseMMX) return(UseMMX);
  UseMMX = isMMX2();
  return(UseMMX);
}

int VirtualScreenSDL::DisableMMX() {
  UseMMX = 0;
  return(UseMMX);
}

inline int VirtualScreenSDL::GetWidth() {
  return(screen->w);
}

inline int VirtualScreenSDL::GetHeight() {
  return(screen->h);
}

int VirtualScreenSDL::Clear(int color) {
  Unlock();
  SDL_FillRect(screen, NULL, color);
}

int VirtualScreenSDL::Show() {
  Unlock();
  SDL_UpdateRect(screen,0,0,0,0);
}

int VirtualScreenSDL::Show(int x1, int y1, int x2, int y2) {
  Unlock();
  SDL_UpdateRect(screen, x1,y1, x2,y2);
}

int VirtualScreenSDL::CopyTo(int destx,int desty, VirtualScreenSDL *dest) {
  Unlock();
  dest->Unlock();
  SDL_Rect dstrect = {destx, desty, 0,0};
  return SDL_BlitSurface(screen, NULL, dest->screen, &dstrect);
}

int VirtualScreenSDL::CopyRectTo(int x1,int y1, int x2,int y2, int destx,int desty, VirtualScreenSDL *dest) {
  Unlock();
  dest->Unlock();
  SDL_Rect srcrect = {x1,y1, abs(x2-x1)+1,abs(y2-y1)+1};
  SDL_Rect dstrect = {destx, desty, 0,0};
  return SDL_BlitSurface(screen, &srcrect, dest->screen, &dstrect);
}

int VirtualScreenSDL::PutPixel(int x, int y, int color) {
  if ((unsigned)x >= screen->w || (unsigned)y >= screen->h) return(-1);
  Lock();

  int bpp = screen->format->BytesPerPixel;
  Uint8 *ptr = (Uint8 *)screen->pixels + y*screen->pitch + x*bpp;
  return HardPutPixel(ptr, bpp, color);
}

int VirtualScreenSDL::GetPixel(int x, int y) {
  if ((unsigned)x >= screen->w || (unsigned)y >= screen->h) return(0);
  Lock();
  
  int bpp = screen->format->BytesPerPixel;
  Uint8 *p = (Uint8 *)screen->pixels + y*screen->pitch + x*bpp;
  switch (bpp) {
    case 1: return *p;
    case 2: return *(Uint16 *)p;
    case 3: 
      if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
        return p[0] << 16 | p[1] << 8 | p[2];
      else
        return p[0] | p[1] << 8 | p[2] << 16;
    case 4: return *(Uint32 *)p;
    default: return 0;
  }
}

int VirtualScreenSDL::Line(int x1, int y1, int x2, int y2, int color) {
  int dx = x2-x1, dy = y2-y1;

  // checks if horizontal or vertical line
  if (dy == 0) return HLine(x1,y1, x2, color);
  if (dx == 0) return VLine(x1,y1, y2, color);

  // checks coord if out of screen
  if (y1 < 0) {
    if (y2 < 0) return(-1);
    x1 = (int)(x1-(long long)y1*dx/dy);
    y1 = 0;
    dy = y2;
    dx = x2-x1;
  } else if (y1 > GetHeight()) {
    int h = GetHeight();
    if (y2 > h) return(-1);
    x1 = (int)(x1+(long long)(h-y1)*dx/dy);
    y1 = h;
    dy = y2-y1;
    dx = x2-x1;
  }
  if (dx == 0) return VLine(x1,y1, y2, color);
  if (x1 < 0) {
    if (x2 < 0) return(-1);
    y1 = (int)(y1-(long)x1*dy/dx);
    x1 = 0;
    dy = y2-y1;
    dx = x2;
  } else if (x1 > GetWidth()) {
    int w = GetWidth();
    if (x2 > w) return(-1);
    y1 = (int)(y1+(long long)(w-x1)*dy/dx);
    x1 = w;
    dy = y2-y1;
    dx = x2-x1;
  }
  if (dy == 0) return HLine(x1,y1, x2, color);
  if (y2 < 0) {
    x2 = (int)(x2-(long long)y2*dx/dy);
    y2 = 0;
    dy = -y1;
    dx = x2-x1;
  } else if (y2 > GetHeight()) {
    int h = GetHeight();
    x2 = (int)(x2+(long long)(h-y2)*dx/dy);
    y2 = h;
    dy = y2-y1;
    dx = x2-x1;
  }
  if (dx == 0) return VLine(x1,y1, y2, color);
  if (x2 < 0) {
    y2 = (int)(y2-(long)x2*dy/dx);
    x2 = 0;
    dy = y2-y1;
    dx = -x1;
  } else if (x2 > GetWidth()) {
    int w = GetWidth();
    y2 = (int)(y2+(long long)(w-x2)*dy/dx);
    x2 = w;
    dy = y2-y1;
    dx = x2-x1;
  }

  int ix = abs(dx), iy = abs(dy),
      inct = (ix>iy?ix:iy),
      plotX = x1, plotY = y1,
      x = 0, y = 0,
      i;
  char plot, drawn=0;

  if (!SkipLineFirst) PutPixel(plotX,plotY, color);
  for (i=0;i<=inct;i++) {
    x += ix;
    y += iy;
    plot = 0;
    if (x > inct) {
      plot = 1;
      x -= inct;
      if (dx > 0) plotX++;
      else plotX--;
    }
    if (y > inct) {
      plot = 1;
      y -= inct;
      if (dy > 0) plotY++;
      else plotY--;
    }
    if (plot) {
      int r = PutPixel(plotX,plotY, color);
      if (drawn && r) return(0);
      if (!r) drawn = 1;
    }
  }
  return( (drawn==0)?(-1):0 );
}

int VirtualScreenSDL::Rect(int x1, int y1, int x2, int y2, int color) {
  if (x1==x2 || y1==y2) {
    Line(x1,y1, x2,y2, color);
  } else {
    SkipLineFirst = 1;
    HLine(x1,y1, x2, color);
    VLine(x2,y1, y2, color);
    HLine(x2,y2, x1, color);
    VLine(x1,y2, y1, color);
    SkipLineFirst = 0;  
  }
}


/* NOT FINISHED YET */
int VirtualScreenSDL::Triangle(int x1, int y1, int x2, int y2, int x3, int y3, int color) {
  SkipLineFirst = 1;
  Line(x1,y1, x2,y2, color);
  Line(x2,y2, x3,y3, color);
  Line(x3,y3, x1,y1, color);
  SkipLineFirst = 0;

return 0;
  // y1 must not be aligned with others y
  if (y1 == y2 && y1 == y3) { 
    Line(min(x1,min(x2,x3)),y1, max(x1,max(x2,x3)),y2, color);
    return 0;
  } else if (y1 == y2) {
    int a,b;
    a = x1; b = y1;
    x1 = x3; y1 = y3;
    x3 = a; y3 = b;
  } else if (y1 == y3) {
    int a,b;
    a = x1; b = y1;
    x1 = x2; y1 = y2;
    x2 = a; y2 = b;
  }
  if ((y1 > y2 && y1 < y3)||(y1 > y3 && y1 < y2)) {
    int a,b;
    a = x1; b = y1;
    x1 = x2; y1 = y2;
    x2 = a; y2 = b;
  }

  int dx[3] = {x2-x1, x3-x1, x3-x2},
      dy[3] = {y2-y1, y3-y1, y3-y2},
      ix[3] = {abs(dx[0]), abs(dx[1]), abs(dx[2])},
      iy[3] = {abs(dy[0]), abs(dy[1]), abs(dy[2])},
      inct[3] = {max(ix[0],iy[0]), max(ix[1],iy[1]), max(ix[2],iy[2])},
      x[3] = {0,0,0},
      y[3] = {0,0,0},
      plotX[3] = {x1,x1,x2},
      plotY[3] = {y1,y1,y2},
      addx[3] = {0,0,0},
      addy[3] = {0,0,0},
      ylimit0 = iy[0],
      ylimit1 = iy[1],
      putx1,puty1, putx2,puty2,
      i,j;
  char goy[2] = {1,1};
  register int idx0 = 0, idx1 = 1;

  if (iy[0]>iy[1]) {
    dx[2] = -dx[2];
    dy[2] = -dy[2];
    plotX[2] = x3;
    plotY[2] = y3;
    ylimit0 = iy[1];
    ylimit1 = iy[0];
  }

  for (i=0; i<3; i++) {
    if (dx[i] > 0) addx[i]++;
    else if (dx[i] < 0) addx[i]--;
    if (dy[i] > 0) addy[i]++;
    else if (dy[i] < 0) addy[i]--;
  }

  for (i=0; i<ylimit1; ) {
    if (goy[0]) {
      x[idx0] += ix[idx0];
      y[idx0] += iy[idx0];
      if (x[idx0] > inct[idx0]) {
        x[idx0] -= inct[idx0];
        plotX[idx0] += addx[idx0];
      }
      if (y[idx0] > inct[idx0]) {
        putx1 = plotX[idx0];
        puty1 = plotY[idx0];
        y[idx0] -= inct[idx0];
        plotY[idx0] += addy[idx0];
        goy[0] = 0;
      }
    }
    
    if (goy[1]) {
      x[idx1] += ix[idx1];
      y[idx1] += iy[idx1];
      if (x[idx1] > inct[idx1]) {
        x[idx1] -= inct[idx1];
        plotX[idx1] += addx[idx1];
      }
      if (y[idx1] > inct[idx1]) {
        putx2 = plotX[idx1];
        puty2 = plotY[idx1];
        y[idx1] -= inct[idx1];
        plotY[idx1] += addy[idx1];
        goy[1] = 0;
      }
    }

    if (!goy[0] && !goy[1]) {
      goy[0] = goy[1] = 1;
      //Line(putx1,puty1, putx2,puty2, color);
      PutPixel(putx1,puty1, color);
      if (putx2 != putx1 || puty2 != puty1) PutPixel(putx2,puty2, color);
      if (i == ylimit0-1) {
        if (iy[0] < iy[1]) idx0 = 2; else idx1 = 2;
      }
      i++;
    }
  }
  //if (iy[0] == iy[1])
    //Line(plotX[idx0],plotY[idx0], plotX[idx1],plotY[idx1], color);
  return 0;
  
}

#ifndef VSSDL_NO_TTF
int VirtualScreenSDL::LoadFont(const char *fontname, int size) {
  if (!TTF_WasInit()) TTF_Init();
  TTF_Font *f = TTF_OpenFont(fontname, size);
  if (f == NULL) return(-1);
  fonts = (TTF_Font**)realloc(fonts, sizeof(TTF_Font*)*(fontsLoaded+1));
  *(fonts+fontsLoaded++) = f;
  return(0);
}

int VirtualScreenSDL::ReloadFont(int fontindex, const char *fontname, int size) {
  if (fontindex >= fontsLoaded) return(-1);
  TTF_Font *f = TTF_OpenFont(fontname, size);
  if (f == NULL) return(-1);
  TTF_CloseFont(*(fonts+fontindex));
  *(fonts+fontindex) = f;
  return(0);
}


int VirtualScreenSDL::SetFontStyle(int fontindex, int style) {
  if (fontindex >= fontsLoaded) return(-1);
  TTF_SetFontStyle(fonts[fontindex], style);
  return(0);
}

int VirtualScreenSDL::GetPrintSize(int fontindex, const char *text, int *sizex, int *sizey) {
  if (fontindex >= fontsLoaded) return(-1);
  int ret = TTF_SizeText(*(fonts+fontindex), text, sizex, sizey);
  return(ret);
}

SDL_Surface *VirtualScreenSDL::Print(int x, int y, const char *text, int color, int colorback) {
  Print(0, x,y, text, color, colorback);
}

SDL_Surface *VirtualScreenSDL::Print(int fontindex, int x, int y, const char *text, int color, int colorback) {
  if (fontindex >= fontsLoaded) return(NULL);

  SDL_Color colf, colb;
  SDL_GetRGB(color, screen->format, &(colf.r), &(colf.g), &(colf.b));
  SDL_GetRGB(colorback, screen->format, &(colb.r), &(colb.g), &(colb.b));
  SDL_Surface *fs;
  switch (TextMode) {
    case VS_TM_NONE: break;
    default:
    case VS_TM_SOLID: fs = TTF_RenderText_Solid(fonts[fontindex], text, colf); break;
    case VS_TM_SHADED: fs = TTF_RenderText_Shaded(fonts[fontindex], text, colf, colb); break; 
    case VS_TM_BLENDED: fs = TTF_RenderText_Blended(fonts[fontindex], text, colf); break;
  }
  
  if (fs && !(TextMode & VS_TM_NORENDER)) {
    Unlock();
    SDL_Rect rc = {x,y,0,0};
    SDL_SetAlpha(fs, SDL_SRCALPHA, (color>>24)&0xff);
    SDL_BlitSurface(fs, NULL, screen, &rc);
    SDL_FreeSurface(fs);
  } else return (fs);
}

int VirtualScreenSDL::PrintRect(int fontindex, int x1, int y1, int x2, int y2, const char *text, int color, int colorback) {
  if (fontindex >= fontsLoaded) return(-1);
  int height = TTF_FontLineSkip(fonts[fontindex]);
  int fx,fy, len;
  int rectx = abs(x2-x1)+1,
      recty = abs(y2-y1)+1;
  int oldtm = TextMode;
  TextMode &= ~VS_TM_NORENDER;
  
  while (text[0] != '\0') {
    if (y1+height > y2) break;
    char *text2 = strdup(text);
    len = strlen(text2);
    TTF_SizeText(fonts[fontindex], text2, &fx, &fy);
    while (fx > rectx) {
      *(text2+--len) = '\0';
      TTF_SizeText(fonts[fontindex], text2, &fx, &fy);
    }
    switch (TextAlign) {
      case VS_TA_LEFT: Print(fontindex, x1,y1, text2, color, colorback); break;
      case VS_TA_CENTER: Print(fontindex, x1+(rectx-fx)/2,y1, text2, color, colorback); break;
      case VS_TA_RIGHT: Print(fontindex, x1+(rectx-fx),y1, text2, color, colorback); break;
    }
    text += len;
    y1 += height;
    free(text2);
  }
  TextMode = oldtm;
  return(0);
}
#endif

#ifndef VSSDL_NO_IMAGE
int VirtualScreenSDL::LoadImage(const char *filename) {
  SDL_Surface *img = IMG_Load(filename);
  if (img == NULL) return(-1);
  Unlock();
  if (FreeSurface) SDL_FreeSurface(screen);
  screen = img;
}

int VirtualScreenSDL::LoadImageXPM(char *xpm[]) {
  SDL_Surface *img = IMG_ReadXPMFromArray(xpm);
  if (img == NULL) return(-1);
  Unlock();
  if (FreeSurface) SDL_FreeSurface(screen);
  screen = img;
}

int VirtualScreenSDL::SaveImage(const char *file) {
  return SDL_SaveBMP(screen, file);
}

int VirtualScreenSDL::SaveImage(int x1, int y1, int x2, int y2, char *file) {
  SDL_Surface *image = SDL_CreateRGBSurface(SDL_SWSURFACE, abs(x2-x1), abs(y2-y1), 32, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);
  SDL_Rect srcrect = {x1,y1, x2,y2};
  
  SDL_BlitSurface(screen, &srcrect, image, NULL);
  int ret = SDL_SaveBMP(image, file);
  SDL_FreeSurface(image);
  return ret;
}
#endif

int VirtualScreenSDL::Blur(int x1, int y1, int x2, int y2, int type) {
  int i,j;
  static int shift = 0;
  if (screen->format->BytesPerPixel != 4) return(-2);
  
  i = min(x1,x2);
  j = max(x1,x2);
  x1 = i; x2 = j;
  i = min(y1,y2);
  j = max(y1,y2);
  y1 = i; y2 = j;
  
  if (x2 < 0 || x1 >= screen->w) return(-1);
  if (y2 < 0 || y1 >= screen->h) return(-1);
  
  if (x1 < 0) x1 = 0;
  if (x2 >= screen->w) x2 = screen->w-1;
  if (y1 < 0) y1 = 0;
  if (y2 >= screen->h) y2 = screen->h-1;
  switch (type) {
    case VS_BLUR_SUPERFAST: y2 >>= 1;
    case VS_BLUR_FAST: x2 >>= 1; break;
    default: ;
  }
  Lock();

  Uint8 *ptr, *ptr2;
  rgbint p1,p2,p3,p4;
  rgbint backCol = {0xff000000};
  
  /* without mmx */
  
  if (type == VS_BLUR_NORMAL) {
    for (j=y1-1; j<y2; j++) {
      ptr = (Uint8 *)screen->pixels + j*screen->pitch + (x1<<2);
      ptr2 = ptr+screen->pitch;
      for (i=x1; i<=x2; i++) {
         if (j >= y1) p1 = *(rgbint*)(ptr);
         else p1 = backCol;
         if (i < x2) p2 = *(rgbint*)(ptr2+4);
         else p2 = backCol;
         if (i > x1) p3 = *(rgbint*)(ptr2-4);
         else p3 = backCol;
         if (j<y2-1) p4 = *(rgbint*)(ptr2+screen->pitch);
         else p4 = backCol;
       
       // Correct, but slow
       /*int outcol;
       __m64 c1, c2, c3, c4;
       __m64 zero, two;
       zero.qw = 0;
       two.qw = 2;
       c1.qw = _mm_unpcklo_pi8(p1.c, zero);
       c2.qw = _mm_unpcklo_pi8(p2.c, zero);
       c3.qw = _mm_unpcklo_pi8(p3.c, zero);
       c4.qw = _mm_unpcklo_pi8(p4.c, zero);
       c1.qw = _mm_add_pi16(c1,c2);
       c2.qw = _mm_add_pi16(c3,c4);
       c1.qw = _mm_add_pi16(c1,c2);
       c1.qw = _mm_srl_pi16(c1, two);
       outcol = _mm_packs_pu16(c1,c1);*/
       
       // Wrong! (why?)
       /*c1.dw[0] = p1.c;
       c1.dw[1] = p3.c;
       c2.dw[0] = p2.c;
       c2.dw[1] = p4.c;
       c1.qw = _mm_avg_pu8(c1,c2);
       outcol = _mm_avg_pu8(c1, c1.dw[1]) & 0x00ffffff;*/
       
         *(Uint32*)ptr2 &= 0xff000000;
         //*(Uint32*)ptr2 |= outcol;
         *(Uint32*)(ptr2) |= 	(((p1.r+p2.r+p3.r+p4.r)&0xfffc)<<14) | 
       				(((p1.g+p2.g+p3.g+p4.g)&0xfffc)<<6) | 
       				((p1.b+p2.b+p3.b+p4.b)>>2);
         ptr += 4;
         ptr2 += 4;
      }
    }
  } else {
    for (j=y1; j<=y2; j++) {
      if (type == VS_BLUR_FAST) ptr = (Uint8 *)screen->pixels + j*screen->pitch + ((x1+(j+shift)&1)<<2);
      else ptr = (Uint8 *)screen->pixels + ((j<<1)+(shift>>1))*screen->pitch + ((x1+(j+shift)&1)<<2); 
      for (i=x1; i<=x2; i++) {
        if (j > y1) p1 = *(rgbint*)(ptr-screen->pitch);
        else p1 = backCol;
        if (i < x2) p2 = *(rgbint*)(ptr+4);
        else p2 = backCol;
        if (i > x1) p3 = *(rgbint*)(ptr-4);
        else p3 = backCol;
        if (j<y2) p4 = *(rgbint*)(ptr+screen->pitch);
        else p4 = backCol;
        *(Uint32*)ptr &= 0xff000000;
        *(Uint32*)ptr |= 	(((p1.r+p2.r+p3.r+p4.r)&0xfffc)<<14) | 
       				(((p1.g+p2.g+p3.g+p4.g)&0xfffc)<<6) | 
       				((p1.b+p2.b+p3.b+p4.b)>>2);
        ptr += 8;
      }
    }
    shift = (shift+1)&0x3;
  }
  return 0;
}

int VirtualScreenSDL::Fadeout(int x1, int y1, int x2, int y2, int color, int speed, int type) {
  int i,j, add;
  static int shift = 0;
  if (screen->format->BytesPerPixel != 4) return(-2);

  i = min(x1,x2);
  j = max(x1,x2);
  x1 = i; x2 = j;
  i = min(y1,y2);
  j = max(y1,y2);
  y1 = i; y2 = j;
  if (x2 < 0 || x1 >= screen->w) return(-1);
  if (y2 < 0 || y1 >= screen->h) return(-1);  
  if (x1 < 0) x1 = 0;
  if (x2 >= screen->w) x2 = screen->w-1;
  if (y1 < 0) y1 = 0;
  if (y2 >= screen->h) y2 = screen->h-1;
  switch (type) {
    case VS_FADE_SUPERFAST: y2 >>= 1;
    case VS_FADE_FAST: x2 >>= 1; add=8; break;
    default: add = 4;
  }
  Lock();
 
  Uint8 *ptr;
  rgbint p1,p2,p3;
  p2.c = color;
  speed &= 0xff;
  
  for (j=y1; j<=y2; j++) {
    switch (type) {
      default:
      case VS_FADE_NORMAL:
        ptr = (Uint8 *)screen->pixels + j*screen->pitch + (x1<<2); break;
      case VS_FADE_FAST:
        ptr = (Uint8 *)screen->pixels + j*screen->pitch + ((x1+(j+shift)&1)<<2); break;
      case VS_FADE_SUPERFAST: 
        ptr = (Uint8 *)screen->pixels + ((j<<1)+(shift>>1))*screen->pitch + ((x1+(j+shift)&1)<<2); break;
    }
    for (i=x1; i<=x2; i++) {
      p1 = *(rgbint*)ptr;
      if (p1.r < p2.r) p3.r = min(p1.r+speed, p2.r);
      else p3.r = max(p1.r-speed, p2.r);
      if (p1.g < p2.g) p3.g = min(p1.g+speed, p2.g);
      else p3.g = max(p1.g-speed, p2.g);
      if (p1.b < p2.b) p3.b = min(p1.b+speed, p2.b);
      else p3.b = max(p1.b-speed, p2.b);
      if (p1.a < p2.a) p3.a = min(p1.a+speed, p2.a);
      else p3.a = max(p1.a-speed, p2.a);
      *(Uint32*)ptr = p3.c;
      ptr += add;
    }
  }

  if (type != VS_FADE_NORMAL) shift = (shift+1)&0x3;
  return 0;
}

int VirtualScreenSDL::StretchRectTo(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4, VirtualScreenSDL *dest) {
  int w1,h1, w2,h2, addx,addy;
  int i,j, nx,ny;

  /* TODO: checks arguments */
  
  w1 = abs(x2-x1)+1;
  h1 = abs(y2-y1)+1;
  w2 = abs(x4-x3)+1;
  h2 = abs(y4-y3)+1;
  
  VirtualScreenSDL *destsurf = dest;
  addx = x3; addy = y3;
  if (this == dest) {
    destsurf = new VirtualScreenSDL(w2, h2);
    SDL_SetAlpha(destsurf->screen, screen->flags & (SDL_RLEACCEL | SDL_SRCALPHA), screen->format->alpha);
    SDL_SetColorKey(destsurf->screen, screen->flags & (SDL_RLEACCEL | SDL_SRCCOLORKEY), screen->format->colorkey);
    addx=addy=0;
  }
  
  for (j=0; j<h2; j++) {
    ny = y1+j*h1/h2;
    for (i=0; i<w2; i++) {
      nx = x1+i*w1/w2;;
      int col = GetPixel(nx, ny);
      destsurf->PutPixel(addx+i,addy+j, col);
    }
  }
  if (this == dest) {
    destsurf->CopyTo(x3,y3, dest);
    delete destsurf;
  }
}

int VirtualScreenSDL::Pixelize(int x1, int y1, int x2, int y2, int size) {
  int endy,endp, i,j, size2=size>>1;
  if (size < 2) return(0);
  
  for (j=y1; j<=y2; j+=size) {
    endy = min(j+size-1, y2);
    endp = min(j+size2, y2);
    for (i=x1; i<=x2; i+=size) {
      FRect(i,j, min(i+size-1, x2),endy, GetPixel(min(i+size2, x2), endp));
    }
  }  
}

int VirtualScreenSDL::Contrast(int x1,int y1, int x2,int y2, char amount) {
  int i,j;
  
  if (screen->format->BytesPerPixel != 4) return(-2);
  i = min(x1,x2);
  j = max(x1,x2);
  x1 = i; x2 = j;
  i = min(y1,y2);
  j = max(y1,y2);
  y1 = i; y2 = j;
  if (x2 < 0 || x1 >= screen->w) return(-1);
  if (y2 < 0 || y1 >= screen->h) return(-1);  
  if (x1 < 0) x1 = 0;
  if (x2 >= screen->w) x2 = screen->w-1;
  if (y1 < 0) y1 = 0;
  if (y2 >= screen->h) y2 = screen->h-1;

  Uint8 *ptr;
  rgbint col, tc;

  for (j=y1; j<=y2; j++) {
    ptr = (Uint8 *)screen->pixels + j*screen->pitch + (x1<<2);
    for (i=x1; i<=x2; i++) {
      col.c = *(Uint32*)ptr;
      tc.c = col.c & 0x80808080;
      if (amount > 0) {
        if (tc.b) col.b = min(col.b+amount, 0xff);
        else col.b = max(col.b-amount, 0x00);
        if (tc.g) col.g = min(col.g+amount, 0xff);
        else col.g = max(col.g-amount, 0x00);
        if (tc.r) col.r = min(col.r+amount, 0xff);
        else col.r = max(col.r-amount, 0x00);
      } else {
        if (tc.b) col.b = max(col.b+amount, 0x7f);
        else col.b = min(col.b-amount, 0x7f);
        if (tc.g) col.g = max(col.g+amount, 0x7f);
        else col.g = min(col.g-amount, 0x7f);
        if (tc.r) col.r = max(col.r+amount, 0x7f);
        else col.r = min(col.r-amount, 0x7f);
      }
      *(Uint32*)ptr = col.c;
      ptr += 4;
    }
  }
  return(0);
}

int VirtualScreenSDL::Brightness(int x1,int y1, int x2,int y2, char amount) {
  int i,j;
  
  if (screen->format->BytesPerPixel != 4) return(-2);
  i = min(x1,x2);
  j = max(x1,x2);
  x1 = i; x2 = j;
  i = min(y1,y2);
  j = max(y1,y2);
  y1 = i; y2 = j;
  if (x2 < 0 || x1 >= screen->w) return(-1);
  if (y2 < 0 || y1 >= screen->h) return(-1);  
  if (x1 < 0) x1 = 0;
  if (x2 >= screen->w) x2 = screen->w-1;
  if (y1 < 0) y1 = 0;
  if (y2 >= screen->h) y2 = screen->h-1;

  Uint8 *ptr;
  rgbint col;

  __m64 add, mcol;
  add.b[0] = abs(amount);
  add.b[1] = add.b[0];
  add.w[1] = add.w[0];
  add.dw[1] = add.dw[0];
  int nend = x1+((x2-x1+1)>>1);
  MMX_RETTYPE (*mmxfunc)(__m64, __m64);
  mmxfunc = (amount>0)?_mm_adds_pu8:_mm_sub_pu8;
  
  for (j=y1; j<=y2; j++) {
    ptr = (Uint8 *)screen->pixels + j*screen->pitch + (x1<<2);
    if (UseMMX) {
      for (i=x1; i<=nend; i++) {
        mcol.qw = *(long long *)ptr;
        mcol.qw = mmxfunc(mcol, add);
        *(long long *)ptr = mcol.qw;
        ptr += 8;
      }
      if ((x2-x1+1) & 1) {
        mcol.dw[0] = *(Uint32*)ptr;
        mcol.dw[0] = mmxfunc(mcol, add);
        *(Uint32*)ptr = mcol.dw[0];
      }
    } else {
      for (i=x1; i<=x2; i++) {
        col.c = *(Uint32*)ptr;
        if (amount > 0) {
          col.b = min(col.b+amount, 0xff);
          col.g = min(col.g+amount, 0xff);
          col.r = min(col.r+amount, 0xff);
        } else {
          col.b = max(col.b+amount, 0x00);
          col.g = max(col.g+amount, 0x00);
          col.r = max(col.r+amount, 0x00);
        }
        *(Uint32*)ptr = col.c;
        ptr += 4;
      }
    }
  }
  return(0);
}

int VirtualScreenSDL::Posterize(int x1,int y1, int x2,int y2, unsigned char levels) {
  int i,j;
  
  if (levels < 2) return(-1);
  if (screen->format->BytesPerPixel != 4) return(-2);
  i = min(x1,x2);
  j = max(x1,x2);
  x1 = i; x2 = j;
  i = min(y1,y2);
  j = max(y1,y2);
  y1 = i; y2 = j;
  if (x2 < 0 || x1 >= screen->w) return(-1);
  if (y2 < 0 || y1 >= screen->h) return(-1);  
  if (x1 < 0) x1 = 0;
  if (x2 >= screen->w) x2 = screen->w-1;
  if (y1 < 0) y1 = 0;
  if (y2 >= screen->h) y2 = screen->h-1;

  Uint8 *ptr;
  rgbint col;
  int d = 255/levels;

  for (j=y1; j<=y2; j++) {
    ptr = (Uint8 *)screen->pixels + j*screen->pitch + (x1<<2);
    for (i=x1; i<=x2; i++) {
      col.c = *(Uint32*)ptr;
      switch (levels) {
        case 2: col.c &= 0xff808080; break;
        case 4: col.c &= 0xffc0c0c0; break;
        case 8: col.c &= 0xffe0e0e0; break;
        case 16: col.c &= 0xfff0f0f0; break;
        default:
          col.r = (col.r/d)*d;
          col.g = (col.g/d)*d;
          col.b = (col.b/d)*d;
      }
      *(Uint32*)ptr = col.c;
      ptr += 4;
    }
  }

  return(0);
}

int VirtualScreenSDL::HLine(int x1, int y1, int x2, int color) {
  if ((unsigned)y1 >= screen->h) return(-1);
  if (x1 > x2) { 
    register int t = x1; x1 = x2; x2 = t; 
    if (SkipLineFirst) x2--;
  } else 
    if (SkipLineFirst) x1++;
  if (x2 < 0 || x1 >= screen->w) return(-1);
  if (x1 < 0) x1 = 0;
  if (x2 >= screen->w) x2 = screen->w-1;
  
  Lock();
  int bpp = screen->format->BytesPerPixel;
  Uint8 *ptr = (Uint8 *)screen->pixels + y1*screen->pitch + x1*bpp;
  for (int i=x1; i<=x2; i++, ptr+=bpp) HardPutPixel(ptr, bpp, color);
  return(0);
}

int VirtualScreenSDL::VLine(int x1, int y1, int y2, int color) {
  if ((unsigned)x1 >= screen->w) return(-1);
  if (y1 > y2) { 
    register int t = y1; y1 = y2; y2 = t; 
    if (SkipLineFirst) y2--;
  } else 
    if (SkipLineFirst) y1++;
  if (y2 < 0 || y1 >= screen->h) return(-1);
  if (y1 < 0) y1 = 0;
  if (y2 >= screen->h) y2 = screen->h-1;
  
  Lock();
  int bpp = screen->format->BytesPerPixel;
  Uint8 *ptr = (Uint8 *)screen->pixels + y1*screen->pitch + x1*bpp;
  for (int i=y1; i<=y2; i++, ptr+=screen->pitch) HardPutPixel(ptr, bpp, color);
  return(0);
}

int VirtualScreenSDL::HardPutPixel(Uint8 *ptr, int bpp, int color) {
  switch (DrawMode) {
    case VS_DM_NONE: break;
    case VS_DM_NORMAL: 
      switch (bpp) {
        case 1: *ptr = (Uint8)color;
        case 4: *(Uint32*)ptr = color; break;
      }
      break;
    case VS_DM_ADD:
      switch (bpp) {
        case 4: {
          if (UseMMX) {
            *(Uint32*)ptr = (Uint32)_mm_adds_pu8((__m64)*(Uint32*)ptr, (__m64)color);
          } else {
            Uint32 c = *(Uint32*)ptr; 
            Uint32 r = min((c & 0xFF)+(color & 0xFF), 255);
            Uint32 g = min(((c>>8) & 0xFF)+((color>>8) & 0xFF), 255);
            Uint32 b = min(((c>>16) & 0xFF)+((color>>16) & 0xFF), 255);
            *(Uint32*)ptr = 0xFF000000 | (b<<16) | (g<<8) | r;
          }
          break;
        }
      }
      break;
    case VS_DM_SUB1:
      switch (bpp) {
        case 4: {
          if (UseMMX) {
            *(Uint32*)ptr = (int)_mm_sub_pu8((__m64)*(Uint32*)ptr, (__m64)color);
          } else {
            Uint32 c = *(Uint32*)ptr; 
            Uint32 r = max((c & 0xFF)-(color & 0xFF), 0);
            Uint32 g = max(((c>>8) & 0xFF)-((color>>8) & 0xFF), 0);
            Uint32 b = max(((c>>16) & 0xFF)-((color>>16) & 0xFF), 0);
            *(Uint32*)ptr = 0xFF000000 | (b<<16) | (g<<8) | r;
          }
          break;
        }
      }
      break;    
    case VS_DM_SUB2:
      switch (bpp) {
        case 4: {
          if (UseMMX) {
            *(Uint32*)ptr = (int)_mm_sub_pu8((__m64)color, (__m64)*(Uint32*)ptr);
          } else {
            Uint32 c = *(Uint32*)ptr; 
            Uint32 r = max(-(c & 0xFF)+(color & 0xFF), 0);
            Uint32 g = max(-((c>>8) & 0xFF)+((color>>8) & 0xFF), 0);
            Uint32 b = max(-((c>>16) & 0xFF)+((color>>16) & 0xFF), 0);
            *(Uint32*)ptr = 0xFF000000 | (b<<16) | (g<<8) | r;
          }
          break;
        }
      }
      break;    
    case VS_DM_MUL:
      switch (bpp) {
        case 4: {
          if (0) { // Is there an artful manner to do this using mmx?
          } else {
            Uint32 c = *(Uint32*)ptr; 
            Uint32 r = min((c & 0xFF)*(color & 0xFF), 255);
            Uint32 g = min(((c>>8) & 0xFF)*((color>>8) & 0xFF), 255);
            Uint32 b = min(((c>>16) & 0xFF)*((color>>16) & 0xFF), 255);
            *(Uint32*)ptr = 0xFF000000 | (b<<16) | (g<<8) | r;
          }
          break;
        }
      }
      break;
    case VS_DM_AND:
      switch (bpp) {
        case 4: *(Uint32*)ptr &= color; break;
      }
      break;    
    case VS_DM_OR:
      switch (bpp) {
        case 4: *(Uint32*)ptr |= color; break;
      }
      break;    
    case VS_DM_XOR:
      switch (bpp) {
        case 4: *(Uint32*)ptr ^= color; break;
      }
      break;
    case VS_DM_MIN:
      switch (bpp) {
        case 4: {
          if (UseMMX) {
            *(Uint32*)ptr = (int)_mm_min_pu8((__m64)*(Uint32*)ptr, (__m64)color);
          } else {
            Uint32 c = *(Uint32*)ptr; 
            Uint32 r = min((c & 0xFF), (color & 0xFF));
            Uint32 g = min(((c>>8) & 0xFF), ((color>>8) & 0xFF));
            Uint32 b = min(((c>>16) & 0xFF), ((color>>16) & 0xFF));
            *(Uint32*)ptr = 0xFF000000 | (b<<16) | (g<<8) | r;
          }
          break;
        }
      }
      break;    
    case VS_DM_MAX:
      switch (bpp) {
        case 4: {
          if (UseMMX) {
            *(Uint32*)ptr = (int)_mm_max_pu8((__m64)*(Uint32*)ptr, (__m64)color);
          } else {
            Uint32 c = *(Uint32*)ptr; 
            Uint32 r = max((c & 0xFF), (color & 0xFF));
            Uint32 g = max(((c>>8) & 0xFF), ((color>>8) & 0xFF));
            Uint32 b = max(((c>>16) & 0xFF), ((color>>16) & 0xFF));
            *(Uint32*)ptr = 0xFF000000 | (b<<16) | (g<<8) | r;
          }
          break;
        }
      }
      break;    
    case VS_DM_BLEND50:
      switch (bpp) {
        case 4: {
          if (UseMMX) {
            *(Uint32*)ptr = (int)_mm_avg_pu8((__m64)*(Uint32*)ptr, (__m64)color);
          } else {
            Uint32 c = *(Uint32*)ptr; 
            Uint32 r = ((c & 0xFF)+(color & 0xFF))>>1;
            Uint32 g = (((c>>8) & 0xFF)+((color>>8) & 0xFF))>>1;
            Uint32 b = (((c>>16) & 0xFF)+((color>>16) & 0xFF))>>1;
            *(Uint32*)ptr = 0xFF000000 | (b<<16) | (g<<8) | r;
          }
          break;
        }
      }
      break;
    default: break;
  }
  return(0);
}
