/* CLASS VIRTUALSCREEN_SDL (header) */

#ifndef CLASS_VIRSCRSDL_H
#define CLASS_VIRSCRSDL_H

#include "class_virscr.h"
#include <SDL/SDL.h>
#ifndef VSSDL_NO_TTF
  #include <SDL/SDL_ttf.h>
#endif
#ifndef VSSDL_NO_IMAGE
  #include <SDL/SDL_image.h>
#endif

#define VSSDL_ERR_NONE		0	// error codes
#define VSSDL_ERR_NULLBUFF	1
#define VSSDL_ERR_NOLOCK	2

#define VS_TM_NONE	0		// text-mode constants
#define VS_TM_SOLID	1
#define VS_TM_SHADED	2
#define VS_TM_BLENDED	3
#define VS_TM_NORENDER	0x100

#define VS_TA_LEFT	0		// text-align constants
#define VS_TA_CENTER	1
#define VS_TA_RIGHT	2

#define VS_BLUR_NORMAL		0	// blur-type constants
#define VS_BLUR_FAST		1	//   2 times faster than NORMAL
#define VS_BLUR_SUPERFAST	2	//   4 times faster than NORMAL

#define VS_FADE_NORMAL		0	// fadeout-type constants
#define VS_FADE_FAST		1
#define VS_FADE_SUPERFAST	2


class VirtualScreenSDL : public VirtualScreen {
 public:

  /* TEXT VARIABLES */ 
 
  /* Controls the type of text-render */
  /* Can be one of VS_TM_* constants */
  int TextMode;
  
  /* Controls the text alignment */
  /* Can be one of VS_TA_* constants */
  int TextAlign;

  
  /* CONSTRUCTORS */
  
  /* Creates a VS from an surface */
  /* Initializes the VS from an existing SDL surface. If screen is
     NULL it stops the execution of program. */
  VirtualScreenSDL(SDL_Surface *screen);

  /* Creates a VS with given sizes */
  /* Initializes the VS with a newly created surface. If it fail the 
     program is halted. */  
  VirtualScreenSDL(int sizex, int sizey);
  
  /* Creates a VS from an image */
  /* Initializes the VS with a newly created surface, filled with the 
     image specified by filename. If an error occurs (filename not 
     accessible or something else) the program is halted. */
#ifndef VSSDL_NO_IMAGE
  VirtualScreenSDL(const char *filename);
#endif
  
  /* DESTRUCTOR */
  
  /* Destroys VS and all the memory allocated.
     The SDL surface is freed only if it was created by VS */
  ~VirtualScreenSDL();
  

  /* SURFACE FUNCTIONS */
  
  /* Change the furface of the VS */
  /* Change the internal buffer with the given newscreen. The old one is
     freed only if it was created by VS. */
  int ChangeSurface(SDL_Surface *newscreen);
  
  int SetDisplayFormat();
  
  /* Returns the width of SDL surface */
  int GetWidth();

  /* Returns the height of SDL surface */
  int GetHeight();

  /* Locks the buffer */
  /* With SDL, it is necessary to lock the surface, before accessing it
     directly. This function does it, but only if it's needed.
     Used mostly internally. Returns 0 if no errors occurred */
  int Lock();
  
  /* Unlocks the buffer */
  /* The counterpart of Lock(); in order to show the surface or do blitting
     the surface must be unlocked. Returns always 0. */
  int Unlock();
  
  
  /* OTHER PROPERTIES */

  /* Enables MMX */
  /* Enables the use of MMX code (only if supported by CPU).
     It returns the new state of MMX-flag (0 if MMX not available) */
  int EnableMMX();

  /* Disables MMX */
  /* Turns off MMX optimizations. Returns always 0 */
  int DisableMMX();


  /* PIXEL PRIMITIVES */
  
  /* Draws a pixel */
  /* Draws a pixel at X,Y with COLOR, according with DrawMode;
     By default MMX is used (if supported by CPU). X and Y are checked 
     before drawing, and -1 is returned if pixel has not been drawn. */
  int PutPixel(int x, int y, int color);
  
  /* Gets a pixel */
  /* Returns the color of the pixel at X,Y.
     X and Y are checked and if the pixel is outside the screen area
     it returns a black opaque pixel (0xff000000) */
  int GetPixel(int x, int y);
  
  
  /* OTHER PRIMITIVES */
  
  /* Draws a line */
  /* Draws a line with COLOR, using Bresenham's algorithm.
     The function is optimized to draw horizontal lines */
  int Line(int x1, int y1, int x2, int y2, int color);
  
  /* Draws an empty rectangle */
  /* This function ensures that drawn pixels do not overlap, which is
     necessary with DrawMode different than VS_DM_NORMAL */
  int Rect(int x1, int y1, int x2, int y2, int color);
  
  /* Draws an empty triangle */
  /* Not yet finished, pixels may overlap */
  int Triangle(int x1, int y1, int x2, int y2, int x3, int y3, int color);

  /* int Circle(int x, int y, int rad, int color); */
  /* int FCircle(int x, int y, int rad, int color);*/
  

  /* TEXT FUNCTIONS */
  
  /**/
#ifndef VSSDL_NO_TTF
  int LoadFont(const char *fontname, int size);
  int ReloadFont(int fontindex, const char *fontname, int size);
  int SetFontStyle(int fontindex, int style);
  SDL_Surface *Print(int x, int y, const char *text, int color, int colorback);
  SDL_Surface *Print(int fontindex, int x, int y, const char *text, int color, int colorback);
  int GetPrintSize(int fontindex, const char *text, int *sizex, int *sizey);
  int PrintRect(int fontindex, int x1, int y1, int x2, int y2, const char *text, int color, int colorback);
#endif

  // image functions
#ifndef VSSDL_NO_IMAGE
  int LoadImage(const char *filename);
  int LoadImageXPM(char *xpm[]);
  int SaveImage(int x1, int y1, int x2, int y2, char *file);
  int SaveImage(const char *file);
#endif

  int Show();
  int Show(int x1, int y1, int x2, int y2);
  int Clear(int color);
  int CopyTo(int destx,int desty, VirtualScreenSDL *dest);
  int CopyRectTo(int x1,int y1, int x2,int y2, int destx,int desty, VirtualScreenSDL *dest);
  
  // effects
  int Blur(int x1,int y1, int x2,int y2, int type);
  int Fadeout(int x1,int y1, int x2,int y2, int color, int speed, int type);
  int StretchRectTo(int x1,int y1, int x2,int y2, int x3,int y3, int x4,int y4, VirtualScreenSDL *dest);
  int Pixelize(int x1,int y1, int x2,int y2, int size);
  int Contrast(int x1,int y1, int x2,int y2, char amount);
  int Brightness(int x1,int y1, int x2,int y2, char amount);
  int Posterize(int x1,int y1, int x2,int y2, unsigned char levels);

 private:
  SDL_Surface *screen;
  char IsLocked, UseMMX, FreeSurface;
  char SkipLineFirst;

#ifndef VSSDL_NO_TTF
  TTF_Font **fonts;
  int fontsLoaded;
#endif
  
  int HardPutPixel(Uint8 *ptr, int bpp, int color);
  int HLine(int x1,int y1, int x2, int color);
  int VLine(int x1,int y1, int y2, int color);
  
  int CatchError(int n, void *data);
};

#endif
