/* CLASS_KRYCUBE.CPP */

#include <stdlib.h>
#include <string.h>

#include "class_krycube.h"
#include "errors.h"

Krycube::Krycube(int size) {
  flag_dontSkimpMem = 1;
  size += size & 0x01; // size must be a multiple of 2
  this->size = size;
  cubesize = size*size*size*sizeof(char);
  bytestoread = cubesize >> 3;
  cube = (char*)malloc(cubesize);
  if (cube == NULL) {
    error(ERR_NOCUBEMEM, NULL);
    exit(1);
  }
}

Krycube::~Krycube() {
  free(cube);
}

int Krycube::Resize(int newsize) {
  int ret = 0;
  int oldcubesize = cubesize;
  size = newsize + (newsize & 0x1);
  cubesize = size*size*size*sizeof(char);
  bytestoread = cubesize >> 3;
  if (flag_dontSkimpMem) { // don't realloc if newsize is smaller than size
    if (oldcubesize >= cubesize) return(0);
  }
  cube = (char*)realloc(cube, cubesize);
  if (cube == NULL) { // out of memory
    exit(ret = error(ERR_NOCUBEMEM, NULL));
  }
  return(ret);
}

int Krycube::Fill(long int length, char *data) {
  if (length > bytestoread) {
    // error?
    length = bytestoread;
  }
  int pos = 0;
  for (int i=0; i<length; i++) {
    char curbyte = *(data+i);
    for (int j=0; j<8; j++, curbyte >>= 1) *(cube+pos++) = curbyte & 0x01;
  }
  return(0);
}

int Krycube::Empty(char *data) {
  int pos = 0;
  for (int i=0; i<bytestoread; i++) {
    char curbyte = 0;
    for (int j=0; j<8; j++) curbyte |= *(cube+pos++) << j;
    *(data+i) = curbyte;
  }
  return(0);
}

void Krycube::Get_shiftvalues(char *pwd, int magic, int *values) {
  int pwdlen = strlen(pwd);
  for (int i=0; i<6; i++)
    for (int j=0; j<size; j++)
      *(values+i*6+j) = (pwd[(i+j+pwd[j%pwdlen])%pwdlen]*magic-i*j)%size;
}

int Krycube::Crypt(char *pwd, int magic) {
  int shval[6*size];
  Get_shiftvalues(pwd, magic, shval);
  for (int i=0; i<size; i++) {
    Shift(PLANE_XY_DIR_X, i, shval[0*6+i]);
    Shift(PLANE_XY_DIR_Y, i, shval[1*6+i]);
    Shift(PLANE_XZ_DIR_X, i, shval[2*6+i]);
    Shift(PLANE_XZ_DIR_Z, i, shval[3*6+i]);
    Shift(PLANE_YZ_DIR_Y, i, shval[4*6+i]);
    Shift(PLANE_YZ_DIR_Z, i, shval[5*6+i]);
  }
  memset(shval, 0, sizeof(shval));		// clear memory
  return(0);
}

int Krycube::Decrypt(char *pwd, int magic) {
  int shval[6*size];
  Get_shiftvalues(pwd, magic, shval);
  for (int i=size-1; i>=0; i--) {
    Shift(PLANE_YZ_DIR_Z, i, -shval[5*6+i]);
    Shift(PLANE_YZ_DIR_Y, i, -shval[4*6+i]);
    Shift(PLANE_XZ_DIR_Z, i, -shval[3*6+i]);
    Shift(PLANE_XZ_DIR_X, i, -shval[2*6+i]);
    Shift(PLANE_XY_DIR_Y, i, -shval[1*6+i]);
    Shift(PLANE_XY_DIR_X, i, -shval[0*6+i]);
  }
  memset(shval, 0, sizeof(shval));		// clear memory
  return(0);
}

char Krycube::Get_cubebit(int x, int y, int z) {
  return(*(cube+z*size*size+y*size+x));
}

void Krycube::Set_cubebit(int x, int y, int z, char bit) {
  *(cube+z*size*size+y*size+x) = bit;
}

int Krycube::Shift(char plane_dir, int xyz, int amount) {
  int ret = 0;
  int i, j, ia, ja, ptr = 0;
  char *tplane = (char*)malloc(sizeof(char)*size*size);
  if (tplane == NULL) exit(ret = error(ERR_NOCUBEMEM, NULL));
  int *ptr_x, *ptr_y, *ptr_z, *ptr_cx, *ptr_cy, *ptr_cz;
  
  switch (plane_dir) {
    case PLANE_XY_DIR_X: 
      ptr_x = &ia;
      ptr_cx = &i;
      ptr_cy = ptr_y = &j;
      ptr_cz = ptr_z = &xyz;
      break;
    case PLANE_XY_DIR_Y: 
      ptr_cx = ptr_x = &i;
      ptr_y = &ja;
      ptr_cy = &j;
      ptr_cz = ptr_z = &xyz;
      break;
    case PLANE_XZ_DIR_X: 
      ptr_x = &ia;
      ptr_cx = &i;
      ptr_cy = ptr_y = &xyz;
      ptr_cz = ptr_z = &j;
      break;
    case PLANE_XZ_DIR_Z: 
      ptr_cx = ptr_x = &i;
      ptr_cy = ptr_y = &xyz;
      ptr_z = &ja;
      ptr_cz = &j;
      break;
    case PLANE_YZ_DIR_Y:
      ptr_cx = ptr_x = &xyz;
      ptr_y = &ia;
      ptr_cy = &i;
      ptr_cz = ptr_z = &j;
      break;
    case PLANE_YZ_DIR_Z:
      ptr_cx = ptr_x = &xyz;
      ptr_cy = ptr_y = &i;
      ptr_z = &ja;
      ptr_cz = &j;
      break;
  }

  amount %= size;
  for (j=0; j<size; j++) {
    ja = (j+size-amount)%size;
    for (i=0; i<size; i++) {
      ia = (i+size-amount)%size;
      *(tplane+ptr++) = Get_cubebit(*ptr_x, *ptr_y, *ptr_z);
    }
  }
  for (j=0, ptr=0; j<size; j++) {
    for (i=0; i<size; i++) Set_cubebit(*ptr_cx, *ptr_cy, *ptr_cz, *(tplane+ptr++));
  }
  free(tplane);
  return(ret);
}
