/* CLASS_KRYFILE.CPP */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fnmatch.h>
#include <utime.h>
#include <time.h>

#include "class_kryfile.h"
#include "class_krycube.h"
#include "errors.h"


char *format_attr(long int attrib) {
  char *a = (char*)malloc(12);
  int t;
  strcpy(a, "----------");
  switch (attrib & 0070000) {
    case S_IFBLK: a[0] = 'b'; break;
    case S_IFDIR: a[0] = 'd'; break;
    case S_IFCHR: a[0] = 'c'; break;
    case S_IFIFO: a[0] = 'f'; break;
  }
  t = attrib & S_IRWXU;
  if ((t & (S_IXUSR << 0)) == S_IXUSR) a[3] = 'x';
  if ((t & (S_IXUSR << 1)) == S_IWUSR) a[2] = 'w';
  if ((t & (S_IXUSR << 2)) == S_IRUSR) a[1] = 'r';
  t = attrib & S_IRWXG;
  if ((t & (S_IXGRP << 0)) == S_IXGRP) a[6] = 'x';
  if ((t & (S_IXGRP << 1)) == S_IWGRP) a[5] = 'w';
  if ((t & (S_IXGRP << 2)) == S_IRGRP) a[4] = 'r';
  t = attrib & S_IRWXO;
  if ((t & (S_IXOTH << 0)) == S_IXOTH) a[9] = 'x';
  if ((t & (S_IXOTH << 1)) == S_IWOTH) a[8] = 'w';
  if ((t & (S_IXOTH << 2)) == S_IROTH) a[7] = 'r';
  return(a);
}

Kryfile::Kryfile() {
  flag_showPercent = 1;
  flag_useMD5Hash = 1;
  archive = NULL;
  table = NULL;
  arc_filename = NULL;
  arc_head = NULL;
}

Kryfile::~Kryfile() {
  if (archive != NULL) fclose(archive);
  if (table != NULL) free(table);
  if (arc_filename != NULL) free(arc_filename);
  if (arc_head != NULL) free(arc_head);
}

long int Kryfile::Checksum(int length, char *buffer) {
  long int sum = 0;
  for (int i=0; i<length; i++) sum += *(buffer+i);
  return(sum);
}

char *Kryfile::Getshortname(char *name) {
  char *ptr = strrchr(name, '/');
  return((ptr == NULL)?name:(ptr+1));
}

long int Kryfile::Get_bytestoread(int edge) { // edge must be multiple of 2
  return((edge*edge*edge) >> 3);
}

long int Kryfile::Get_edgeofcube(int datasize) {
  int ret = 0;
  while (ret*ret*ret <= (datasize<<3)) ret += 2;
  ret -= 2;
  if (ret*ret*ret != (datasize<<3)) ret += 2; // check if exact cubic root
  return(ret);
}

int Kryfile::Calccubes(char *pwd, int magic, long int filesize, int *ncubes, int **cubesizes) {
  const int maxcube = 128;
  const int maxdata = 262144; // 256 Kb
  const int mincube = 4;
  const int mindata = 8; // 8 bytes
  int i, blocks, maxsize;
  int *cubes, sizes[4], *sizes2;
  int pwdlen = strlen(pwd);
  
  if (filesize <= 512) {
    blocks = filesize/mindata+1; // calc necessary blocks
    cubes = (int*)malloc(sizeof(int)*blocks);
    for (i=0; i<blocks; i++) cubes[i] = mincube+(i&1)*2;
  } else {
    // 1st pass: create 4 blocks
    blocks = 4;
    for (i=0; i<blocks-1; i++) sizes[i] = filesize >> 2;
    sizes[blocks-1] = (filesize >> 2)+(filesize & 3);
    // 2nd pass: divide each of the 4 blocks in up to 16 blocks
    int pos2 = 0;
    int newblocks = 0;
    sizes2 = (int*)malloc(sizeof(int)*64); // allocate array
    for (i=0; i<4; i++) {
      int div = ((magic >> (i<<2)) & 0x0F) + 1;
      int cursize = 0;
      for (int j=0; j<div-1; j++) {
        int subdiv = ((*(pwd+((j+i)>>1)%pwdlen)>>(((j+i)&1)<<2)) & 0x0F) + 1;
        sizes2[pos2+j] = subdiv*sizes[i]/256;
        cursize += sizes2[pos2+j];
      }
      sizes2[pos2+div-1] = sizes[i]-cursize;
      if (div > 1) {
        int last = sizes[i]/div;
        int tolow = sizes2[pos2+div-1]-last;
        sizes2[pos2+div-1] = last;
        for (int j=0; j<div-2; j++) {
          sizes2[pos2+j] += tolow/(div-1);
        }
        sizes2[pos2+div-2] += tolow-(div-2)*(tolow/(div-1));
      }
      pos2 += div;
      newblocks += div;
    }
    // 3rd pass: resize blocks if too big
    int iblocks = newblocks;
    for (i=0; i<iblocks; i++) {
      if (sizes2[i] > maxdata) {
        int newslots = sizes2[i]/maxdata+1;
        int lastslot = sizes2[i]%maxdata;
        sizes2 = (int*)realloc(sizes2, sizeof(int)*(newblocks+newslots));
        sizes2[i] = sizes2[newblocks-1];
        for (int j=0; j<newslots-1; j++) {
          sizes2[newblocks-1+j] = maxdata;
        }
        if (lastslot > 0) {
          sizes2[newblocks+newslots-2] = lastslot;
          newblocks += newslots-1;
        } else newblocks += newslots-2;
      }
    }
    // copy to the final cube-array
    cubes = (int*)malloc(sizeof(int)*newblocks);
    for (i=0; i<newblocks; i++) cubes[i] = Get_edgeofcube(sizes2[i]);
    blocks = newblocks;
    free(sizes2);
  }
  // checks if too many cubes
  int totsize = 0;
  maxsize = 0;
  for (i=0; i<blocks; i++) {
    if (cubes[i] > maxsize) maxsize = cubes[i];
    totsize += Get_bytestoread(cubes[i]);
    if (totsize >= filesize) break;
  }
  blocks = i+1;
  // resize last cube
  totsize -= Get_bytestoread(cubes[blocks-1]);
  int lastcube = cubes[blocks-1];
  do {
    lastcube -= 2;
  } while (totsize+Get_bytestoread(lastcube) >= filesize);
  cubes[blocks-1] = lastcube+2;

  *ncubes = blocks;
  *cubesizes = cubes;
  return(maxsize);
}

int Kryfile::CalcMD5Hash(char *filename, char *hash) { // hash has to be almost 32 bytes large
  int desc, ret = 0;
  char tname[50] = "kryXXXXXX\0";
  printf("Calculating MD5 hash... ");
  memset(hash, 0, 32);
  if ((desc = mkstemp(tname)) == -1) {
    return(ret = error(ERR_NOMD5, filename));
  } else {
    char cmdline[strlen(filename)+50];
    sprintf(cmdline, "md5sum %s > %s", filename, tname);
    if ((ret = system(cmdline)) != 0) error(ERR_NOMD5, filename);
    else read(desc, hash, 32);
    close(desc);
    remove(tname);
  }
  if (ret == 0) printf("ok\n");
  return(ret);
}

int Kryfile::Createfile(const char *filename) {
#ifdef DEBUG
  printf("Creating archive %s...\n", filename);
#endif
  int ret = 0;
  archive = fopen(filename, "r");
  if (archive != NULL) {		// archive already exists
    fclose(archive);
    ret = warn(WARN_FILEEXISTS, (char*)filename);
    int aret;
    do { aret = ask(ASK_OVERWRITE, NULL); } while (aret == -1);
    if (aret == 1) {	// no
      exit(ret);
    }
  }
  
  archive = fopen(filename, "wb+");
  if (archive == NULL) exit(ret = error(ERR_ARCCREAT, (void *)filename));
  arc_filename = strdup(filename);
  return(ret);
}

int Kryfile::Openfile(const char *filename) {
#ifdef DEBUG
  printf("Opening archive %s...\n", filename);
#endif
  int ret = 0;
  filesadded = 1; 	// DO NOT CHANGE, otherwise the archive will be erased!
  
  archive = fopen(filename, "rb");
  if (archive == NULL) exit(ret = error(ERR_OPENFILE, (char*)filename));
  return(ret);
}

int Kryfile::Readheader() {
#ifdef DEBUG
  printf("Reading header...\n");
#endif
  int ret = 0;
  if (archive == NULL) exit(ret = error(ERR_FILECLOSE, NULL));

  arc_head = (Kryheader_090 *)malloc(sizeof(Kryheader_090));
  if (fread(arc_head, sizeof(Kryheader_090), 1, archive) != 1) {	// too small
    exit(ret = error(ERR_BADARCHIVE, NULL));
  };
  if (*((long int *)(arc_head->signature)) != HEADER_SIGN) {		// bad signature
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }
  if (arc_head->size != sizeof(Kryheader_090) || arc_head->version != 0x0090) {		// version mismatch
    // TODO: check for other versions
    ret = warn(WARN_BADVERSION, NULL);
  }
  if (arc_head->checksum != Checksum(sizeof(Kryheader_090)-6, (char*)arc_head)) {	// bad checksum
    ret = warn(WARN_ARCCORRUPT, NULL);
  }
  if (*(short*)(arc_head->signature_end) != HEADER_SIGNEND) {		// bad end-signature
    ret = warn(WARN_ARCCORRUPT, NULL);
  }

  filesadded = arc_head->files;
  if (fseek(archive, arc_head->table_ptr, SEEK_SET) != 0) {		// can't find the table
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }
  int tabsize = sizeof(Krytable_090)+4*(filesadded-1);
  table = (Krytable_090*)realloc(table, tabsize);
  memset(table, 0x66, tabsize);
  if (fread(table, tabsize, 1, archive) != 1) {		// table too small, but try to continue
    ret = warn(WARN_ARCCORRUPT, NULL);
  }
  if ((table->sign == 0x66) || (table->pointers == 0x66666666)) {  // nothing good
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }
  if (table->sign != 0x00) {			// table possibly corrupted
    ret = warn(WARN_ARCCORRUPT, NULL);
  }
  if (table->pointers != filesadded) {		// wrong amount of files, take the smaller
    ret = warn(WARN_ARCCORRUPT, NULL);
    if (filesadded > table->pointers) filesadded = table->pointers;
  }

  return(ret);
}

int Kryfile::Writeheader(int count) {
#ifdef DEBUG
  printf("Writing header...\n");
#endif
  int ret = 0;
  if (archive == NULL) exit(ret = error(ERR_FILECLOSE, NULL));

  // Writes the Archive Header
  Kryheader_090 head;
  *((long int *)(head.signature)) = HEADER_SIGN; 		// signature
  head.size = sizeof(head); 					// size of header
  head.version = 0x0090; 					// program version
  head.files = count; 						// must be changed if error occurs!
  head.properties = 0; 						// properties of archive
  head.table_ptr = sizeof(head); 				// file-header pointer table follows the header;
  head.reserved = 0; 						// be compatible with future versions
  head.checksum = Checksum(sizeof(head)-6, (char*)&head); 	// calculates checksum
  *((short *)(head.signature_end)) = HEADER_SIGNEND;		// signature end
  if (fwrite(&head, sizeof(head), 1, archive) != 1) {		// writes to file
    exit(ret = error(ERR_WRITE, NULL));
  };

  // Writes the pointers table
  // Remember to adjust it if some files fail
  int tabsize = sizeof(Krytable_090)+4*(count-1);
  table = (Krytable_090 *)malloc(tabsize);	// allocates table
  table->sign = 0;
  table->pointers = count;
  table->files_ptr[0] = 0;					// must be changed later
  filesadded = 0;
  if (fwrite(&table, tabsize, 1, archive) != 1) {		// writes to file
    exit(ret = error(ERR_WRITE, NULL));
  };
  return(ret);
}

int Kryfile::Cryptdata(char *filename, char *pwd) {
#ifdef DEBUG
  printf("Crypting file %s...\n", filename);
#endif
  int ret = 0;
  if (archive == NULL) exit(ret = error(ERR_FILECLOSE, NULL));
  
  table->files_ptr[filesadded] = ftell(archive);
  
  FILE *file = fopen(filename, "rb");
  if (file == NULL) return(ret = error(ERR_OPENFILE, filename));
  Kryfileheader_090_1 *filehead1;
  Kryfileheader_090_2 filehead2;
  srand(time(NULL));
  srand48(time(NULL));
  
  char *shortname = Getshortname(filename);
  int fh1size = sizeof(Kryfileheader_090_1)+strlen(shortname);
  filehead1 = (Kryfileheader_090_1 *)malloc(fh1size);
  
  filehead1->sign = FILEHEADER_SIGN;
  filehead1->size = fh1size + sizeof(Kryfileheader_090_2);
  filehead1->namelength = strlen(shortname)+1;
  strcpy(filehead1->name, shortname);
  
  struct stat fileinfo;
  if (stat(filename, &fileinfo) != 0) warn(WARN_NOSTAT, filename); 
  filehead2.filesize = fileinfo.st_size;
  filehead2.attributes = fileinfo.st_mode;
  filehead2.date_create = fileinfo.st_mtime;
  filehead2.date_access = fileinfo.st_atime;
  filehead2.randval = lrand48();
  filehead2.properties = flag_useMD5Hash & 0x01;
  filehead2.data_ptr = table->files_ptr[filesadded] + filehead1->size; 
  if (flag_useMD5Hash)
    if (CalcMD5Hash(filename, filehead2.md5sum) != 0) filehead2.properties ^= 0x01;
  *((short*)(filehead2.sign_end)) = FILEHEADER_SIGNEND;

  int written = fwrite(filehead1, fh1size, 1, archive);
  written += fwrite(&filehead2, sizeof(filehead2), 1, archive);
  free(filehead1);
  if (written != 2) return(ret = error(ERR_WRITE, NULL));
  
  int blocks, maxsize, *cubesizes;
  maxsize = Calccubes(pwd, filehead2.randval, filehead2.filesize, &blocks, &cubesizes);
  long int filecsize, bufsize = Get_bytestoread(maxsize);
  char *buf = (char *)malloc(bufsize);
  Krycube *cube = new Krycube(cubesizes[0]);
  
#ifdef DEBUG
  printf("Blocks: ");
  for (int i=0; i<blocks; i++) printf("%4i", cubesizes[i]);
  printf("\n");
#endif

  for (int i=0,filecsize=0; i<blocks; i++) {
#ifdef DEBUG
  printf("Crypting cube #%i\n", i);
#endif
    if (feof(file)) break;
    if (flag_showPercent) {
      printf("%i%%\r", (i*100)/blocks);
      fflush(stdout);
    }  
    cube->Resize(cubesizes[i]);
    int blocksize = cube->Get_bytestoread();
    fread(buf, blocksize, 1, file);
    if (feof(file)) { // eof, fill buffer with garbage
      for (int j=(filehead2.filesize-filecsize); j<blocksize; j++) *(buf+j) = rand();
    }
    cube->Fill(blocksize, buf);
    cube->Crypt(pwd, i*filesadded+filehead2.filesize+filehead2.randval);
    cube->Empty(buf);
    if (fwrite(buf, blocksize, 1, archive) != 1) {
      return(ret = error(ERR_WRITE, NULL));
    }
    filecsize += blocksize;
  }
  if (flag_showPercent) printf("100%%\r");

  delete cube;
  free(cubesizes);
  free(buf);
  fclose(file);
  filesadded++;
  return(ret);
}

int Kryfile::Readfilehead(Kryfileheader_090_1 **fh1, Kryfileheader_090_2 **fh2, int file) {
#ifdef DEBUG
  printf("Reading file-header #%i...\n", file);
#endif

  int ret = 0;
  *fh1 = (Kryfileheader_090_1*)malloc(sizeof(Kryfileheader_090_1));
  *fh2 = (Kryfileheader_090_2*)malloc(sizeof(Kryfileheader_090_2));

  if (fseek(archive, table->files_ptr[file], SEEK_SET) != 0) {
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }

  if (fread(*fh1, sizeof(Kryfileheader_090_1), 1, archive) != 1) {
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }
  if ((*fh1)->sign != FILEHEADER_SIGN) 
    ret = warn(WARN_ARCCORRUPT, NULL);
  if ((*fh1)->size != sizeof(Kryfileheader_090_1)+sizeof(Kryfileheader_090_2)-1+(*fh1)->namelength) {
    // TODO: next versions
    ret = warn(WARN_ARCCORRUPT, NULL);
  }
  *fh1 = (Kryfileheader_090_1*)realloc(*fh1, sizeof(Kryfileheader_090_1)+(*fh1)->namelength);
  if (fread((*fh1)->name+1, (*fh1)->namelength-1, 1, archive) != 1) {
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }
  if (fread(*fh2, sizeof(Kryfileheader_090_2), 1, archive) != 1) {
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }
  if (*((short*)((*fh2)->sign_end)) != FILEHEADER_SIGNEND)
    ret = warn(WARN_ARCCORRUPT, NULL);
  if (fseek(archive, (*fh2)->data_ptr, SEEK_SET) != 0) {
    exit(ret = error(ERR_BADARCHIVE, NULL));
  }
  return(ret);
}

int Kryfile::Decryptdata(char *out_path, int out_patcount, char *patterns[], char *pwd) {
#ifdef DEBUG
  printf("Decrypting archive...\n");
#endif

  int ret = 0;
  int evermatch = 0;
  if (archive == NULL) exit(ret = error(ERR_FILECLOSE, NULL));

  Kryfileheader_090_1 *filehead1;
  Kryfileheader_090_2 *filehead2;
  
  for (int i=0; i<filesadded; i++) {
  
    Readfilehead(&filehead1, &filehead2, i);
    
    int matched = 0;
    for (int j=0; j<out_patcount; j++) {
      if (fnmatch(patterns[j], filehead1->name, 0) == 0) {
        matched = 1;
        evermatch = 1;
        break;
      }
    }
    if (!matched) continue;

    char *path = (char*)malloc(sizeof(char)*(strlen(out_path)+strlen(filehead1->name)+2));
    strcpy(path, out_path);
    strcat(path, filehead1->name);
    
    printf("Sto estraendo %s...\n", path);
    
    FILE *file = fopen(path, "r");
    if (file != NULL) {		// file exists!
      fclose(file);
      ret = warn(WARN_FILEEXISTS, path);
      int aret;
      do { aret = ask(ASK_OVERWRITE, NULL); } while (aret == -1);
      if (aret == 1) {	// no
        continue;
      }
    }

    file = fopen(path, "wb");
    if (file == NULL) {
      ret = error(ERR_ARCCREAT, path);
      continue;
    }
    
    int blocks, maxsize, *cubesizes;
    maxsize = Calccubes(pwd, filehead2->randval, filehead2->filesize, &blocks, &cubesizes);
    long int filecsize, bufsize = Get_bytestoread(maxsize);
    char *buf = (char *)malloc(bufsize);
    Krycube *cube = new Krycube(cubesizes[0]);

    for (int j=0,filecsize=0; j<blocks; j++) {
#ifdef DEBUG
  printf("Decrypting cube #%i\n", j);
#endif
      if (flag_showPercent) {
        printf("%i%%\r", (j*100)/blocks);
        fflush(stdout);
      }  
      cube->Resize(cubesizes[j]);
      int blocksize = cube->Get_bytestoread();
      if (fread(buf, blocksize, 1, archive) != 1) {
        ret = warn(WARN_ARCCORRUPT, NULL);
        break;
      };
      cube->Fill(blocksize, buf);
      cube->Decrypt(pwd, j*i+filehead2->filesize+filehead2->randval);
      cube->Empty(buf);
      if (fwrite(buf, (j==blocks-1)?(filehead2->filesize-filecsize):blocksize, 1, file) != 1) {
        ret = error(ERR_WRITE, NULL);
        break;
      }
      filecsize += blocksize;
    }
    if (flag_showPercent) printf("100%%\r");
    
    delete cube;
    free(cubesizes);    
    free(buf);
    fclose(file);

    if (chmod(path, filehead2->attributes) != 0) {
      ret = warn(WARN_NOWRITESTAT, path);
    }
    struct utimbuf times;
    times.actime = filehead2->date_access;
    times.modtime = filehead2->date_create;
    if (utime(path, &times) != 0) {
      ret = warn(WARN_NOWRITESTAT, path);
    }
    // checks MD5 hash
    if (flag_useMD5Hash && (filehead2->properties == 0x01)) {
      char hash[40];
      int r = CalcMD5Hash(path, hash);
      if (r != 0) ret = error(ERR_NOMD5, path);
      else {
        if (memcmp(hash, filehead2->md5sum, 32) != 0)
          ret = error(ERR_BADMD5, path); 
      }
    }
    free(path);
  }
  if (!evermatch) error(ERR_NOMATCH, NULL);
  free(filehead1);
  free(filehead2);
  return(ret);
}

int Kryfile::List(int out_patcount, char *patterns[]) {
#ifdef DEBUG
  printf("Listing archive...\n");
#endif
  int ret = 0;
  int evermatch = 0, counter = 0;
  if (archive == NULL) exit(ret = error(ERR_FILECLOSE, NULL));

  Kryfileheader_090_1 *filehead1;
  Kryfileheader_090_2 *filehead2;
  
  for (int i=0; i<filesadded; i++) {
    Readfilehead(&filehead1, &filehead2, i);
    
    int matched = 0;
    for (int j=0; j<out_patcount; j++) {
      if (fnmatch(patterns[j], filehead1->name, 0) == 0) {
        matched = 1;
        evermatch = 1;
        break;
      }
    }
    if (!matched) continue;
    time_t date = filehead2->date_create;
    struct tm *infotime = localtime(&date);
    char *attr = format_attr(filehead2->attributes);
    printf("%s  %10lu  %4i-%02i-%02i %02i:%02i  %s\n", 
        format_attr(filehead2->attributes),				// attributes
        filehead2->filesize,						// size
    	1900+infotime->tm_year, 1+infotime->tm_mon, infotime->tm_mday, 	// date
    	infotime->tm_hour, infotime->tm_min, 				// time
    	filehead1->name);						// name
    free(attr);	

    counter++;
  }
  if (!evermatch) error(ERR_NOMATCH, NULL);
  else printf("%d files trovati\n", counter);
  free(filehead1);
  free(filehead2);
  return(ret);
}

int Kryfile::Finalize() {
#ifdef DEBUG
  printf("Completing archive...\n");
#endif
  int ret = 0;
  if (archive == NULL) exit(ret = error(ERR_FILECLOSE, NULL));
  if (filesadded == 0) return(ret = error(ERR_NOFILESADDED, NULL));

  Kryheader_090 head;
  fseek(archive, 0, SEEK_SET);
  fread(&head, sizeof(head), 1, archive);
  head.files = filesadded;
  head.checksum = Checksum(sizeof(head)-6, (char*)&head);
  fseek(archive, 0, SEEK_SET);
  fwrite(&head, sizeof(head), 1, archive);

  table->pointers = filesadded;
  fwrite(table, sizeof(Krytable_090)+4*(filesadded-1), 1, archive);
}

int Kryfile::Close() {
#ifdef DEBUG
  printf("Closing archive...\n");
#endif
  int ret = 0;
  if (archive == NULL) {
    exit(ret = error(ERR_FILECLOSE, NULL));
  } else {
    fclose(archive);
    archive = NULL;
    if (filesadded == 0) remove(arc_filename);
  }  
  return(ret);
}
