#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <glob.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>

/* to enable debug use the -DDEBUG switch in g++ command line */
//#define DEBUG

#include "errors.h"
#include "class_kryfile.h"

#define min(a,b)	(((a)<(b))?(a):(b))
#define max(a,b)	(((a)>(b))?(a):(b))

#define VERSION_HI  0
#define VERSION_LO  9

#define ACTION_NONE		0
#define ACTION_CRYPT		1
#define ACTION_DECRYPT		2
#define ACTION_LIST		3

const char *txt_welcome = "Kry  -  Versione 0.9 BETA2\n";

char flag_percent = 1;
char flag_md5 = 1;

int action = ACTION_NONE;
char *arc_filename = NULL;
char **fil_filename = NULL;
int fil_count = 0;
char pwd[100];
int pwd_complexity = 0;

// clears the buffer containing the clear password and exits
void exitfunc(void) {
  memset(pwd, 0, sizeof(pwd));
  pwd_complexity = 0;
}

// shows a brief usage
void show_syntax() {
  printf("Uso:\n"\
  	 "  kry <azione> [opzioni] <archivio> [ <dir> <files> ]\n\n"\
  	 "Azioni:\n"\
  	 "  -c    Cripta i <files> nell'<archivio>\n"\
  	 "  -d    Decripta i files in <archivio> che corrispondono al modello\n"\
  	 "        <files> nella directory <dir>\n"\
  	 "  -l    Mostra i files contenuti in <archivio> che corrispondono al\n"\
  	 "        modello <files>\n\n"\
  	 "Opzioni:\n"\
  	 "  -p    Non visualizza la percentuale di avanzamento\n"\
  	 "  -m    Salta il controllo MD5 dei files\n"\
  	 "  -h    Mostra questa guida\n\n"\
  	 "Note:\n"\
  	 "  Se <files> contiene caratteri jolly, includerlo tra virgolette.\n\n");
}

// parses the command line
// returns 0 if no errors occurred, otherwise the error/warning number 
int parse_cmd(int argc, char *argv[]) {
  int ret = 0;
  if (argc <= 1) {
    show_syntax();
    exit(0);
  }
  for (int i=1; i<argc; i++) {
    if (argv[i][0] == '-') { // it is an option
      if (argv[i][2] != '\0') { // too long
	ret = error(ERR_BADOPT, argv[i]);
	continue;
      }
      switch (argv[i][1]) { // parse option
	case 'c': action = ACTION_CRYPT; break;
	case 'd': action = ACTION_DECRYPT; break;
	case 'l': action = ACTION_LIST; break;
	case 'p': flag_percent = 0; break;
	case 'm': flag_md5 = 0; break;
	case 'h': show_syntax(); exit(0); break;
	default: ret = error(ERR_BADOPT, argv[i]);
      }
    } else { // it is a filename
      if (action == ACTION_NONE) exit(ret = error(ERR_NOACTION, NULL));
      if (arc_filename == NULL) {		// it is the archive name
        arc_filename = strdup(argv[i]);
      } else {					// it is a file name
        fil_filename = (char**)realloc(fil_filename, (fil_count+1)*sizeof(char*));
        fil_filename[fil_count++] = strdup(argv[i]);
      }
    }
  }
  if (arc_filename == NULL) exit(ret = error(ERR_NOARCHIVE, NULL));
  if (action == ACTION_CRYPT && fil_count == 0) exit(ret = error(ERR_NOFILES, NULL));
  if (action == ACTION_CRYPT) {  // expands file-names only if in crypt-mode
    glob_t globs;
    glob(fil_filename[0], GLOB_NOSORT | GLOB_BRACE | GLOB_TILDE, NULL, &globs);
    for (int i=1; i<fil_count; i++) 
      glob(fil_filename[i], GLOB_NOSORT | GLOB_APPEND | GLOB_BRACE | GLOB_TILDE, NULL, &globs);
    for (int i=0; i<fil_count; i++) free(fil_filename[i]);		// frees previous filenames
    fil_count = globs.gl_pathc;
    fil_filename = (char**)realloc(fil_filename, fil_count*sizeof(char*));
    for (int i=0; i<fil_count; i++) fil_filename[i] = strdup(globs.gl_pathv[i]);
    globfree(&globs);
    if (fil_count == 0) exit(ret = error(ERR_NOMATCH, NULL));
  }
  return(ret);
}

// expands strings in pattern in directories or files
// returns 0 if no errors occurred, otherwise the error/warning number 
int expand_dirfile(int patcount, char *pattern[], char **out_dir, int *out_filecount, char **out_files[]) {
  int ret = 0;
  DIR *mydir;
  *out_filecount = 0;
  for (int i=0; i<patcount; i++) {
    mydir = opendir(pattern[i]);
    if (mydir != NULL) {	// directory exists!
      closedir(mydir);
      if (*out_dir == NULL) {
        *out_dir = (char*)malloc(strlen(pattern[i])+3);
        strcpy(*out_dir, pattern[i]);
      }
      else ret = warn(WARN_MANYOUTDIR, pattern[i]);
    } else {			// so it's a file pattern
      *out_files = (char**)realloc(*out_files, sizeof(char*)*((*out_filecount)+1));
      (*out_files)[(*out_filecount)++] = strdup(pattern[i]);
    }
  }
  if (*out_dir == NULL) *out_dir = strdup("./"); 
  else if ((*out_dir)[strlen(*out_dir)-1] != '/') strcat(*out_dir, "/");
  if (*out_filecount == 0) {
    *out_files = (char**)realloc(*out_files, sizeof(char*));
    (*out_files)[(*out_filecount)++] = strdup("*");
  }
  return(ret);
}

// asks the password to the user
int ask_pwd(char repeat) {
  int ret = 0;
  int len, maxpwd = sizeof(pwd); // max length of password (including final '\0')
  do {
    char *pass;
    pass = getpass("Inserisci la password: ");
    len = strlen(pass);
    strncpy(pwd, pass, maxpwd);
    pwd[maxpwd-1] = '\0';
    if (repeat) {
      pass = getpass("Conferma la password: ");
      if (strncmp(pass, pwd, maxpwd-1) != 0) {
        ret = error(ERR_PWDNOMATCH, NULL);
        len = 0;
      }
    }
  } while (len == 0);  
  if (len > maxpwd) ret = warn(WARN_PWDCUT, &maxpwd);
  return(ret);
}

// calculates the password complexity (a bit useless)
int calc_pwd_complexity() {
  int pwdlen = strlen(pwd);
  int n_digit = 0, n_symbol = 0, n_upper = 0, n_len;
  // calculates % of numbers, uppercase letters an symbols
  for (int i=0; i<pwdlen; i++) {
    if (isdigit(pwd[i])) n_digit++;
    else if (isupper(pwd[i])) n_upper++;
    else if (ispunct(pwd[i])) n_symbol++;
  }
  n_digit = 100*n_digit/pwdlen;
  n_upper = 100*n_upper/pwdlen;
  n_symbol = 100*n_symbol/pwdlen;
  n_len = min(pwdlen*10-10,100);
  if (n_digit <= 30) n_digit = 10*n_digit/3; else n_digit = (1000-10*n_digit)/7;
  if (n_upper <= 40) n_upper = 10*n_upper/4; else n_upper = (1000-10*n_upper)/6;
  if (n_symbol <= 30) n_symbol = 10*n_symbol/3; else 
    if (n_symbol <= 70) n_symbol = 100; else n_symbol = (1000-10*n_symbol )/3;
  pwd_complexity = (n_len*50+n_digit*10+n_upper*15+n_symbol*25)/100;
  //pwd_complexity = max(n_len, max(n_digit, max(n_upper, n_symbol)));
  return(0);
}

int crypt() {
  int ret = 0;
  ask_pwd(1);
  calc_pwd_complexity();
#ifdef DEBUG
  printf("Password complexity: %i%%\n", pwd_complexity);
#endif
  if (pwd_complexity < 20) warn(WARN_PWDWEAK, NULL);
  
  Kryfile *archive = new Kryfile();
  archive->flag_showPercent = flag_percent;
  archive->flag_useMD5Hash = flag_md5;
  archive->Createfile(arc_filename);
  archive->Writeheader(fil_count);
  for (int i=0; i<fil_count; i++) {
    printf("Sto aggiungendo %s...\n", fil_filename[i]);
    archive->Cryptdata(fil_filename[i], pwd);
  }
  memset(pwd, 0, strlen(pwd));		// clear password from memory
  archive->Finalize();
  archive->Close();
  delete archive;
  
  return(ret);
}

int decrypt() {
  int ret = 0;
  ask_pwd(0);
  
  Kryfile *archive = new Kryfile();
  archive->flag_showPercent = flag_percent;
  archive->flag_useMD5Hash = flag_md5;
  archive->Openfile(arc_filename);
  archive->Readheader();

  int patcount;
  char **pats = NULL;
  char *dirout = NULL;

  expand_dirfile(fil_count, fil_filename, &dirout, &patcount, &pats);
  archive->Decryptdata(dirout, patcount, pats, pwd);

  memset(pwd, 0, strlen(pwd));		// clear password from memory
  free(dirout);
  for (int i=0; i<patcount; i++) free(pats[i]);
  free(pats);

  archive->Close();
  delete archive;
  return(ret);
}

int list() {
  int ret = 0;
  Kryfile *archive = new Kryfile();
  archive->Openfile(arc_filename);
  archive->Readheader();

  int patcount;
  char **pats = NULL;
  char *dirout = NULL;

  expand_dirfile(fil_count, fil_filename, &dirout, &patcount, &pats);
  archive->List(patcount, pats);

  free(dirout);
  for (int i=0; i<patcount; i++) free(pats[i]);
  free(pats);
  delete archive;
  return(ret);
}

int main(int argc, char *argv[]) {
#ifdef DEBUG
  printf("## DEBUG MODE ACTIVE ##\n");
#endif
  int ret = 0;
  atexit(exitfunc);		// sets exit function
  puts(txt_welcome);		// prints a short message
  parse_cmd(argc, argv);	// parses command line
#ifdef DEBUG
  printf("Archive name: %s\n", arc_filename);
  printf("Files/Dirs list: ");
  for (int i=0; i<fil_count;i++) printf("%s ", fil_filename[i]);
  printf("\nAction: ");
  switch (action) {
    case ACTION_CRYPT: printf("Crypt"); break;
    case ACTION_DECRYPT: printf("Decrypt"); break;
    case ACTION_LIST: printf("List"); break;
  }
  printf("\n");
#endif
  switch (action) {
    case ACTION_CRYPT: crypt(); break;
    case ACTION_DECRYPT: decrypt(); break;
    case ACTION_LIST: list(); break;
  }
  return(ret);
}
