/*
VGA font
*/

#ifndef VGAFONT_H
#define VGAFONT_H

#include <string.h>
#include <vga.h>

typedef struct {
  BYTE code; // corresponding ASCII code
  short width, height;
  BYTE *data;
} VGACHAR, *PVGACHAR;

class Font {
 public:
  char spacing;
  char align; //0=left, 1=center, 2=right

  Font(VirtualScreen *);
  ~Font();
  void setTextColor(BYTE col) { textColor = col; }
  void setBackColor(BYTE col) { backColor = col; }
  BYTE getTextColor() { return(textColor); }

  void addChar(VGACHAR);
  short getCharWidth(char);
  short getCharHeight(char);
  void print(VGACOORD, VGACOORD, char *);
  void printRect(VGACOORD, VGACOORD, VGACOORD, VGACOORD, char *);
  void readLine(VGACOORD, VGACOORD, short, char *);
 private:
  BYTE textColor, backColor;
  char isMonochrome;
  short spaceWidth;
  int n_chars;
  short table[256];
  PVGACHAR chars;
  VirtualScreen *screen;

  PERRPROC errFunc;
  short getStrnWidth(char *, int);
  short getCharIntCount(char *, int); // gets number of integer chars that fit into given space
  void printc(VGACOORD, VGACOORD, char);
  static int errorHandler(int, int);
};

Font::Font(VirtualScreen *scr) {
  spacing = 1;
  align = 0;
  isMonochrome = 1;
  textColor = 7;
  backColor = 0;
  spaceWidth = 0;
  n_chars = 0;
  chars = 0;
  errFunc = errorHandler;
  memset(table, 0xFF, sizeof(short)*256);
  screen = scr;
}

Font::~Font() {
  int i;
  if (chars != 0) {
    for (i=0; i<n_chars; i++) free((chars+i)->data);
    free(chars);
  }
}

short Font::getCharWidth(char ch) {
  if (table[ch] == -1) return(spaceWidth);
  return((chars+(table[(BYTE)ch] & 0xFF))->width);
}

short Font::getCharHeight(char ch) {
  if (table[ch] == -1) return(spaceWidth);
  return((chars+(table[(BYTE)ch] & 0xFF))->height);
}

void Font::addChar(VGACHAR ch) {
  if (table[ch.code] != -1) { // already exist
  // 2do
  } else {
    chars = (PVGACHAR)realloc(chars, sizeof(VGACHAR)*(n_chars+1));
    if (chars == 0) errFunc(5,1);
    (chars+n_chars)->code = ch.code;
    (chars+n_chars)->width = ch.width;
    (chars+n_chars)->height = ch.height;
    (chars+n_chars)->data = (BYTE *)malloc(ch.width*ch.height);
    if ((chars+n_chars)->data == 0) errFunc(5,1);
    memcpy((chars+n_chars)->data, ch.data, ch.width*ch.height);
    table[ch.code] = n_chars;
    n_chars++;
    if (spaceWidth == 0) spaceWidth = ch.width;
    if (ch.code == ' ') spaceWidth = ch.width;
  }
}

void Font::print(VGACOORD x, VGACOORD y, char *text) {
  int i, cpos = x;
  i = 0;
  while (text[i] != '\0') {
    printc(cpos, y, text[i]);
    cpos += getCharWidth(text[i])+spacing;
    i++;
  }
}

void Font::printRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, char *text) {
  int i, len, cpos = 0;
  int totx = x2-x1+1;
  char t;

  for (;;) {
    len = getStrnWidth(text+cpos,0x7FFE);
    if (len > totx) {
      char *space;
      len = getCharIntCount(text+cpos, totx);
      for (i=len-1; i>=0; i--) {
        if (text[cpos+i] == ' ') break;
      }
      if (i > 0) { // found a space
        len = i+1;
      }
      t = text[cpos+len];
      text[cpos+len] = '\0';
      switch (align) {
        case 0: print(x1, y1, text+cpos); break;
        case 1: print(x1+((totx-getStrnWidth(text+cpos,len))>>1), y1, text+cpos); break;
        case 2: print(x2-getStrnWidth(text+cpos,len), y1, text+cpos); break;
      }
      text[cpos+len] = t;
      cpos += len;
      y1 += chars->height;
      if (y1+chars->height > y2) return;
    } else {
      switch (align) {
        case 0: print(x1, y1, text+cpos); break;
        case 1: print(x1+((totx-getStrnWidth(text+cpos,len))>>1), y1, text+cpos); break;
        case 2: print(x2-getStrnWidth(text+cpos,len), y1, text+cpos); break;
      }
      break;
    }
  }
}

void Font::readLine(VGACOORD x1, VGACOORD y1, short maxlen, char *buffer) {
  char key, reading = 1;
  int curLen = 0, sizex=0;

  screen->show();
  do {
    key = getch();
    switch (key) {
      case 0x1B:
        curLen = 0;
      case 0x0D:
        reading = 0;
        break;
      case 0x08:
        curLen--;
        if (curLen < 0) curLen = 0; else {
          sizex -= getCharWidth(buffer[curLen])+spacing;
          screen->fRect(x1+sizex, y1, x1+sizex+getCharWidth(buffer[curLen])+spacing+getCharWidth('_'),y1+getCharHeight(buffer[curLen]), backColor);
        }
        printc(x1+sizex, y1, '_');
        break;
      default: {
        screen->fRect(x1+sizex, y1, x1+sizex+getCharWidth('_'),y1+getCharHeight('_'), backColor);
        printc(x1+sizex, y1, key);
        buffer[curLen++] = key;
        if (curLen > maxlen) curLen = maxlen; else {
          sizex += getCharWidth(key)+spacing;
          printc(x1+sizex, y1, '_');
        }
      }
    }
    screen->show();
  } while (reading);
  buffer[curLen] = '\0';
}

short Font::getStrnWidth(char *text, int len) {
  int ret = 0, i;

  if ((len == 0)||(text[0] == '\0')) return(0);
  for (i=0;;) {
    ret += getCharWidth(text[i++]);
    if ((i >= len)||(text[i] == '\0')) break;
    ret += spacing;
  }
  return(ret);
}

short Font::getCharIntCount(char *text, int interval) {
  int ret = 0, i;

  if (text[0] == '\0') return(0);
  for (i=0;;) {
    ret += getCharWidth(text[i++]);
    if (ret > interval) {
      i--;
      break;
    }
    if (text[i] == '\0') break;
    ret += spacing;
  }

  return(i);
}

void Font::printc(VGACOORD x, VGACOORD y, char ch) {
  int i,j;
  BYTE col;
  PVGACHAR pc;

  if (table[(BYTE)ch] == -1) return; // char does not exist
  pc = chars+(table[(BYTE)ch] & 0xFF);
  for (j=0; j<pc->height; j++) {
    for (i=0; i<pc->width; i++) {
      col = *(pc->data+j*pc->width+i);
      if (col != 0) {
        screen->putPixel(x+i,y+j, (isMonochrome)?textColor:col);
      }
    }
  }
}

int Font::errorHandler(int src, int code) {
  src;
  code;
  abort();
  return(0);
}

#endif