#include "polinomio.h"
#include "monomio.h"

Polinomio::SmartP::SmartP(Nodo *p = 0): ptr(p) {
  if (ptr) ptr->refs++;
}

Polinomio::SmartP::SmartP(const SmartP& sp): ptr(sp.ptr) {
  if (ptr) ptr->refs++;
}

Polinomio::SmartP::~SmartP() {
  if (ptr) {
    ptr->refs--;
    if (ptr->refs == 0) delete ptr;
  }
}

Polinomio::SmartP& Polinomio::SmartP::operator=(const SmartP& sp) {
  if (this != &sp) {
    Nodo *oldptr = ptr;
    ptr = sp.ptr;
    if (ptr) ptr->refs++;
    if (oldptr) {
      oldptr->refs--;
      if (oldptr->refs == 0) delete oldptr;
    }
  }
  return *this;
}

Polinomio::Nodo& Polinomio::SmartP::operator* () const {
  // must not be 0
  return const_cast<Nodo&>(*ptr);
}

Polinomio::Nodo* Polinomio::SmartP::operator-> () const {
  // must not be 0
  return const_cast<Nodo*>(ptr);
}

bool Polinomio::SmartP::operator== (const SmartP& sp) const {
  return (ptr == sp.ptr);
}

bool Polinomio::SmartP::operator!= (const SmartP& sp) const {
  return (ptr != sp.ptr);
}

// ---- CLASS NODO ------------------------------------------------------------

Polinomio::Nodo::Nodo(): refs(0) {
}

Polinomio::Nodo::Nodo(const Monomio& m, const SmartP& sp): 
  mon(m), next(sp), refs(0) { 
}


// ---- CLASS ITERATOR --------------------------------------------------------

bool Polinomio::Iterator::operator == (const Iterator& it) const {
  return ptr == it.ptr;
}

bool Polinomio::Iterator::operator != (const Iterator& it) const {
  return ptr != it.ptr;
}

Polinomio::Iterator& Polinomio::Iterator::operator ++() {
  if (ptr) ptr = &(*(ptr->next));
  return *this;
}

Polinomio::Iterator Polinomio::Iterator::operator ++(int) {
  Iterator o = *this;
  if (ptr) ptr = &(*(ptr->next));
  return o;
}

// ---- CLASS POLINOMIO ------------------------------------------------------- 

Polinomio::Polinomio(): head(0) {
}

Polinomio::Polinomio(const Monomio& m) {
  head = new Nodo(m, head);
}

Polinomio::Polinomio(int coeff, int esp = 1) {
  Monomio m(coeff, esp);
  head = new Nodo(m, head);
}

Monomio& Polinomio::operator [] (Iterator it) const {
  return (it.ptr)->mon;
}

int Polinomio::operator () (int x) const {
  SmartP p = head;
  int r = 0;
  while (p != 0) {
    r += (p->mon)(x);
    p = p->next;
  }
  return r;
}

bool Polinomio::is_null() const {
  return (head == 0);
}

Polinomio::Iterator Polinomio::begin() const {
  Iterator it;
  it.ptr = &(*head);
  return it;
}

Polinomio::Iterator Polinomio::end() const {
  Iterator it;
  it.ptr = 0;
  return it;
}

void Polinomio::debug_printlist() {
  SmartP p = head;
  int i = 0;
  while (p != 0) {
    std::cout << i << ")  " << p->mon << " (REFS=" << p->refs-1 << ")\n";
    p = p->next;
    i++;
  }
}

ostream& operator << (ostream& os, const Polinomio& p) {
  if (p.is_null()) return os << "0";
  Polinomio::Iterator it = p.begin();
  os << p[it++];
  for (; it!=p.end(); it++) {
    os << " + " << p[it];
  }
  return os;
}

Polinomio operator + (const Polinomio& p1, const Polinomio& p2) {
  Polinomio sum;
  Polinomio::Iterator i1 = p1.begin(), i2 = p2.begin();
  if (i1 == p1.end()) return p2;
  if (i2 == p2.end()) return p1;
  Monomio m1, m2;
  do {
    m1 = p1[i1];
    m2 = p2[i2];
    if (m1.esp > m2.esp) {
      sum.add_tail(m1);
      i1++;
    } else if (m1.esp == m2.esp) {
      Monomio mt(m1.coeff+m2.coeff, m1.esp);
      sum.add_tail(mt);
      i1++, i2++;
    } else {
      sum.add_tail(m2);
      i2++;
    }
  } while (i1 != p1.end() && i2 != p2.end());
  if (i1 == p1.end()) {
    for ( ; i2!=p2.end(); i2++) sum.add_tail(p2[i2]);
  } else {
    for ( ; i1!=p1.end(); i1++) sum.add_tail(p1[i1]);
  }
  return sum;
}

Polinomio operator - (const Polinomio& p1, const Polinomio& p2) {
  Polinomio sum;
  Polinomio::Iterator i1 = p1.begin(), i2 = p2.begin();
  if (i1 == p1.end()) return p2;
  if (i2 == p2.end()) return p1;
  Monomio m1, m2;
  do {
    m1 = p1[i1];
    m2 = p2[i2];
    if (m1.esp > m2.esp) {
      sum.add_tail(m1);
      i1++;
    } else if (m1.esp == m2.esp) {
      Monomio mt(m1.coeff-m2.coeff, m1.esp);
      sum.add_tail(mt);
      i1++, i2++;
    } else {
      sum.add_tail(m2);
      i2++;
    }
  } while (i1 != p1.end() && i2 != p2.end());
  if (i1 == p1.end()) {
    for ( ; i2!=p2.end(); i2++) sum.add_tail(-p2[i2]);
  } else {
    for ( ; i1!=p1.end(); i1++) sum.add_tail(p1[i1]);
  }
  return sum;
}

Polinomio operator * (int a, const Polinomio& p) {
  Polinomio prod;
  Polinomio::Iterator i = p.begin();
  for (; i!=p.end(); i++) {
    Monomio m(p[i].coeff*a, p[i].esp);
    prod.add_tail(m);
  }
  return prod;
}

Monomio Polinomio::del_head() {
  Monomio ret = head->mon;
  head = head->next;
  return ret;
}

Monomio Polinomio::del_monomio(SmartP p) {
}

void Polinomio::add_head(const Monomio& m) {
  if (m == 0) return;
  head = new Nodo(m, head);
}

void Polinomio::add_tail(const Monomio& m) {
  if (m == 0) return;
  if (is_null()) add_head(m);
  else {
    SmartP nm = new Nodo(m, 0);
    SmartP t = head;
    while (t != 0) {
      if (t->next == 0) break;
      t = t->next;
    }
    t->next = nm;
  }
}

void Polinomio::reduct() {
  SmartP s = head;
  while (s != 0) {
    if (s->mon == 0) {
    }
  }
}
