/* Matrix3D-def.h */

#include <math.h>


// ---- COSTRUTTORI ----------------------------------------------------------

// costruttore vuoto o con lista
template <class Type>
Matrix3D<Type>::Matrix3D(const Type* list) {
  if (list) {
    memcpy(data, list, sizeof(Type)*9);
  } else {
    for (int i=0; i<9; i++) data[i] = 0;
  }
}

// costruttore con 3 punti (le colonne)
template <class Type>
Matrix3D<Type>::Matrix3D(const Point3D<Type>& p1, const Point3D<Type>& p2, const Point3D<Type>& p3) {
  data[0] = p1.get_x();  data[1] = p2.get_x();  data[2] = p3.get_x();
  data[3] = p1.get_y();  data[4] = p2.get_y();  data[5] = p3.get_y();
  data[6] = p1.get_z();  data[7] = p2.get_z();  data[8] = p3.get_z();
}

// costruttore con i 9 elementi
template <class Type>
Matrix3D<Type>::Matrix3D(Type v1, Type v2, Type v3, Type v4, Type v5, Type v6, Type v7, Type v8, Type v9) {
  data[0] = v1; data[1] = v2; data[2] = v3;
  data[3] = v4; data[4] = v5; data[5] = v6;
  data[6] = v7; data[7] = v8; data[8] = v9;
}

// costruttore di copia, anche con tipo diverso
template <class Type>
template <class Type2>
Matrix3D<Type>::Matrix3D(const Matrix3D<Type2>& m) {
  for (int i=0; i<9; i++) data[i] = m.get(i/3, i%3);
}


// ---- OPERATORI ------------------------------------------------------------

// == : uguale
template <class Type>
bool Matrix3D<Type>::operator == (const Matrix3D<Type> &m) const {
  return memcmp(data, m.data, sizeof(Type)*9) == 0;
}

// != : diverso
template <class Type>
bool Matrix3D<Type>::operator != (const Matrix3D<Type> &m) const {
  return memcmp(data, m.data, sizeof(Type)*9) != 0;
}

// + : somma
template <class Type>
Matrix3D<Type> Matrix3D<Type>::operator +(const Matrix3D<Type> &m) const {
  Matrix3D<Type> r(data);
  for (int i=0; i<9; i++) r.data[i] += m.data[i];
  return r;
}

// - : differenza
template <class Type>
Matrix3D<Type> Matrix3D<Type>::operator -(const Matrix3D<Type> &m) const {
  Matrix3D<Type> r(data);
  for (int i=0; i<9; i++) r.data[i] -= m.data[i];
  return r;
}

// * : prodotto
template <class Type>
Matrix3D<Type> Matrix3D<Type>::operator *(const Matrix3D<Type> &m) const {
  Matrix3D<Type> r;
  for (int i=0; i<3; i++) {
    for (int j=0; j<3; j++) {
      for (int k=0; k<3; k++) {
        r.data[i*3+j] += get(i,k)*m.get(k,j);
      }
    }
  }
  return r;
}

// * : prodotto con punto
template <class Type>
Point3D<Type> Matrix3D<Type>::operator *(const Point3D<Type> &p) const {
  return Point3D<Type>( 
    data[0]*p.get_x() + data[1]*p.get_y() + data[2]*p.get_z(),
    data[3]*p.get_x() + data[4]*p.get_y() + data[5]*p.get_z(),
    data[6]*p.get_x() + data[7]*p.get_y() + data[8]*p.get_z()
  );
}

// assegnazione, anche con tipo diverso
template <class Type>
template <class Type2>
Matrix3D<Type>& Matrix3D<Type>::operator =(const Matrix3D<Type2> &m) {
  for (int i=0; i<9; i++) data[i] = m.get(i/3, i%3);
  return *this;
}

// output su ostream
template <class Type>
std::ostream & operator << (std::ostream& os, const Matrix3D<Type> &m) {
  char delim[3] = {'/', '|', '\\'};
  for (int i=0; i<3; i++) {
    os << delim[i] << " ";
    for (int j=0; j<3; j++) {
      os << m.get(i,j) << " ";
    }
    os << delim[2-i] << "\n";
  }
  return os;
}


// ---- METODI PUBBLICI ------------------------------------------------------

// legge l'elemento (i,j)
template <class Type>
Type Matrix3D<Type>::get(int i, int j) const throw(std::invalid_argument) {
  if (i<0 || j<0 || i>2 || j>2)
    throw std::invalid_argument("Class Matrix3D: Indici fuori dalla matrice.");
  return data[i*3+j];
}

// calcola il determinante
template <class Type>
Type Matrix3D<Type>::det() const {
  return ( data[0]*data[4]*data[8] + data[1]*data[5]*data[6] + data[2]*data[3]*data[7] -
           data[2]*data[4]*data[6] - data[1]*data[3]*data[8] - data[0]*data[5]*data[7] );
}

// traspone la matrice
template <class Type>
Matrix3D<Type>& Matrix3D<Type>::transpose() {
  for (int i=0; i<3; i++) {
    for (int j=i+1; j<3; j++) {
      Type t = get(j,i);
      set(j,i, get(i,j));
      set(i,j, t);
    }
  }
  return *this;
}

// esegue il complemento algebrico
template <class Type>
Matrix3D<Type>& Matrix3D<Type>::complement() {
  Type d[9];
  d[0] = data[4]*data[8]-data[5]*data[7];
  d[1] = data[5]*data[6]-data[3]*data[8];
  d[2] = data[3]*data[7]-data[4]*data[6];
  d[3] = data[2]*data[7]-data[1]*data[8];
  d[4] = data[0]*data[8]-data[2]*data[6];
  d[5] = data[1]*data[6]-data[0]*data[7];
  d[6] = data[1]*data[5]-data[2]*data[4];
  d[7] = data[2]*data[3]-data[0]*data[5];
  d[8] = data[0]*data[4]-data[1]*data[3];
  memcpy(data, d, sizeof(Type)*9);
  return *this;
}

// inverte la matrice (se possibile)
template <class Type>
Matrix3D<Type>& Matrix3D<Type>::invert() throw(std::domain_error) {
  Type d = det();
  if (d == 0) throw std::domain_error("Class Matrix3D: Matrice non invertibile.");
  complement();
  for (int i=0; i<9; i++) data[i] /= d;
  return *this;
}

// calcola la matrice di rotazione
template <class Type>
Matrix3D<Type> Matrix3D<Type>::rotation_matrix(double alpha, double beta, double gamma) {
  double s[3],c[3];
  #ifdef _GNU_SOURCE
    sincos(alpha, s+0, c+0);
    sincos(beta, s+1, c+1);
    sincos(gamma, s+2, c+2);
  #else
    s[0] = sin(alpha), c[0] = cos(alpha);
    s[1] = sin(beta) , c[1] = cos(beta);
    s[2] = sin(gamma), c[2] = cos(gamma);
  #endif
  Matrix3D<Type> ret;
  ret.data[0] = static_cast<Type>(c[1]*c[2]);
  ret.data[1] = static_cast<Type>(-(s[0]*s[1]*c[2]+c[0]*s[2]));
  ret.data[2] = static_cast<Type>(s[0]*s[2]-c[0]*s[1]*c[2]);
  ret.data[3] = static_cast<Type>(c[1]*s[2]);
  ret.data[4] = static_cast<Type>(c[0]*c[2]-s[0]*s[1]*s[2]);
  ret.data[5] = static_cast<Type>(-(c[0]*s[1]*s[2]+s[0]*c[2]));
  ret.data[6] = static_cast<Type>(s[1]);
  ret.data[7] = static_cast<Type>(s[0]*c[1]);
  ret.data[8] = static_cast<Type>(c[0]*c[1]);
  return ret;
}


// ---- METODI PROTETTI ------------------------------------------------------

// imposta a v l'elemento in posizione (i,j)
template <class Type>
void Matrix3D<Type>::set(int i, int j, Type v) throw(std::invalid_argument) {
  if (i<0 || j<0 || i>2 || j>2)
    throw std::invalid_argument("Class Matrix3D: Indici fuori dalla matrice.");
  data[i*3+j] = v;
}

