/* Point3D.h */

#ifndef POINT3D_H
#define POINT3D_H

#include <string>
#include <iostream>
#include <stdexcept>

// dichiarazione incompleta
template <class Type> class Point3D;

// dichiarazioni incomplete per funzioni friend
template <class Type>
std::ostream &operator << (std::ostream& os, const Point3D<Type> &p);

template <class Type>
Point3D<Type> operator+ (Type v, const Point3D<Type> &p);
template <class Type>
Point3D<Type> operator* (Type v, const Point3D<Type> &p);


/* 
  Classe Point3D:
    Rappresenta un punto nello spazio euclideo.
    Può essere istanziata con float, double e long double oppure
    con tipi interi (int, long, ... ).
    Le operazioni con istanziazioni di tipi diversi sono limitate alla copia
    di oggetti e all'assegnazione. Se la conversione fa perdere precisione
    vengono visualizzati degli warnings.
*/
template <class Type>
class Point3D {
 
 public:
  
  // ---- COSTRUTTORI ----
  
  // vuoto
  Point3D();
  
  // inizializzazione con x,y,z
  Point3D(Type, Type, Type);
  
  // copia, anche con tipo diverso
  template <class Type2>
  Point3D(const Point3D<Type2> &p);
  
  // ---- DISTRUTTORE ----
  // ... quello standard
  

  // ---- OPERATORI ----
  
  // confronto
  bool operator == (const Point3D<Type> &p) const;
  bool operator != (const Point3D<Type> &p) const;
  
  // somma e prodotto, versioni per la commutatività
  friend Point3D<Type> operator+ <Type>(Type v, const Point3D<Type> &p);
  friend Point3D<Type> operator* <Type>(Type v, const Point3D<Type> &p);
 
  // + e - unari
  Point3D<Type> operator +() const;
  Point3D<Type> operator -() const;
  
  // operazioni con scalari
  Point3D<Type> operator +(Type v) const;
  Point3D<Type> operator -(Type v) const;
  Point3D<Type> operator *(Type v) const;
  Point3D<Type> operator /(Type v) const throw(std::domain_error);
  
  // somma e differenza tra punti
  Point3D<Type> operator +(const Point3D<Type> &p) const;
  Point3D<Type> operator -(const Point3D<Type> &p) const;
  
  // prodotto scalare
  Type operator *(const Point3D<Type> &p) const;
  
  // assegnazione, anche con tipo diverso
  template <class Type2>
  Point3D<Type>& operator =(const Point3D<Type2> &p);
  
  // assegnazioni con operatori (scalari)
  Point3D<Type>& operator +=(Type v);
  Point3D<Type>& operator -=(Type v);
  Point3D<Type>& operator *=(Type v);
  Point3D<Type>& operator /=(Type v) throw(std::domain_error);
  
  // assegnazioni con operatori (punti)
  Point3D<Type>& operator +=(const Point3D<Type> &p);
  Point3D<Type>& operator -=(const Point3D<Type> &p);
  
  // conversione a stringa
  operator std::string() const;
  
  // output su ostream
  friend std::ostream &operator << <Type>(std::ostream& os, const Point3D<Type> &p);
  
  
  // ---- METODI ----
  
  // accesso alle coordinate
  inline Type get_x() const { return x; }
  inline Type get_y() const { return y; }
  inline Type get_z() const { return z; }
  
  // norma (distanza dall'origine)
  Type norm() const;
  
  // normalizza il punto (divide per la norma)
  Point3D<Type>& normalize();
  
  // distanza da un altro punto
  Type dist(const Point3D<Type> &p) const;
  
  // distanza dal piano contenente p1,p2,p3
  Type plane_dist(const Point3D<Type> &p1, const Point3D<Type> &p2, const Point3D<Type> &p3) const;
  
  // punto medio con p
  Point3D<Type> midpoint(const Point3D<Type> &p) const;
  
  // prodotto vettore
  Point3D<Type> cross(const Point3D<Type> &p) const;
  
  // rotazione attorno all'origine
  Point3D<Type>& rotate(double, double, double);
  
  // rotazione attorno a c
  Point3D<Type>& rotate(double, double, double, const Point3D<Type> &c);
  
  // ---- METODI STATICI ----
  
  // area del triangolo p1,p2,p3
  static Type area3p(const Point3D<Type> &p1, const Point3D<Type> &p2, const Point3D<Type> &p3);
  
 
 private:
  
  // coordinate del punto
  Type x,y,z;
  
  // calcola il determinante di u,v,w
  static Type det(const Point3D<Type> &u, const Point3D<Type> &v, const Point3D<Type> &w);

};

// definizioni dei membri
#include "Point3D-def.h"

// qualche tipo comodo
typedef Point3D<int> Point3Di;
typedef Point3D<double> Point3Dd;

// l'origine
const Point3Di P3Di_ZERO(0, 0, 0);
const Point3Dd P3Dd_ZERO(0.0, 0.0, 0.0);

#endif
