/* Polyhedron.cpp */

#include "Polyhedron.h"


// ---- COSTRUTTORI ----------------------------------------------------------

// costruttore vuoto
Polyhedron::Polyhedron(const std::string& n):
 Solid(n), vertices(0), verts(NULL) {

}

// costruttore con lista di vertici
Polyhedron::Polyhedron(int numverts, const Point3Dd *listv, const std::string& n):
 Solid(n), vertices(numverts) {
  if (vertices > 0) {
    verts = (Point3Dd *)malloc(vertices*sizeof(Point3Dd));
    for (int i=0; i<vertices; i++) verts[i] = Point3Dd(listv[i]);
  } else {
    verts = NULL;
  }
}

// costruttore di copia
Polyhedron::Polyhedron(const Polyhedron& p):
 Solid(p.name, p.color), vertices(p.vertices) {
  if (vertices > 0) {
    verts = (Point3Dd *)malloc(vertices*sizeof(Point3Dd));
    memcpy(verts, p.verts, vertices*sizeof(Point3Dd));
  } else {
    verts = NULL;
  }
}


// ---- DISTRUTTORE ----------------------------------------------------------

Polyhedron::~Polyhedron() {
  if (verts) free(verts);
}


// ---- OPERATORI ------------------------------------------------------------

// = : assegnazione profonda
Polyhedron& Polyhedron::operator = (const Polyhedron &p) {
  if (&p == this) return *this;
  if (verts) free(verts);
  name = p.name;
  color = p.color;
  vertices = p.vertices;
  if (vertices > 0) {
    verts = (Point3Dd *)malloc(vertices*sizeof(Point3Dd));
    memcpy(verts, p.verts, vertices*sizeof(Point3Dd));
  } else {
    verts = NULL;
  }
  return *this;
}

// [] : indicizzazione
const Point3Dd& Polyhedron::operator [] (Polyhedron::const_iterator it) const {
  if (it.i < 0 || it.i >= vertices) ; // throw exception
  return verts[it.i];
}

// << : output su stream
std::ostream& operator << (std::ostream &os, const Polyhedron& ph) {
  os << "Polyhedron \"" << ph.name << "\":\n";
  for (int i=0; i<ph.vertices; i++)
    os << ph.verts[i] << ", ";
  return os << "\n";
}


// ---- METODI PUBBLICI ------------------------------------------------------

// controlla se il poliedro è "piatto" (sta su un piano)
bool Polyhedron::is_flat() const {
  if (vertices <= 3) return true;
  Point3Dd v0 = verts[1]-verts[0];
  Point3Dd n0 = v0.cross(verts[2]-verts[1]);
  for (int i=3; i<vertices; i++) {
    Point3Dd n = v0.cross(verts[i]-verts[0]);
    if (n0.cross(n) != Point3Dd(0,0,0)) return false;
  }
  return true;
}

// ritorna un iteratore costante per il primo vertice
Polyhedron::const_iterator Polyhedron::begin() const {
  const_iterator ci;
  ci.i = 0;
  return ci;
}

// ritorna un iteratore costante per l'ultimo vertice+1
Polyhedron::const_iterator Polyhedron::end() const {
  const_iterator ci;
  ci.i = vertices;
  return ci;
}

// restituisce il valore minimo e massimo della componente z
void Polyhedron::get_minmax_z(double *minz, double *maxz) const {
  if (vertices == 0) {
    *minz = *maxz = 0;
    return;
  }
  *minz = *maxz = verts[0].get_z();
  for (int i=1; i<vertices; i++) {
    if (verts[i].get_z() > *maxz) *maxz = verts[i].get_z();
    if (verts[i].get_z() < *minz) *minz = verts[i].get_z();
  }
}

// calcola il baricentro del poliedro
Point3Dd Polyhedron::get_center() const {
  Point3Dd ret(0,0,0);
  for (int i=0; i<vertices; i++) ret += verts[i];
  return (ret/vertices);
}

// trasla il poliedro di v
void Polyhedron::move(const Point3Dd& v) {
  for (int i=0; i<vertices; i++) verts[i] += v;
}

// ruota il poliedro attorno al punto c
void Polyhedron::rotate(double alpha, double beta, double gamma, const Point3Dd& c) {
  if (c == P3Dd_ZERO) {
    for (int i=0; i<vertices; i++) verts[i].rotate(alpha, beta, gamma);
  } else {
    for (int i=0; i<vertices; i++) verts[i].rotate(alpha, beta, gamma, c);
  }
}

