/* Polyhedron.h */

#ifndef POLYHEDRON_H
#define POLYHEDRON_H

#include <string>
#include "Point3D.h"


/* 
  Classe Solid:
   classe astratta che rappresenta un solido. fornisce alcune funzionalità
   di base
*/
class Solid {
 public:
  // costruttore
  Solid(const std::string& n = "Solido", unsigned int col = 0): name(n), color(col) {};
  // distruttore
  virtual ~Solid() {};
  // metodi pubblici
  inline std::string get_name() const { return name; }
  inline void        set_name(const std::string& n) { name = n; }
  inline unsigned int get_color() const { return color; }
  inline void         set_color(unsigned int c) { color = c; }
  inline void *get_extradata() const { return extra; }
  inline void  set_extradata(void *e) { extra = e; }
  // metodi virtuali puri
  virtual double get_area() const = 0;
  virtual double get_volume() const = 0;
  virtual void move(const Point3Dd& v) = 0;
  virtual void rotate(double, double, double, const Point3Dd& c = P3Dd_ZERO) = 0;
 protected:
  // il nome del solido
  std::string name;
  // il colore del solido (rgb)
  unsigned int color;
  // informazioni aggiuntive
  void *extra;
};


/* 
  Classe Polyhedron:
   Rappresenta un poliedro
*/
class Polyhedron: public Solid {
 public:  
  
  
  // ---- COSTRUTTORI -----
  
  // vuoto o con nome
  Polyhedron(const std::string& name = "Poliedro");
  
  // inizializzazione con lista di vertici
  Polyhedron(int numverts, const Point3Dd *listv, const std::string& name = "Poliedro");
  
  // di copia
  Polyhedron(const Polyhedron& p);

  
  // ---- DISTRUTTORE ----
  
  virtual ~Polyhedron();
  
  // ---- CLASSI ITERATORI ----
  
  // iteratore read-only
  class const_iterator {
   public:
    // amicizie
    friend class Polyhedron;
    // operatori
    bool operator == (const const_iterator& it) const { return i == it.i; }
    bool operator != (const const_iterator& it) const { return i != it.i; }
    bool operator <  (const const_iterator& it) const { return i <  it.i; }
    bool operator <= (const const_iterator& it) const { return i <= it.i; }
    bool operator >  (const const_iterator& it) const { return i >  it.i; }
    bool operator >= (const const_iterator& it) const { return i >= it.i; }
    const_iterator &operator ++() { // prefisso
      i++; return *this;
    }
    const_iterator  operator ++(int) { // postfisso
      const_iterator ret = *this;
      i++; return ret;
    }
   private:
    int i;
  };
  

  // ---- OPERATORI ----
  
  // assegnazione profonda
  Polyhedron& operator = (const Polyhedron &p);
  
  // indicizzazione (per accedere ai vertici)
  const Point3Dd& operator [] (const_iterator) const;
  
  // output su ostream
  friend std::ostream& operator << (std::ostream &os, const Polyhedron& ph);
  
  
  // ---- METODI PUBBLICI ----
  
  // calcola il volume
  // ritorna -1 perché per un poliedro generico è troppo complicato calcolarlo 
  virtual double get_volume() const { return -1.0; }
  
  // calcola la superficie (-1 come sopra)
  virtual double get_area() const { return -1.0; }
  
  // ritorna vero se è convesso
  // anche qui ritorna sempre falso, essendo troppo complicato stabilirlo
  virtual bool is_convex() const { return false; }
  
  // ritorna vero se è piatto (sta su un piano)
  virtual bool is_flat() const;
  
  // ritornano un iteratore per l'inizio e la fine
  const_iterator begin() const;
  const_iterator end() const;
  
  // vero se non ci sono vertici
  inline bool is_empty() const { return vertices == 0; }
  
  // ritorna il numero di vertici
  inline int get_vertices() const { return vertices; }
  
  // calcola il baricentro del poliedro
  Point3Dd get_center() const;
  
  // calcola il valore minimo e massimo della componente z
  void get_minmax_z(double*, double*) const;
  
  // trasla il polidero del vettore v
  void move(const Point3Dd& v);
  
  // ruota il poliedro attorno a c
  void rotate(double, double, double, const Point3Dd& c = P3Dd_ZERO);
  
 protected:
  
  // numero di vertici
  int vertices;
  
  // vettore dinamico di vertici
  Point3Dd *verts;

};


#endif
