/* QDrawSolid.cpp */

#include "QDrawSolid.h"

// macro min/max
#ifndef MIN
  #define MIN(a,b)  (((a)<(b))?(a):(b))
#endif
#ifndef MAX
  #define MAX(a,b)  (((a)>(b))?(a):(b))
#endif

// ---- COSTRUTTORI ----------------------------------------------------------

QDrawSolid::QDrawSolid(QWidget *parent):
 QWidget(parent), perspective(true), paused(false), mouse_scaling(false), 
 scale(10.0), polycolor(QColor(0,0,0)), poly(NULL) {
  
  setBackgroundRole(QPalette::Window);
  setAutoFillBackground(true);
  setMinimumSize(40,40);
  setCursor(Qt::OpenHandCursor);

  timer.start(100, this);
}


// ---- DISTRUTTORE ----------------------------------------------------------

QDrawSolid::~QDrawSolid() {
  if (poly) delete poly;
}


// ---- METODI PUBBLICI ------------------------------------------------------
QSize QDrawSolid::sizeHint() const {
  return QSize(100,80);
}


// ---- SLOT PUBBLICI --------------------------------------------------------

// imposta il poliedro da visualizzare
void QDrawSolid::setPolyhedron(const Polyhedron& p) {
  if (poly) delete poly;
  poly = new Polyhedron(p);
  center = poly->get_center();
}

// imposta il colore
void QDrawSolid::setColor(const QColor& c) {
  polycolor = c;
}

// cancella il solido
void QDrawSolid::clear() {
  if (poly) delete poly;
  poly = NULL;
}


// ---- METODI PROTETTI ------------------------------------------------------

// disegna
void QDrawSolid::paintEvent(QPaintEvent *) {
  if (!poly) return;
  QPainter painter(this);
  QColor back = palette().color(QPalette::Window);
  painter.setBrush(Qt::SolidPattern);
  //painter.setRenderHint(QPainter::Antialiasing, true);
  
  int j = 0;
  Point3Dd verts[poly->get_vertices()];
  // sort on z-axis, to draw back points first
  for (Polyhedron::const_iterator i = poly->begin(); i<poly->end(); i++,j++) verts[j] = (*poly)[i];
  qsort(verts, poly->get_vertices(), sizeof(Point3Dd), point_zcomp);
  double minz = verts[0].get_z();
  double maxz = verts[poly->get_vertices()-1].get_z();
  for (j=0; j<poly->get_vertices(); j++) {
    int x,y, size;
    QColor color;
    double f = ((maxz-minz)<0.001)?1.0:((verts[j].get_z()-minz)/(maxz-minz));
    color.setRgb(static_cast<int>(back.red()*MIN(1.0-f, 0.8) + polycolor.red()*f),
                 static_cast<int>(back.green()*MIN(1.0-f, 0.8) + polycolor.green()*f), 
                 static_cast<int>(back.blue()*MIN(1.0-f, 0.8) + polycolor.blue()*f) );
    painter.setPen(color);
    painter.setBrush(color);
    to2D(verts[j], &x, &y);
    size = static_cast<int>(round(1.5 + f*4));
    painter.drawEllipse(x,y, size, size);
  }
}

// evento timer
void QDrawSolid::timerEvent(QTimerEvent *event) {
  if (event->timerId() == timer.timerId()) {
    if (!poly || paused) return;
    poly->rotate(0.1,0,0.2, center);
    update();
  } else {
    QWidget::timerEvent(event);
  }
}

// clic col mouse
void QDrawSolid::mousePressEvent(QMouseEvent *event) {
  paused = true;
  clickpos = event->pos();
  if (event->button() == Qt::RightButton) {
    mouse_scaling = true;
    setCursor(Qt::SizeVerCursor);
  } else {
    setCursor(Qt::ClosedHandCursor);
  }
}

// movimento del mouse
void QDrawSolid::mouseMoveEvent(QMouseEvent *event) {
  if (!poly) return;
  paused = true;
  double fact = 0.1;
  int sgnx = (clickpos.x()>event->x())?+1:((clickpos.x()<event->x())?-1:0);
  int sgny = (clickpos.y()>event->y())?+1:((clickpos.y()<event->y())?-1:0);
  if (mouse_scaling) {
    scale += 0.5*sgny;
    if (scale < 0) scale = 0.0;
    else if (scale > 50.0) scale = 50.0;
  } else {
    poly->rotate(-sgny*fact, sgnx*fact, 0, center);
  }
  update();
  clickpos = event->pos();
}

// rilascio pulsante del mouse
void QDrawSolid::mouseReleaseEvent(QMouseEvent *) {
  paused = false;
  mouse_scaling = false;
  setCursor(Qt::OpenHandCursor);
}


// ---- METODI PRIVATI -------------------------------------------------------

// proietta il punto p su un piano (in modo molto semplice)
void QDrawSolid::to2D(const Point3Dd &p, int *x, int *y) const {
  int cx = width()/2;
  int cy = height()/2;
  if (perspective) {
    int mex = 0, mey = 0, mez = 80, sizez = 200;
    int distz = abs(sizez-(mez+sizez/2));
    *x = cx + static_cast<int>((p.get_x()*scale-mex)*(p.get_z()*scale+distz+sizez/2)/sizez);
    *y = cy - static_cast<int>((p.get_y()*scale-mey)*(p.get_z()*scale+distz+sizez/2)/sizez);
  } else {
    *x = cx + static_cast<int>(p.get_x()*scale);
    *y = cy - static_cast<int>(p.get_y()*scale);
  }
}

int QDrawSolid::point_zcomp(const void *p1, const void *p2) {
  const Point3Dd *a = static_cast<const Point3Dd*>(p1);
  const Point3Dd *b = static_cast<const Point3Dd*>(p2);
  if (a->get_z() < b->get_z()) return -1;
  else if (a->get_z() > b->get_z()) return +1;
  else return 0;
}
