/*
  Schedulatore di turni per Raffaello
  
  Changelog:
  
  2.2: Reintrodotta verifica risolvibilità del problema
  2.1: Miglioramenti vari, aggiunta profondità di ricerca
  2.0: Versione corretta
  1.0: Versione preliminare, che faceva la cosa sbagliata

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define VERSION         2
#define VERSION_MINOR   3

#include "bitvector.h"


#define PROGNAME        "rapisched"

#ifdef __WINDOWS__
  #include <windows.h>
  #define ENDL          "\r\n"           // fine riga - windows
  // definisco random:
  #define srandom(a)    (srand(a))
  #define random(a)     (rand(a))
#else
  #define ENDL          "\n"             // fine riga - unix
#endif

/* Alcuni limiti */
#define MIN_WORKERS       1              // Minimo numero di lavoratori
#define MAX_WORKERS      50              // Massimo numero di lavoratori (deve star dentro un int64)
#define MIN_CONSEC        1              // Minimo numero di giorni di lavoro consecutivi
#define MAX_CONSEC       20              // Massimo numero di giorni di lavoro consecutivi
#define MIN_M             0              // Minimo numero di slot alla mattina
#define MAX_M           100              // Massimo numero di slot alla mattina
#define MIN_P             0              // Minimo numero di slot al pomeriggio
#define MAX_P           100              // Massimo numero di slot al pomeriggio
#define MIN_RELAX         0              
#define MAX_RELAX         7
#define MIN_TURNS         1              // Numero minimo di turni da generare
#define MAX_TURNS        50              // Numero massimo di turni da generare
#define MIN_REC_DEEP      1              // Minimo livello di profondità
#define MAX_REC_DEEP      5              // Massimo livello di profondità

/* Costanti degli slot */
#define SLOT_NULL       0x00             // Slot non inizializzato
#define SLOT_FREE       0x01             // Slot di riposo
#define SLOT_WORK       0x03             // Slot di lavoro
#define SLOT_M          0x05             // Slot di lavoro - mattina
#define SLOT_P          0x0b             // Slot di lavoro - pomeriggio
#define SLOT_SCORE_MIN   100             // Punteggio minimo 100-200

/* Alcuni valori di default */
#define DEF_N             13             // Numero di lavoratori
#define DEF_TURNS          4             // Numero di turni da generare
#define DEF_M_G            7             // Slot di lavoro mattina lun-ven
#define DEF_M_S            4             // Slot di lavoro mattina sabato
#define DEF_M_D            4             // Slot di lavoro mattina domenica
#define DEF_P_G            4             // Slot di lavoro pomeriggio lun-ven
#define DEF_P_S            2             // Slot di lavoro pomeriggio sabato
#define DEF_P_D            0             // Slot di lavoro pomeriggio domenica
#define DEF_MIN_RELAX      1             // Minimo numero di giorni di riposo a settimana
#define DEF_MAX_RELAX      3             // Massimo numero di giorni di riposo a settimana
#define DEF_MAX_CONSEC     5             // Massimo numero di giorni di lavoro consecutivi
#define DEF_MAX_M_CONSEC   3             // Massimo numero di mattine consecutive
#define DEF_MAX_P_CONSEC   3             // Massimo numero di pomeriggi consecutivi
#define DEF_MAX_PM_WEEK    2             // Massimo numero di P-M alla settimana
#define DEF_REC_DEEP       2             // Livello di profondità

/* Lavoratore */ 
typedef struct {
  /* Dati globali */
  int work_days;                // numero di giorni di lavoro fatti
  int relax_days;               // numero di giorni di riposo fatti
  int relax_consec;             // massimo numero di giorni di riposo consecutivi
  int relax_we_days;            // numero di gg di riposo durante il finesettimana
  int relax_split_days;         // numero di gg di riposo staccati
  int last_relax_day;           // ultimo giorno di riposo
  int mornings;                 // numero di turni mattutini
  int afternoons;               // numero di turni pomeridiani
  //int last_m;
  //int last_p;
  int pm_consec;                // numero di volte in cui un P è seguito da una M
} WEEK;

typedef struct {
  int slots[MAX_WORKERS][7];
  WEEK w[MAX_WORKERS];
  int score;
  char relax_distrib[8];        // Distribuzione dei riposi consecutivi
  char filled;                  // livello di completezza 0=vuoto, 1=con riposi, 2=completa
} TURN;

/* Dati del problema */
int N = DEF_N;                          // Numero di lavoratori totali
int M[7] = {DEF_M_G,DEF_M_G,DEF_M_G,DEF_M_G,DEF_M_G,DEF_M_S,DEF_M_D};
                                        // Ore di lavoro per mattina
int P[7] = {DEF_P_G,DEF_P_G,DEF_P_G,DEF_P_G,DEF_P_G,DEF_P_S,DEF_P_D};
                                        // Ore di lavoro per pomeriggio
int turns = DEF_TURNS;                  // Turni da generare
int min_relax = DEF_MIN_RELAX;          // Minimo numero di giorni di riposo a settimana
int max_relax = DEF_MAX_RELAX;          // Massimo numero di giorni di riposo a settimana
int max_consec = DEF_MAX_CONSEC;        // Massimo numero di giorni di lavoro consecutivi
int max_m_consec = DEF_MAX_M_CONSEC;
int max_p_consec = DEF_MAX_P_CONSEC;
int max_pm_week = DEF_MAX_PM_WEEK;
int max_recursive_deep = DEF_REC_DEEP;

/* Parametri vari */


/* Altre variabili */
TURN plan[MAX_TURNS];                   // Piano dei turni
int L[7];                               // Ore di lavoro (mattina + pomeriggio)
int best_relax;                         // Numero ideale di giorni di riposo
char *outfile = "-";                    // File per l'output
FILE *fout;
char html = 0;                          // Output in html



/* Funzioni */

/* Scambio */
inline void XCHG(int *a, int *b) {
  register int t = *a; *a = *b; *b = t;
}

/* Minimo */
inline int MIN(int a, int b) { return (a<=b)?a:b; }

/* Massimo */
inline int MAX(int a, int b) { return (a>=b)?a:b; }

/* Fattoriale */
long long int fact(int n) {
  register int i;
  long long int r = 1;
  for (i=2; i<=n; ++i) r *= i;
  return r;
}

/* Mul n...m */
long long int fact_part(int n, int m) {
  register int i;
  long long int r = 1;
  for (i=n; i<=m; ++i) r *= i;
  return r;
}

/* Coefficiente binomiale */
long long int binom(int a, int b) {
  if (b > a) XCHG(&a, &b);
  return fact_part(b+1, a)/fact(a-b);
}


// TODO: usare id globali
void warning(int wid) {
  fprintf(stderr, "ATTENZIONE: ");
  switch (wid) {
    case 0: fprintf(stderr, "Il numero di lavoratori non e' ottimale."); break;
    default: fprintf(stderr, "C'è qualcosa che non va, ma non so cosa."); break;
  }
  fprintf(stderr, ENDL);
}

// TODO: usare id globali
void error(int eid) {
  fprintf(stderr, "ERRORE: ");
  switch (eid) {
    case 0: fprintf(stderr, "Non ci sono sufficienti lavoratori."); break;
    case 1: fprintf(stderr, "Ci sono troppi lavoratori."); break;
    case 2: fprintf(stderr, "Impossibile generare i turni."); break;
    case 3: fprintf(stderr, "Opzione sconosciuta."); break;
    case 4: fprintf(stderr, "Opzione non valida."); break;
    case 5: fprintf(stderr, "Impossibile assegnare il turno."); break;
    case 6: fprintf(stderr, "Impossibile creare il file."); break;
    case 7: fprintf(stderr, "Impossibile aprire il file."); break;
    case 8: fprintf(stderr, "Formato del file non valido."); break;
    default: fprintf(stderr, "E' accaduto qualcosa di molto brutto."); break;
  }
  fprintf(stderr, ENDL);
  exit(1);
}

int find_best_T() {
  return 1;
}

void clear_turn(TURN *t) {
  int j;
  memset(t->slots, SLOT_NULL, sizeof(t->slots));
  // Svuoto i dati dei lavoratori
  for (j=0; j<N; ++j) {
    t->w[j].work_days = 0;
    t->w[j].relax_days = 0;
    t->w[j].relax_we_days = 0;
    t->w[j].relax_consec = 0;
    t->w[j].relax_split_days = 0;
    t->w[j].last_relax_day = -1;
    t->w[j].mornings = 0;
    t->w[j].afternoons = 0;
    t->w[j].pm_consec = 0;
    /*t->w[j].last_m = -1;
    t->w[j].last_p = -1;*/
  }
  t->score = 0;
  t->filled = 0;
  memset(t->relax_distrib, 0, sizeof(t->relax_distrib));
}

void init_vars() {
  int i;
  // Seme per i numeri pseudo casuali
  srandom(time(NULL)+clock());
  // Ore di lavoro totali al giorno
  for (i=0; i<7; ++i) L[i] = M[i]+P[i];
  // Svuoto il piano di lavoro
  for (i=0; i<turns; ++i) {
    clear_turn(plan+i);
  }
  // quantità ottimale di giorni di riposo settimanali
  best_relax = (min_relax + max_relax)/2;
  // apro i files
  if (strcmp(outfile, "-") == 0) fout = stdout; else {
    fout = fopen(outfile, "w");
    if (fout == NULL) error(6);
  }
}

void print_turn_text(TURN *t) {
  int i,j;
  char days[7] = {'L', 'M', 'M', 'G', 'V', 'S', 'D'};

  // Stampa intestazione
    fprintf(fout, "    ");
    for (i=0; i<7; ++i) {
      fprintf(fout, " %c ", days[i%7]);
      if ((i%7) == 6) fprintf(fout, " | ");
    }
    fprintf(fout, ENDL"----");
    for (i=0; i<1; ++i) {
      fprintf(fout, "------------------------");
    }
    fprintf(fout, ENDL);
    // Stampo turni
    for (j=0; j<N; ++j) {
      fprintf(fout, "%3i:", j+1);
      for (i=0; i<7; ++i) {
        char out;
        //char str[2] = {0,0};
        switch (t->slots[j][i]) {
          case SLOT_NULL: out = ' '; break;       // non inizializzata
          case SLOT_M   : out = 'M'; break;       // mattina
          case SLOT_P   : out = 'P'; break;       // pomeriggio
          case SLOT_FREE: out = '-'; break;       // riposo
          case SLOT_WORK: out = 'L'; break;       // lavoro (mattina o pomeriggio)
          //default: str[0] = plan[i][j]-SLOT_SCORE_MIN; out = atoi(str)+'0'; break;
          default:        out = '?'; break;
        }
        fprintf(fout, " %c ", out);
        if ((i%7) == 6) fprintf(fout, " | ");
      }
      fprintf(fout, ENDL);
    }
    //fprintf(fout, "%i\n", t->score);
}

void print_plan_text() {
  int w;
  for (w=0; w<turns; ++w) {
    print_turn_text(plan+w);
  }
}


void print_turn_html(TURN *t) {
  int i,j;
  char days[8] = {' ', 'L', 'M', 'M', 'G', 'V', 'S', 'D'};

  if (t->filled == 0) fprintf(fout, "Impossibile generare il turno<br>");
  else if (t->filled == 1) fprintf(fout, "Turno generato parzialmente<br>");
  
  fprintf(fout, "<table border=\"1\"><tr><td></td>");
    for (i=0; i<8; ++i) {
      fprintf(fout, "<td class=\"td_day\">%c</td>", days[i]);
    }
    fprintf(fout, "</tr>\n");

    // Stampo turni
    for (j=0; j<N; ++j) {
      fprintf(fout, "<tr><td class=\"td_worker\">%i</td>", j+1);
      for (i=0; i<7; ++i) {
        char out;
        char cls[16];
        switch (t->slots[j][i]) {
          case SLOT_NULL: out = ' '; strcpy(cls, "td_null"); break;       // non inizializzata
          case SLOT_M   : out = 'M'; strcpy(cls, "td_m"); break;       // mattina
          case SLOT_P   : out = 'P'; strcpy(cls, "td_p"); break;       // pomeriggio
          case SLOT_FREE: out = 'R'; strcpy(cls, "td_free"); break;       // riposo
          case SLOT_WORK: out = 'L'; strcpy(cls, "td_work"); break;       // lavoro (mattina o pomeriggio)
          default:        out = '?'; strcpy(cls, "td_other"); break;
        }
        if ((i%7) == 0) fprintf(fout, "<td>&nbsp;</td>");
        fprintf(fout, "<td class=\"%s\">&nbsp;%c&nbsp;</td>", cls, out);
      }
      fprintf(fout, "</tr>\n");
    }
    fprintf(fout, "</table>\n");
}

void print_stats_html(TURN *t) {
  int i,j;
  fprintf(fout, "<b>Statistiche</b><br>\n");
  //fprintf(fout, "<i>Punteggio:
  fprintf(fout, "<table cellspacing=0 cellpadding=0 border=0><tr><td align=\"center\" width=\"300\" colspan=\"12\">Distribuzione dei riposi consecutivi</td></tr>\n");
  for (i=2; i<=5; ++i) {
    fprintf(fout, "<tr><td width=\"100\">%i</td>", i);
    for (j=0; j<10; ++j) {
      char bgcol[10];
      if (t->relax_distrib[i] <= j) strcpy(bgcol, "#ffffff");
      else strcpy(bgcol, "#ff00ff");
      fprintf(fout, "<td bgcolor=\"%s\">&nbsp;</td>",bgcol);
    }
    fprintf(fout, "<td>(%i)</td>", t->relax_distrib[i]);
    fprintf(fout, "</tr>\n");
  }
  fprintf(fout, "</table>\n");
  // conta riposi totali:
  int totrel = 0;
  for (i=0; i<N; ++i) totrel += t->w[i].relax_days;
  fprintf(fout, "<p>Riposi totali: %i</p>\n", totrel);
}

void print_plan_html() {
  int w;
  
  // Stampa intestazione
  fprintf(fout, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n");
  fprintf(fout, "<html><head>\n");
  fprintf(fout, "<style type=\"text/css\">\
    body {color: black; background-color: white}\
    h3 {}\
    .td_title {text-align: center}\
    .td_turn {width:220}\
    .td_day {background-color: cyan}\
    .td_worker {background-color: blue; color: white; font-weight: bold}\
    .td_null {background-color: white} \
    .td_free {background-color: green}\
    .td_work {background-color: red}\
    .td_m {background-color: yellow}\
    .td_p {background-color: orange}\
    .td_other {background-color: purple}\
    </style>\n");
  fprintf(fout, "</head><body>\n");
  
  if (turns == 0) {
    fprintf(fout, "<b>Non &egrave; stato possibile generare i turni.</b><br><br>\n");
  } else {
    for (w=0; w<turns; ++w) {
      fprintf(fout, "<table><tr><td class=\"td_title\" align=\"center\" colspan=\"2\"><h3>Turno %i</h3></td></tr>\n", w+1);
      fprintf(fout, "<tr><td class=\"td_turn\" width=\"220\">\n");
      print_turn_html(plan+w);
      fprintf(fout, "</td><td>\n");
      print_stats_html(plan+w);
      fprintf(fout, "</td></tr></table>\n");
      fprintf(fout, "<hr><br>\n");
    }
    // legenda
    fprintf(fout, "<b>Legenda:</b>\n");
    fprintf(fout, "<table border=\"1\"><tr><td class=\"td_free\">R</td><td>Giorno di riposo</td></tr>\n");
    fprintf(fout, "<tr><td class=\"td_m\">M</td><td>Turno di mattina</td></tr>\n");
    fprintf(fout, "<tr><td class=\"td_p\">P</td><td>Turno di pomeriggio</td></tr>\n");
    fprintf(fout, "<tr><td class=\"td_work\">L</td><td>Turno lavorativo (M o P)</td></tr>\n");
    fprintf(fout, "<tr><td class=\"td_null\">&nbsp;</td><td>Turno non generato</td></tr>\n");
    fprintf(fout, "<tr><td class=\"td_other\">?</td><td>Errore di generazione</td></tr>\n");
    fprintf(fout, "</table>\n");
  }
  // chiudo
  fprintf(fout, "</body></html>\n");
}

void print_plan() {
  if (html) print_plan_html(); else print_plan_text();
}

/* Calcola quanti slot sono di riposo per un certo giorno */
inline int day_relax_slots(int day) { return (N - L[day%7]); }

/* Calcola quanti slot sono di lavoro per un certo giorno */
inline int day_work_slots(int day) { return (L[day%7]); }

/* Calcola quanti slot sono di riposo in turno */
int turn_relax_slots() {
  register int i, t = 0;
  for (i=0; i<7; ++i) t += day_relax_slots(i);
  return t;
}

/* Calcola quanti slot mattutini in un turno */
int turn_morning_slots() {
  register int i, t=0;
  for (i=0; i<7; ++i) t += M[i];
  return t;
}

/* Calcola quanti slot pomeridiani in un turno */
int turn_afternoon_slots() {
  register int i, t=0;
  for (i=0; i<7; ++i) t += P[i];
  return t;
}

/* Suggerisce nuove combinazioni di ore lavorative per ottimizzare i turni */
void find_best_L() {
  const int min_work_hours[3] = {4,2,1};
  const int max_work_hours[3] = {16,12,10};
  int best_work = (7-best_relax)*N;
  int starta = (L[0]+L[1]+L[2]+L[3]+L[4])/5;
  int a,b,c, da,db,dc;
  int nsol = 0;

  for (da=1,a = starta; a>=min_work_hours[0] && a<=max_work_hours[0]; a = starta+da) {
    for (db=1,b = L[5]; b>=min_work_hours[1] && b<=max_work_hours[1]; b = L[5]+db) {
      for (dc=1,c = L[6]; c>=min_work_hours[2] && c<=max_work_hours[2]; c = L[6]+dc) {
        int workslots = (5*a+b+c);
        if (workslots == best_work) {
          if (nsol == 0) fprintf(stderr, "Suggerimenti per le ore di lavoro:"ENDL" L  M  M  G  V  S  D"ENDL);
          fprintf(stderr, "%2i %2i %2i %2i %2i %2i %2i"ENDL, a,a,a,a,a,b,c);
          ++nsol;
          if (nsol > 10) return;
        }
        if (dc < 0) --dc;
        dc *= -1;
      }
      if (db < 0) --db;
      db *= -1;
    }
    if (da < 0) --da;
    da *= -1;
  }
}

/* Esegue un controllo preliminare sulla risolvibilità del problema */
void check_vars() {
  int workslots = (7*N) - turn_relax_slots();
  int min_work = (7-max_relax)*N;
  int best_work = (7-best_relax)*N;
  int max_work = (7-min_relax)*N;
  if (max_work < workslots) error(0);
  if (min_work > workslots) error(1);
  if (workslots != best_work) {
    warning(0);
    find_best_L();
  }
}

/* Ricostruisce le statistiche del turno t fino al giorno day */
void rebuild_workers_data(TURN *t, int day) {
  /*int w,d, maxrelax, relaxcons;
  for (w=0; w<N; ++w) {
    workers[w].work_days = 0;
    workers[w].relax_days = 0;
    workers[w].relax_consec = 0;
    workers[w].relax_we_days = 0;
    workers[w].relax_split_days = 0;
    workers[w].last_relax_day = -1;
    workers[w].pm_consec = 0;
    workers[w].mornings = 0;
    workers[w].afternoons = 0;
    maxrelax = 0;
    relaxcons = 0;
    for (d=0; d<=day; ++d) {
      if ((d%7) == 0) { // begin a new week
        workers[w].week_relax_days = 0;
      }
      if (plan[week][w][d] == SLOT_NULL) continue;
      if (plan[week][w][d] != SLOT_FREE) { // giorno di lavoro
        ++(workers[w].work_days);
        if (plan[week][w][d] == SLOT_M) {
          ++(workers[w].mornings);
          if (d > 0 && plan[week][w][d-1] == SLOT_P) ++(workers[w].pm_consec);
        }
        if (plan[week][w][d] == SLOT_P) ++(workers[w].afternoons);
        if (relaxcons > maxrelax) maxrelax = relaxcons;
        relaxcons = 0;
      } else { // giorno di riposo
        ++relaxcons;
        ++(workers[w].relax_days);
        ++(workers[w].week_relax_days);
        workers[w].last_relax_day = d;
        if ((d%7) >= 5) ++(workers[w].relax_we_days); // nel we
        int d1 = (d>0)?plan[week][w][d-1]:SLOT_NULL;
        int d2 = (d<day)?plan[week][w][d+1]:SLOT_NULL;
        if (d1 == SLOT_FREE || d2 == SLOT_FREE) { // riposo attaccato
        } else { // riposo staccato
          ++(workers[w].relax_split_days);
        }
      }
    }
    workers[w].relax_consec = maxrelax;
  }*/
}





/* Imposta i riposi per il tal giorno usando il bitset e aggiorna statistiche */
void set_day_bitset(TURN *t, int d, BITVEC *bits) {
  int w;
  for (w=0; w<N; ++w) {
    if (get_bit(bits, w) == 1) { // riposo
      t->slots[w][d] = SLOT_FREE;
      ++(t->w[w].relax_days);
      t->w[w].last_relax_day = d;
      if ((d%7) >= 5) ++(t->w[w].relax_we_days);
    } else { // lavoro
      t->slots[w][d] = SLOT_WORK;
      ++(t->w[w].work_days);
    }
  }
}

/* 0 = non accettabile, 100 = obbligatorio */
int calc_bitset_score(TURN *t, int d, BITVEC *bits) {
  int w, ret;
  TURN tw = *t;
  set_day_bitset(t, d, bits);
  ret = random()%100;
  for (w=0; w<N; ++w) {
    if (t->w[w].relax_days > max_relax) { ret = 0; break; } // troppi riposi
    int nextwork = 0;
    if (d == 6) { // domenica, calcolo lavorativi consecutivi la settimana dopo
      while (nextwork < 6 && t->slots[(w+1)%N][nextwork] == SLOT_WORK) ++nextwork;
    }
    if ( nextwork + (d - t->w[w].last_relax_day) > max_consec) { ret = 0; break; } // troppi lavori consecutivi
  }
  *t = tw;
  bits->score = ret;
  return ret;
}

/* Aggiunge un bitset */
int add_bitset(TURN *t, int d, int maxbs, BITVEC bits[], int curlen, BITVEC *bs) {
  int score = calc_bitset_score(t, d, bs);


  if (score == 0) return curlen; // non accettabile
  if (score == 100) { // obbligatoria
    memcpy(bits, bs, sizeof(BITVEC));
    return 1;
  }
  if (curlen < maxbs) { // aggiunge e basta
    memcpy(bits+curlen, bs, sizeof(BITVEC));
    if (curlen == maxbs-1) { // se l'ultimo allora ordina
      qsort(bits, maxbs, sizeof(BITVEC), cmp_bits_reverse);
    }
    return curlen+1;
  } else { // cerca dove inserire
    if (bits[curlen-1].score > score) return curlen;
    int ipos = curlen-2;
    while (ipos >= 0 && bits[ipos].score <= score) --ipos;
    if (ipos < curlen-2) { // in mezzo
      memmove(bits+(ipos+2), bits+(ipos+1), sizeof(BITVEC)*(maxbs-ipos-2));     
    }
    memcpy(bits+(ipos+1), bs, sizeof(BITVEC));
    return maxbs;
  }
  return curlen;
}

/* Calcola i possibili bitsets. Ritorna il numero di bitset generati */
int calc_bitsets(TURN *t, int d, int maxbs, BITVEC bits[]) {
  int i, nbits = 0;
  const int max_combin = 2000;
  if (maxbs == 0) return 0;
  
  if (d == 0) { // primo giorno, a caso
    set_bits(bits+0, 0);
    int l = day_relax_slots(d);
    do {
      int b = random()%N;
      if (get_bit(bits+0, b) == 0) {
        set_bit(bits+0, b, 1);
        --l;
      }
    } while (l > 0);
    return 1;
  } else {
    BITVEC db;
    BITVEC_REC bvr;
    bvr.n = N;
    bvr.m = day_relax_slots(d);
    bvr.list[N] = -1;
    long long combin = binom(N, day_relax_slots(d));
    if (combin <= max_combin) { // calcolo tutte le combinazioni
      for (i=0; i<combin; ++i) {
        if (gen_next_bitset(&db, &bvr) == 0) break;
        nbits = add_bitset(t, d, maxbs, bits, nbits, &db);
        if (db.score == 100) break; // obbligatoria
      }
    } else { // calcolo un po' di combinazioni
      //fprintf(stderr, "Salto delle combinazioni\n");
      long long counter = 0;
      for (i=0; i<max_combin; ++i) {
        int r;
        int skip = random()%(combin/(combin-counter) + 1);
        for ( ; skip > 0 && r != 0; --skip, ++counter) {
          r = gen_next_bitset(&db, &bvr);
        }
        if (r == 0) break;
        nbits = add_bitset(t, d, maxbs, bits, nbits, &db);
        if (db.score == 100) break; // obbligatoria
      }
    }
  }
  return nbits;
}

/* Confronta i turni sulla base del punteggio (decrescente) */
int cmp_turn_reverse(const void *w1, const void *w2) {
  int s1 = ((const TURN*)w1)->score;
  int s2 = ((const TURN*)w2)->score;
  if (s1 == s2) return 0;
  if (s1 < s2) return +1;
  return -1;
}

int calc_turn_score(TURN *t) {
  int i;
  int sc = 0;

  // calcolo distribuzione
  int lastwork;
  for (i = 7*N-1; i>=0; --i) if (t->slots[i/7][i%7] == SLOT_WORK) { lastwork = i; break; }
  int lim = lastwork;
  lastwork -= 7*N;
  for (i=0; i<lim; ++i) {
    int sl = t->slots[i/7][i%7];
    if (sl == SLOT_WORK) {
      ++(t->relax_distrib[i-lastwork-1]);
      lastwork = i;
    }
  }
  t->relax_distrib[0] = 0;
  if (t->relax_distrib[4] > 0) sc += 40; // riposo da 4
  sc += 20*MIN(t->relax_distrib[3], 2); // riposo da 3
  sc += 2*MIN(t->relax_distrib[2], 10); // riposo da 2
  sc += random()%MAX(MIN(10, 100-sc), 1);
  // calcolo giorni di riposo dopo 5 lavori consecutivi:
  /*int lastworks = 0;
  int lastrelax = 0;
  int inrelax = 0;
  for (i=0; i<7*(N+1); ++i) {
    if (t->slots[(i/7)%N][i%7] == SLOT_WORK) {
      if (inrelax) {
        if (lastworks >= max_consec && lastrelax < 2) sc -= 10;
        if (lastworks >= max_consec && lastrelax >= 2) sc += 20;
        lastworks = 0;
        inrelax = 0;
        lastrelax = 0;
      }
      lastworks++;
    } else if (t->slots[(i/7)%N][i%7] == SLOT_FREE) {
      inrelax = 1;
      ++lastrelax;
    }
  }
  */
  // calcolo riposo più lungo
  for (i=6; i>=0; --i) if (t->relax_distrib[i] != 0) break;
  if (i < 3) sc = 0; // deve esserci un riposo da almeno 3 gg consecutivi
  t->score = sc; 
  return sc;
}

void generate_rec(TURN *t, int day) {
  int i;
  static int nturn = 0;
  
  if (day == 7) {
    // TODO: unificare con add_bits
    //print_turn_text(t);
    t->filled = 1;
    if (nturn < turns) { // aggiungi in coda
      plan[nturn++] = *t;
      if (nturn == turns) { // se l'ultimo allora ordina
        qsort(plan, turns, sizeof(TURN), cmp_turn_reverse);
      }
    } else {
      int sc = calc_turn_score(t);
      if (sc > 0 && plan[turns-1].score <= sc) {
        int ipos = turns-2;
        while (ipos >= 0 && plan[ipos].score <= sc) --ipos;
        if (ipos < turns-2) { // in mezzo
          memmove(plan+(ipos+2), plan+(ipos+1), sizeof(TURN)*(turns-ipos-2));     
        }
        memcpy(plan+(ipos+1), t, sizeof(TURN));
      }
    }
    return;
  }
  BITVEC bs[max_recursive_deep];
  int r = calc_bitsets(t, day, max_recursive_deep, bs);
  for (i=0; i<r; ++i) {
    TURN tt = *t;
    set_day_bitset(t, day, bs+i);
    generate_rec(t, day+1);
    *t = tt;
  }
}

void generate() {
  TURN t;
  clear_turn(&t);
  generate_rec(&t, 0);
}

// ---------------------------------------------------------

/* Imposta i P/M per il tal giorno usando il bitset e aggiorna statistiche */
void set_day_bitset_mp(TURN *t, int d, BITVEC *bits) {
  int w, b;
  for (b=0,w=0; w<N; ++w) {
    if (t->slots[w][d] == SLOT_WORK) { // deve essere di lavoro
      if (get_bit(bits, b) == 1) { // pomeriggio
        t->slots[w][d] = SLOT_P;
        ++(t->w[w].afternoons);
        //t->w[w].last_p = d;
      } else { // mattina
        t->slots[w][d] = SLOT_M;
        ++(t->w[w].mornings);
        if (d > 0 && t->slots[w][d-1] == SLOT_P) ++(t->w[w].pm_consec);
        //t->w[w].last_m = d;
      }
      ++b;
    }
  }
}

int calc_turn_score_mp(TURN *t) {
  return random()%100;
}

int calc_bitset_score_mp(TURN *t, int day, BITVEC *bs) {
  int w, ret, i;
  TURN tw = *t;
  set_day_bitset_mp(t, day, bs);
  ret = random()%100;
  for (w=0; w<N; ++w) {
    if (t->w[w].pm_consec > max_pm_week) { ret = 0; break; }

    int nextm = 0, nextp = 0;
    int prevm = 0, prevp = 0;
    if (day == 6) { // domenica, calcolo p/m consecutivi la settimana dopo
      while (nextm < 6 && t->slots[(w+1)%N][nextm] == SLOT_M) ++nextm;
      while (nextp < 6 && t->slots[(w+1)%N][nextp] == SLOT_P) ++nextp;
    }
    for (i=day; i>=0; --i) if (t->slots[w][i] == SLOT_M) ++prevm; else break;
    for (i=day; i>=0; --i) if (t->slots[w][i] == SLOT_P) ++prevp; else break;
    if ( nextm + prevm > max_m_consec) { ret = 0; break; } // troppe mattine consecutive
    if ( nextp + prevp > max_p_consec) { ret = 0; break; } // troppi pomeriggi consecutive
  }
  *t = tw;
  bs->score = ret;
  return ret;
}

/* Aggiunge un bitset */
int add_bitset_mp(TURN *t, int d, int maxbs, BITVEC bits[], int curlen, BITVEC *bs) {
  int score = calc_bitset_score_mp(t, d, bs);


  if (score == 0) return curlen; // non accettabile
  if (score == 100) { // obbligatoria
    memcpy(bits, bs, sizeof(BITVEC));
    return 1;
  }
  if (curlen < maxbs) { // aggiunge e basta
    memcpy(bits+curlen, bs, sizeof(BITVEC));
    if (curlen == maxbs-1) { // se l'ultimo allora ordina
      qsort(bits, maxbs, sizeof(BITVEC), cmp_bits_reverse);
    }
    return curlen+1;
  } else { // cerca dove inserire
    if (bits[curlen-1].score > score) return curlen;
    int ipos = curlen-2;
    while (ipos >= 0 && bits[ipos].score <= score) --ipos;
    if (ipos < curlen-2) { // in mezzo
      memmove(bits+(ipos+2), bits+(ipos+1), sizeof(BITVEC)*(maxbs-ipos-2));     
    }
    memcpy(bits+(ipos+1), bs, sizeof(BITVEC));
    return maxbs;
  }
  return curlen;
}

/* Calcola i possibili bitsets per M/P. Ritorna il numero di bitset generati */
/* TODO: portare in file .h esterno */
int calc_bitsets_mp(TURN *t, int d, int maxbs, BITVEC bits[]) {
  int i, nbits = 0;
  const int max_combin = 2000;
  if (maxbs == 0) return 0;
  
  if (d == 0) { // primo giorno, a caso
    set_bits(bits+0, 0);
    int l = P[d];
    do {
      int b = random()%day_work_slots(d);
      if (get_bit(bits+0, b) == 0) {
        set_bit(bits+0, b, 1);
        --l;
      }
    } while (l > 0);
    return 1;
  } else {
    BITVEC db;
    BITVEC_REC bvr;
    bvr.n = day_work_slots(d);
    bvr.m = P[d%7];
    bvr.list[day_work_slots(d)] = -1;
    long long combin = binom(day_work_slots(d), P[d%7]);
    if (combin <= max_combin) { // calcolo tutte le combinazioni
      for (i=0; i<combin; ++i) {
        if (gen_next_bitset(&db, &bvr) == 0) break;
        nbits = add_bitset_mp(t, d, maxbs, bits, nbits, &db);
        if (db.score == 100) break; // obbligatoria
      }
    } else { // calcolo un po' di combinazioni
      //fprintf(stderr, "Salto delle combinazioni\n");
      long long counter = 0;
      for (i=0; i<max_combin; ++i) {
        int r;
        int skip = random()%(combin/(combin-counter) + 1);
        for ( ; skip > 0 && r != 0; --skip, ++counter) {
          r = gen_next_bitset(&db, &bvr);
        }
        if (r == 0) break;
        nbits = add_bitset_mp(t, d, maxbs, bits, nbits, &db);
        if (db.score == 100) break; // obbligatoria
      }
    }
  }
  return nbits;
}


void set_mp_rec(TURN *t, int nturn, int day) {
  int i;
  
  if (day == 7) {
    int sc = calc_turn_score_mp(t);
    if (sc > 0 && sc > plan[nturn].score) {
      t->filled = 2;
      plan[nturn] = *t;
    }
    //print_turn_text(t);
    return;
  }
  BITVEC bs[max_recursive_deep];
  int r = calc_bitsets_mp(t, day, max_recursive_deep, bs);
  for (i=0; i<r; ++i) {
    TURN tt = *t;
    set_day_bitset_mp(t, day, bs+i);
    set_mp_rec(t, nturn, day+1);
    *t = tt;
  }
}

void set_mp() {
  int it,tt;
  
  for (it=0; it<turns; ++it) {
    if (plan[it].filled < 1) continue;
    TURN t = plan[it];
    for (tt=0; tt<max_recursive_deep && plan[it].filled < 2; ++tt) {
      set_mp_rec(&t, it, 0);
    }
  }
}

void show_version() {
  printf(ENDL PROGNAME" -- versione %i.%i"ENDL ENDL, VERSION, VERSION_MINOR);
}

void show_syntax() {
//        1........|.........|.........|........|.........|.........|.........|.........|81
  printf(ENDL"Sintassi: "PROGNAME" [opzioni] [files]"ENDL ENDL);
  printf("Opzioni:"ENDL);
  printf("  -n X          : Imposta il numero di lavoratori a X"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_WORKERS, MAX_WORKERS, DEF_N);
  printf("  -t X          : Il numero di turni da generare"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_TURNS, MAX_TURNS, DEF_TURNS);
  printf("  -c X          : Massimo numero di giorni lavorativi consecutivi"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_CONSEC, MAX_CONSEC, DEF_MAX_CONSEC);
  printf("  -r X          : Minimo numero di giorni di riposo settimanali"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_RELAX, MAX_RELAX, DEF_MIN_RELAX);
  printf("  -s X          : Massimo numero di giorni di riposo settimanali"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_RELAX, MAX_RELAX, DEF_MAX_RELAX);
  printf("  -m G,S,D      : Quantità di lavoratori al mattino. Lun-Ven=G, Sab=S, Dom=D"ENDL);
  printf("                    valori=%i-%i, default=%i,%i,%i"ENDL, MIN_M, MAX_M, DEF_M_G, DEF_M_S, DEF_M_D);
  printf("  -p G,S,D      : Come sopra, ma per i pomeriggi"ENDL);
  printf("                    valori=%i-%i, default=%i,%i,%i"ENDL, MIN_P, MAX_P, DEF_P_G, DEF_P_S, DEF_P_D);
  printf("  -x X          : Massimo numero di M consecutive"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_CONSEC, MAX_CONSEC, DEF_MAX_M_CONSEC);
  printf("  -y X          : Massimo numero di P consecutivi"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_CONSEC, MAX_CONSEC, DEF_MAX_P_CONSEC);
  printf("  -a X          : Massimo numero di P-M consecutivi alla settimana"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_CONSEC, MAX_CONSEC, DEF_MAX_PM_WEEK);
  printf("  -d X          : Massima profondità della ricerca"ENDL);
  printf("                    valori=%i-%i, default=%i"ENDL, MIN_REC_DEEP, MAX_REC_DEEP, DEF_REC_DEEP);
  printf("  -o file       : Scrive il risultato sul file (standard output se -)"ENDL);
  printf("  --html        : Output in formato html"ENDL);
  printf("  -h, --help    : Mostra questo messaggio"ENDL);
  printf("  -v, --version : Mostra il numero di versione"ENDL);
  printf("  --            : Fine delle opzioni"ENDL ENDL);
}

void read_cmdline(int argc, char *argv[]) {
  int i;
  char endopt = 0;
  
  for (i=1; i<argc; ++i) {
    if (!endopt && argv[i][0] == '-') { // inizia opzione
      switch (argv[i][1]) {
        case 'n': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_WORKERS || n > MAX_WORKERS) error(4);
          N = n;
          break;
        }
        case 't': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_TURNS || n > MAX_TURNS) error(4);
          turns = n;
          break;
        }
        case 'r': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_RELAX || n > MAX_RELAX) error(4);
          min_relax = n;
          break;
        }
        case 's': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_RELAX || n > MAX_RELAX) error(4);
          max_relax = n;
          break;
        }
        case 'c': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_CONSEC || n > MAX_CONSEC) error(4);
          max_consec = n;
          break;
        }        
        case 'a': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_CONSEC || n > MAX_CONSEC) error(4);
          max_pm_week = n;
          break;
        }        
        case 'x': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_CONSEC || n > MAX_CONSEC) error(4);
          max_m_consec = n;
          break;
        }  
        case 'y': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_CONSEC || n > MAX_CONSEC) error(4);
          max_p_consec = n;
          break;
        }  
        case 'd': {
          if (i == (argc-1)) error(3);
          int n = atoi(argv[++i]);
          if (n < MIN_REC_DEEP || n > MAX_REC_DEEP) error(4);
          max_recursive_deep = n;
          break;
        }          case 'm':
        case 'p': {
          if (i == (argc-1)) error(3);
          int g,s,d;
          g = s = d = MIN(MIN_M, MIN_P)-1;
          sscanf(argv[++i], "%u,%u,%u", &g, &s, &d);
          
          if (argv[i-1][1] == 'm') {
            if (g < MIN_M || g > MAX_M || s < MIN_M || s > MAX_M || d < MIN_M || d > MAX_M) error(4);
            M[0] = M[1] = M[2] = M[3] = M[4] = g;
            M[5] = s;
            M[6] = d;
          } else {
            if (g < MIN_P || g > MAX_P || s < MIN_P || s > MAX_P || d < MIN_P || d > MAX_P) error(4);
            P[0] = P[1] = P[2] = P[3] = P[4] = g;
            P[5] = s;
            P[6] = d;
          }
          break;
        }
        case 'o': {
          if (i == (argc-1)) error(3);
          outfile = strdup(argv[++i]);
          break;
        }
        case 'h': show_syntax(); exit(0); break;
        case 'v': show_version(); exit(0); break;
        case '-': { // opzione lunga
          if (argv[i][2] == 0) { // end of options
            endopt = 1;
          } else if (strcmp(argv[i]+2, "html") == 0) {
            html = 1;
          } else if (strcmp(argv[i]+2, "help") == 0) {
            show_syntax();
            exit(0);
          } else if (strcmp(argv[i]+2, "version") == 0) {
            show_version();
            exit(0);
          } else {
            error(3);
          }
          break;
        }
        default: error(3); break;
      }
    } else { // filename
      // TODO
    }
  }
  if (min_relax > max_relax) error(4);
}

/* Funzione main */

int main(int argc, char *argv[]) {

  read_cmdline(argc, argv);
 
  init_vars();
  check_vars();
  
  generate();
  set_mp();
  // controlla se ne vanno tolti
  int i;
  for (i=turns-1; i>=0; --i) {
    if (plan[i].filled == 0) --turns;
  }
  
  //rebuild_workers_data(7*T-1);
  print_plan();
  //print_workers();
  
  // TODO: mettere on-exit
  if (fout != NULL && fout != stdout) fclose(fout);
  
  return 0;
}
