#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define SUDOSIZE2	3
#define SUDOSIZE	(SUDOSIZE2*SUDOSIZE2)

int timeout = 10;

typedef char Sudotype;
typedef Sudotype Sudoku[SUDOSIZE][SUDOSIZE];

struct Sudotree {
  Sudotree *left, *right;
  Sudoku sudo;
};

int compare(Sudoku *a1, Sudoku *a2) {
  for (int i=0; i<SUDOSIZE*SUDOSIZE; i++) {
    Sudotype b1 = *((Sudotype *)(*a1)+i), 
        b2 = *((Sudotype*)(*a2)+i);
    if (b1 > b2) return(1);
    if (b1 < b2) return(-1);
  }
  return(0);
}

class SudoSet {
 public:
  SudoSet();
  ~SudoSet();
  int Add(Sudoku *sudo);
  int IsPresent(Sudoku *sudo);
  int GetCount() { return(count); }
 private:
  int count;
  Sudotree *root;
  Sudotree *Find(Sudotree *root, Sudoku *sudo);
  Sudotree *FindAdd(Sudotree *root, Sudoku *sudo);
};

SudoSet::SudoSet() {
  count = 0;
  root = NULL;
}

SudoSet::~SudoSet() {
  // free the tree ?
}

Sudotree *SudoSet::Find(Sudotree *root, Sudoku *sudo) {
  if (root == NULL) return(NULL);
  int r = compare(sudo, &(root->sudo));
  if (r == 0) return(root);
  if (r < 0) 
    return(Find(root->left, sudo));
  else
    return(Find(root->right, sudo));
}

Sudotree *SudoSet::FindAdd(Sudotree *root, Sudoku *sudo) {
  if (root == NULL) return(NULL);
  int r = compare(sudo, &(root->sudo));
  Sudotree *ret, **child;
  if (r == 0) return(root);
  child = (r<0) ? &(root->left) : &(root->right);
  ret = FindAdd(*child, sudo);
  if (ret == NULL) {
    *child = (Sudotree *)malloc(sizeof(Sudotree));
    (*child)->left = NULL;
    (*child)->right = NULL;
    memcpy((*child)->sudo, *sudo, sizeof(Sudoku));
    ret = *child;
  }
  return(ret);
}


int SudoSet::Add(Sudoku *sudo) {
  if (root == NULL) { // 1st element
    root = (Sudotree *)malloc(sizeof(Sudotree));
    root->left = NULL;
    root->right = NULL;
    memcpy(root->sudo, sudo, sizeof(Sudoku));
  } else {  
    FindAdd(root, sudo);
  }
  count++;
  return(0);
}

int SudoSet::IsPresent(Sudoku *sudo) {
  Sudotree *t = Find(root, sudo);
  return (t == NULL)?0:1;
}

/*----------------------------------------------------------------------------*/

class SudoSolver {
 public:
  SudoSolver(SudoSet *sset, time_t starttime, FILE *fout);
  ~SudoSolver();
  int FillSudoku(Sudoku *sudo);
  int PrintSudoku();
  int Solve();
 private:
  int IsComplete();
  int IsCorrect();
  // sets a number in the sudoku
  int SetBit(int x, int y, Sudotype n) { table[y][x] = n; }
  // sets a number and clears the possibility table
  int PutBit(int x, int y, Sudotype n,  Sudotype *tsudo[SUDOSIZE][SUDOSIZE], Sudoku *tsudolen);
  Sudoku table;
  SudoSet *sset;
  time_t time0;
  FILE *fout;
};

SudoSolver::SudoSolver(SudoSet *sset, time_t starttime, FILE *fout) {
  memset(table, 0, sizeof(Sudoku));
  this->sset = sset;
  this->fout = fout;
  time0 = starttime;
}

SudoSolver::~SudoSolver() {
}

int SudoSolver::FillSudoku(Sudoku *sudo) {
  memcpy(table, *sudo, sizeof(Sudoku));
}

int SudoSolver::PrintSudoku() {
  for (int j=0; j<SUDOSIZE; j++) {
    for (int i=0; i<SUDOSIZE; i++) {
      fprintf(fout, (table[j][i] != 0)?"%2i":"  ", (int)(table[j][i]));
      //fprintf(fout, "%3i", (int)(table[j][i]));
    }
    fprintf(fout, "\n");
  }
  fprintf(fout, "\n");
}

int SudoSolver::IsComplete() {
  for (int j=0; j<SUDOSIZE; j++) {
    for (int i=0; i<SUDOSIZE; i++) {
      if (table[j][i] == 0) return(0);
    }
  }
  return(1);
}

int SudoSolver::IsCorrect() {
  for (int j=0; j<SUDOSIZE; j++) {
    for (int i=0; i<SUDOSIZE; i++) {
      Sudotype n = table[j][i];
      if (n != 0) {
        // checks rows
        for (int u=i+1; u<SUDOSIZE; u++)
          if (table[j][u] == n) return(0);
        // checks columns
        for (int u=j+1; u<SUDOSIZE; u++)
          if (table[u][i] == n) return(0);
        // checks square
        for (int u2=0; u2<SUDOSIZE2; u2++)
          for (int u1=0; u1<SUDOSIZE2; u1++) {
            int py = SUDOSIZE2*(j/SUDOSIZE2)+u2,
                px = SUDOSIZE2*(i/SUDOSIZE2)+u1;
            if ((px == i) && (py == j)) continue;
            if (table[py][px] == n) return(0);
          }
      }
    }
  }
  return(1);
}

int SudoSolver::PutBit(int x, int y, Sudotype n,  Sudotype *tsudo[SUDOSIZE][SUDOSIZE], Sudoku *tsudolen) {

  // sets n
  table[y][x] = n;
  int oldlen = (*tsudolen)[y][x];
  (*tsudolen)[y][x]--;
  if (oldlen == 1) free(tsudo[y][x]);
  else {
    for (int v=0; v<oldlen; v++)
      if (*(tsudo[y][x]+v) == n) { // ok, free
        *(tsudo[y][x]+v) = *(tsudo[y][x]+(*tsudolen)[y][x]);
        break;
      }
  }

  // free row & column
  for (int u=0; u<SUDOSIZE; u++) {
    // free row
    for (int v=0; v<(*tsudolen)[y][u]; v++)
      if (*(tsudo[y][u]+v) == n) { // ok, free
        *(tsudo[y][u]+v) = *(tsudo[y][u]+--(*tsudolen)[y][u]);
        break;
      }
    // free column
    for (int v=0; v<(*tsudolen)[u][x]; v++)
      if (*(tsudo[u][x]+v) == n) { // ok, free
        *(tsudo[u][x]+v) = *(tsudo[u][x]+--(*tsudolen)[u][x]);
        break;
      }
  }    

  // free square
  for (int u2=0; u2<SUDOSIZE2; u2++)
    for (int u1=0; u1<SUDOSIZE2; u1++) {
      int py = SUDOSIZE2*(y/SUDOSIZE2)+u2,
          px = SUDOSIZE2*(x/SUDOSIZE2)+u1;
      for (int v=0; v<(*tsudolen)[py][px]; v++)
        if (*(tsudo[py][px]+v) == n) { // ok, free
          *(tsudo[py][px]+v) = *(tsudo[py][px]+--(*tsudolen)[py][px]);
          break;
        }
    }
}


int SudoSolver::Solve() {
  int i, j;
  Sudotype *tsudo[SUDOSIZE][SUDOSIZE];
  Sudoku tsudolen;
  
  // checks timeout
  if (time(NULL)-time0 > timeout) return(2);
  
  // checks if the table is correct
  if (!IsCorrect()) return(1);
  
  // checks if we have already calculate this
  if (sset->IsPresent(&table)) return(0);
  sset->Add(&table);
  
  // fills the possibility-table
  for (j=0; j<SUDOSIZE; j++) {
    for (i=0; i<SUDOSIZE; i++) {
      Sudotype k;
      tsudo[j][i] = NULL;
      tsudolen[j][i] = 0;
      // checks if cell is empty
      if (table[j][i] == 0) {
        // checks all numbers from 1 to SUDOSIZE
        for (k=1; k<=SUDOSIZE; k++) {
          int kok = 1;
          // horiz & vertic
          for (int u=0; u<SUDOSIZE; u++) {
            if (table[j][u] == k) { kok = 0; break; }
            if (table[u][i] == k) { kok = 0; break; }
          }
          if (!kok) continue;
          // same square
          for (int u=0; u<SUDOSIZE2; u++) {
            for (int v=0; v<SUDOSIZE2; v++) {
              if (table[SUDOSIZE2*(j/SUDOSIZE2)+u][SUDOSIZE2*(i/SUDOSIZE2)+v] == k) { kok = 0; break; }
            }
          }
          if (!kok) continue;
          // todo: some better checks
          // add k to the possibilities table
          tsudo[j][i] = (Sudotype*)realloc(tsudo[j][i], sizeof(Sudotype)*(tsudolen[j][i]+1));
          *(tsudo[j][i]+tsudolen[j][i]++) = k;
        }
      }
    }
  }
  
  // cunning cuts to possibility table
  /*
  for (j=0; j<SUDOSIZE2; j++) {
    for (i=0; i<SUDOSIZE2; i++) {
      for (k=1; k<=SUDOSIZE; k++) {
        for (int u=0; u<SUDOSIZE2; u++) {
          for (int v=0; v<SUDOSIZE2; v++) {
            table[j*SUDOSIZE2+u][i*SUDOSIZE2+v]
          }
        }
      }
    }
  }
  */
  
  // fills the sudoku according with possibility table
  int somedone = 1;
  while (!IsComplete() && somedone) {
    somedone = 0;
    for (j=0; j<SUDOSIZE; j++) {
      int irow, countrow = 0;
      for (i=0; i<SUDOSIZE; i++) {
        if (table[j][i] == 0) { countrow++; irow=i; }
        if (tsudolen[j][i] == 1) {
          somedone = 1;
          PutBit(i,j, *(tsudo[j][i]), tsudo, &tsudolen);
        }
      }
      // checks if there is only one free cell in a row
      if (countrow == 1) {
        if (table[j][irow] == 0) {
          if (tsudolen[j][irow] == 0) return(1); // not correct
          /* todo: fill */
        }
      }
    }
    
    if (somedone == 0) {
      SudoSolver *subsudo = new SudoSolver(sset, time0, fout);
      if (subsudo == NULL) exit(-2);
      // brute-force recursion
      for (j=0; j<SUDOSIZE; j++)
        for (i=0; i<SUDOSIZE; i++) {
          for (int u=0; u<tsudolen[j][i]; u++) {
            subsudo->FillSudoku(&table);
            subsudo->SetBit(i,j, *(tsudo[j][i]+u));
            int ret = subsudo->Solve();
            if (ret == 2) { // timeout
              delete subsudo;
              return(ret);
            }
          }
        }
      delete subsudo;
    }
  }

  if (somedone != 0) {
    if (!sset->IsPresent(&table)) {
      sset->Add(&table);
      PrintSudoku();
    }
  }
  return(0);
}
/*----------------------------------------------------------------------------*/

int showCmdline() {
  printf("Usage: sudoku [options] filein [fileout]\n"\
  	 "Options:\n"\
  	 "  -i        uses standard input as inputfile\n"\
  	 "  -t time   sets the timeout to time (in seconds)\n"\
  	 "  -h        shows this brief help\n\n");
}

int main(int argc, char *argv[]) {
  int i;
  Sudoku sudotab;
  FILE *fin = NULL;
  FILE *fout = stdout;

  // parses command line
  for (i=1; i<argc; i++) {
    if (*(argv[i]) == '-') { // it's an option
      switch (*(argv[i]+1)) {
        case 'h': showCmdline(); exit(0);
        case 'i': fin = stdin; break;
        case 't': {
          if (i < (argc-1)) {
            timeout = atoi(argv[++i]);
          } else {
            showCmdline();
            exit(0);
          }
          break;
        }
        case 's': /* todo */ break;
      }
    } else {
      if (fin == NULL) {
        fin = fopen(argv[i], "r");
      } else
        fout = fopen(argv[i], "w");
    }
  }
  if (fin == NULL) {
    showCmdline();
    exit(0);
  }
  if (fout == NULL) fout = stdout;
  
  for (i=0; i<SUDOSIZE*SUDOSIZE; i++) {
    int a;
    fscanf(fin, "%i", &a);
    *((Sudotype*)sudotab+i) = a;
  }
  fclose(fin);

  SudoSet *sudoset = new SudoSet();
  SudoSolver *sudo = new SudoSolver(sudoset, time(NULL), fout);
  sudo->FillSudoku(&sudotab);
  if (sudo->Solve() == 2) fprintf(fout, "Timeout reached.\n");
  fclose(fout);
  delete sudo;
  delete sudoset ;
}
